/*
 * ksi_regexp.c
 * regexp extension
 *
 * Copyright (C) 1998-2000,2001, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written
 * agreement is hereby granted, provided that the above copyright notice
 * and this paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Sat Jul  4 05:54:47 1998
 * Last Update:   Tue Jun 19 17:09:35 2001
 *
 */

/* Comments:
 *
 */

#include <ksi_conf.h>
#include <ksi.h>
#include <ksi_type.h>
#include <ksi_prim.h>
#include "ksi_regexp.h"
#include "pcre.h"

static char rcsid[] = "$Id: ksi_regexp.c,v 1.3 2001/06/19 10:12:02 demakov Exp $";


static struct Ksi_RE
{
  const unsigned char *table;
} *re;

struct Ksi_Regexp
{
  unsigned		itag;
  ksi_tag		etag;
  const char*		pattern;
  pcre*			regexp;
  pcre_extra*		extra;
  int			subnum;
};

typedef struct Ksi_Regexp	*ksi_regexp;


#define KSI_REGEXP_P(x)	KSI_EXT_IS((x), &tc_regexp)


static int
ksi_regexp_tag_equal (ksi_tag tag, ksi_obj x1, ksi_obj x2, int deep)
{
  return 0;
}

static const char*
ksi_regexp_tag_print (ksi_tag tag, ksi_obj x, int slashify)
{
  return ksi_aprintf ("#<regexp %s>", ((ksi_regexp) x) -> pattern);
}

static struct Ksi_Tag tc_regexp =
{
  "regexp",
  ksi_regexp_tag_equal,
  ksi_regexp_tag_print
};


ksi_obj
ksi_regexp_p (ksi_obj x)
{
  return KSI_REGEXP_P (x) ? ksi_true : ksi_false;
}


/*
 * ksi_make_regexp -- Compile a regexp.
 */

ksi_obj
ksi_make_regexp (ksi_obj str, ksi_obj opt)
{
  ksi_regexp x;
  pcre *regexp;
  const char *pattern;
  const char *errptr;
  int erroff, i, flg = 0;

  KSI_TYPE_CHECK (str, KSI_STR_P (str), "make-regexp: invalid string in arg1");
  if (opt != 0)
    {
      KSI_TYPE_CHECK (opt, KSI_STR_P (opt), "make-regexp: invalid string in arg2");
      for (i = 0; i < KSI_STR_LEN (opt); i++)
	switch (KSI_STR_PTR (opt) [i])
	  {
	  case 'i': flg |= PCRE_CASELESS;  break;
	  case 's': flg |= PCRE_DOTALL;    break;
	  case 'x': flg |= PCRE_EXTENDED;  break;
	  case 'm': flg |= PCRE_MULTILINE; break;
	  }
    }

  pattern = ksi_safe_string_ptr (str);
  regexp  = pcre_compile (pattern, flg, &errptr, &erroff, re->table);

  if (regexp == 0)
    ksi_exn_error ("misc", str, "make-regexp: %s", errptr);

  x = KSI_NEW_TYPED (struct Ksi_Regexp, pattern, extra);
  x->itag    = KSI_TAG_EXTENDED;
  x->etag    = &tc_regexp;
  x->pattern = pattern;
  x->regexp  = regexp;
  x->extra   = pcre_study (regexp, 0, &errptr);
  pcre_fullinfo (regexp, x->extra, PCRE_INFO_CAPTURECOUNT, &x->subnum);

  return (ksi_obj) x;
}


/*
 * ksi_regexp_exec -- Match a compiled regexp with a string.
 *
 * Return `ksi_false' if the string does not match the regexp
 * or a vector of 3 elements if the string matches.
 * First elementh of that vector is the matched string itself,
 * second and third are vectors of starting and ending indexes
 * of substrings within the string.
 * The first substring is a substring that matched the whole
 * regular expression.  The others are those substrings that
 * matched parenthesized expressions numbered in left-to-right
 * order.
 */

ksi_obj
ksi_regexp_exec (ksi_obj reg, ksi_obj str, ksi_obj beg, ksi_obj end)
{
  int bof, eof;

  KSI_TYPE_CHECK (reg, KSI_REGEXP_P(reg), "regexp-exec: invalid regexp in arg1");
  KSI_TYPE_CHECK (str, KSI_STR_P(str), "regexp-exec: invalid string in arg2");

  if (beg == 0)
    bof = 0;
  else
    {
      KSI_TYPE_CHECK (beg, KSI_SINT_P(beg), "regexp-exec: invalid integer in arg3");
      bof = KSI_SINT_COD (beg);
      KSI_RANGE_CHECK (beg, 0 <= bof, "regexp-exec: index out of range");
    }

  if (end == 0)
    eof = KSI_STR_LEN (str);
  else
    {
      KSI_TYPE_CHECK (end, KSI_SINT_P(end), "regexp-exec: invalid integer in arg4");
      eof = KSI_SINT_COD (end);
      KSI_RANGE_CHECK (end, bof <= eof, "regexp-exec: index out of range");

      if (eof > KSI_STR_LEN (str))
	eof = KSI_STR_LEN (str);
    }

  if (bof < eof)
    {
      ksi_regexp  exp = (ksi_regexp) reg;
      const char *ptr = KSI_STR_PTR (str) + bof;
      int         num = (exp->subnum + 1) * 3;
      int        *sub = (int*) alloca (num * sizeof *sub);

      num = pcre_exec (exp->regexp, exp->extra, ptr, eof - bof, 0, 0, sub, num);
      if (num > 0)
	{
	  int i;
	  ksi_obj v, v1, v2;

	  v1 = ksi_make_vector (KSI_MK_SINT (num), ksi_false);
	  v2 = ksi_make_vector (KSI_MK_SINT (num), ksi_false);
	  for (i = 0; i < num; i++)
	    if (sub[i*2] >= 0 && sub[i*2+1] >= 0)
	      {
		KSI_VEC_REF (v1, i) = KSI_MK_SINT (sub[i*2]);
		KSI_VEC_REF (v2, i) = KSI_MK_SINT (sub[i*2+1]);
	      }

	  v = ksi_make_vector (KSI_MK_SINT (3), ksi_false);
	  KSI_VEC_REF (v, 0) = ksi_str2string (ptr, eof - bof);
	  KSI_VEC_REF (v, 1) = v1;
	  KSI_VEC_REF (v, 2) = v2;
	  return v;
	}
    }

  return ksi_false;
}


ksi_obj
ksi_string_match (ksi_obj pattern, ksi_obj str, ksi_obj opt)
{
  ksi_obj exp = ksi_make_regexp (pattern, opt);
  return ksi_regexp_exec (exp, str, 0, 0);
}


static struct Ksi_Prim_Def defs [] =
{
  { "regexp?",		ksi_regexp_p,		KSI_CALL_ARG1, 1 },
  { "make-regexp",	ksi_make_regexp,	KSI_CALL_ARG2, 1 },
  { "regexp-exec",	ksi_regexp_exec,	KSI_CALL_ARG4, 2 },
  { "string-match",	ksi_string_match,	KSI_CALL_ARG3, 2 },

  { 0 }
};

void
ksi_init_regexp ()
{
  (void) rcsid;

  pcre_malloc = ksi_malloc_data;
  pcre_free   = ksi_free;

  re = (struct Ksi_RE*) ksi_malloc_eternal (sizeof *re);
  re->table = pcre_maketables ();

  ksi_reg_unit (defs);
}


void
ksi_term_regexp ()
{
  ksi_free (re);
  re = 0;
}


 /* End of code */
