/*
 * ksi_bool.c
 * booleans
 *
 * Copyright (C) 1998-2000,2001, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Sun Jan  4 21:27:23 1998
 * Last Update:   Wed Dec 12 21:05:16 2001
 *
 */

/* Comments:
 *
 */

#include "ksi_int.h"
#include "ksi_comp.h"
#include "ksi_klos.h"


static char rcsid[] = "$Id: ksi_bool.c,v 1.5 2002/01/15 04:47:13 ksion Exp $";


ksi_obj
ksi_bool_p (ksi_obj x)
{
  return KSI_BOOL_P (x) ? ksi_true : ksi_false;
}

ksi_obj
ksi_int2bool (int x)
{
  return x ? ksi_true : ksi_false;
}

int
ksi_bool2int (ksi_obj x)
{
  return x == ksi_false ? 0 : 1;
}

ksi_obj
ksi_not (ksi_obj x)
{
  return x == ksi_false ? ksi_true : ksi_false;
}

ksi_obj
ksi_eq_p (ksi_obj x1, ksi_obj x2)
{
  if (x1 == x2)
    return ksi_true;
  return ksi_false;
}


static inline int
eff_tag (ksi_obj x)
{
  int tag = x->itag;
  if (tag == KSI_TAG_CONST_STRING)
    return KSI_TAG_STRING;
  if (tag == KSI_TAG_CONST_VECTOR)
    return KSI_TAG_VECTOR;
  if (tag == KSI_TAG_CONST_PAIR)
    return KSI_TAG_PAIR;
  if (tag == KSI_TAG_LOCAL || tag == KSI_TAG_IMPORTED || tag == KSI_TAG_GLOBAL)
    return KSI_TAG_FREEVAR;

  return tag;
}

ksi_obj
ksi_eqv_p (ksi_obj x1, ksi_obj x2)
{
  if (x1 == x2)
    return ksi_true;

  if (KSI_OBJ_P (x1) && KSI_OBJ_P (x2))
    {
      if (eff_tag (x1) != eff_tag (x2))
	return ksi_false;

      /* x1->itag == x2->itag */

      if (x1->itag == KSI_TAG_EXTENDED)
	{
	  ksi_tag etag = KSI_EXT_TAG (x1);
	  return (etag -> equal (etag, x1, x2, 0) ? ksi_true : ksi_false);
	}

      switch (x1->itag)
	{
	case KSI_TAG_STRING:
	case KSI_TAG_CONST_STRING:
	  return ksi_string_eqv_p (x1, x2);

	case KSI_TAG_BIGNUM:
	case KSI_TAG_FLONUM:
	case KSI_TAG_RATIONAL:
	case KSI_TAG_COMPLEX:
	  return ksi_num_eqv_p (x1, x2);

	case KSI_TAG_INSTANCE:
	  return ksi_inst_eqv_p (x1, x2);
	}
    }

  return ksi_false;
}


static int
frame_equal (ksi_frame f1, ksi_frame f2)
{
  int i;

again:
  if (f1 == f2)
    return 1;

  if (!f1 || !f2)
    return 0;
  if (f1->num != f2->num)
    return 0;
  if (f1->env != f2->env)
    return 0;

  for (i = 0; i < f1->num; i++)
    if (ksi_equal_p (f1->vals[i], f2->vals[i]) == ksi_false)
      return 0;

  f1 = f1->next;
  f2 = f2->next;
  goto again;
}

ksi_obj
ksi_equal_p (ksi_obj x1, ksi_obj x2)
{
  int i;
  ksi_obj z1, z2;

again:
  if (x1 == x2)
    return ksi_true;

  if (KSI_OBJ_P (x1) && KSI_OBJ_P (x2))
    {
      if (eff_tag (x1) != eff_tag (x2))
	return ksi_false;

      /* x1->itag == x2->itag */

      if (x1->itag == KSI_TAG_EXTENDED)
	{
	  ksi_tag etag = KSI_EXT_TAG (x1);
	  return (etag -> equal (etag, x1, x2, 1) ? ksi_true : ksi_false);
	}

      switch (x1->itag)
	{
	case KSI_TAG_STRING:
	case KSI_TAG_CONST_STRING:
	  return ksi_string_equal_p (x1, x2);

	case KSI_TAG_BIGNUM:
	case KSI_TAG_FLONUM:
	case KSI_TAG_RATIONAL:
	case KSI_TAG_COMPLEX:
	  return ksi_num_equal_p (x1, x2);

	case KSI_TAG_PAIR:
	case KSI_TAG_CONST_PAIR:
	  for (i = 0, z1 = x1, z2 = x2; ; )
	    {
	      if (ksi_equal_p (KSI_CAR (x1), KSI_CAR (x2)) == ksi_false)
		return ksi_false;

	      x1 = KSI_CDR (x1);
	      x2 = KSI_CDR (x2);

	      /* if cdr's are equal then lists equal, if both lists end,
	       * x1 and x2 are ksi_nil (and equal too).
	       */
	      if (x1 == x2)
		return ksi_true;

	      if (!KSI_PAIR_P (x1) || !KSI_PAIR_P (x2))
		goto again;

	      if (ksi_equal_p (KSI_CAR (x1), KSI_CAR (x2)) == ksi_false)
		return ksi_false;

	      x1 = KSI_CDR (x1);
	      x2 = KSI_CDR (x2);

	      if (x1 == x2)
		return ksi_true;

	      if (!KSI_PAIR_P (x1) || !KSI_PAIR_P (x2))
		goto again;

	      z1 = KSI_CDR (z1);
	      z2 = KSI_CDR (z2);

	      /* if both lists are circular they are equal */
	      if (z1 == x1) i |= 1;
	      if (z2 == x2) i |= 2;
	      if (i == 3)
		return ksi_true;
	      CHECK_EVENTS;
	    }

	case KSI_TAG_VECTOR:
	case KSI_TAG_CONST_VECTOR:
	case KSI_TAG_CORE:
	  if (KSI_VEC_LEN (x1) == KSI_VEC_LEN (x2))
	    {
	      for (i = 0; i < KSI_VEC_LEN (x1) - 1; i++)
		{
		  if (ksi_equal_p (KSI_VEC_REF (x1, i), KSI_VEC_REF (x2, i)) == ksi_false)
		    return ksi_false;
		  CHECK_EVENTS;
		}
	      x1 = KSI_VEC_REF (x1, i);
	      x2 = KSI_VEC_REF (x2, i);
	      goto again;
	    }
	  return ksi_false;

	case KSI_TAG_INSTANCE:
	  return ksi_inst_equal_p (x1, x2);

	case KSI_TAG_CLOSURE:
	  if (CLOS_NUMS (x1) == CLOS_NUMS (x2)
	      && CLOS_NARY (x1) == CLOS_NARY (x2)
	      && CLOS_OPTS (x1) == CLOS_OPTS (x2)
	      && frame_equal (CLOS_FRM (x1), CLOS_FRM (x2)))
	    {
	      x1 = CLOS_BODY (x1);
	      x2 = CLOS_BODY (x2);
	      goto again;
	    }
	  return ksi_false;

	case KSI_TAG_VAR0:
	case KSI_TAG_VAR1:
	case KSI_TAG_VAR2:
	case KSI_TAG_VARN:
	  if (VAR_NUM (x1) == VAR_NUM (x2) && VAR_LEV (x1) == VAR_LEV (x2))
	    return ksi_true;
	  return ksi_false;

	case KSI_TAG_FREEVAR:
	case KSI_TAG_LOCAL:
	case KSI_TAG_IMPORTED:
	case KSI_TAG_GLOBAL:
	  if (VAR_SYM (x1) == VAR_SYM (x2))
	    return ksi_true;
	  return ksi_false;

	default:
	  if (KSI_TAG_FIRST_CODE <= x1->itag && x1->itag <= KSI_TAG_LAST_CODE
	      && CODE_NUM (x1) == CODE_NUM (x2))
	    {
	      for (i = 0; i < CODE_NUM (x1); ++i)
		{
		  if (ksi_equal_p (CODE_VAL (x1, i),
				   CODE_VAL (x2, i)) == ksi_false)
		    return ksi_false;
		  CHECK_EVENTS;
		}
	      x1 = CODE_VAL (x1, i);
	      x2 = CODE_VAL (x2, i);
	      goto again;
	    }
	}
    }

  return ksi_false;
}


static struct Ksi_Prim_Def defs[] =
{
  { "boolean?",		ksi_bool_p,		KSI_CALL_ARG1, 1 },

  { 0 }
};

void
ksi_init_bools ()
{
  ksi_ignore (rcsid);
  ksi_reg_unit (defs);

  ksi_defsym ("not", ksi_new_imm (KSI_TAG_NOT), 0);
  ksi_defsym ("eq?", ksi_new_imm (KSI_TAG_EQP), 0);
  ksi_defsym ("eqv?", ksi_new_imm (KSI_TAG_EQVP), 0);
  ksi_defsym ("equal?", ksi_new_imm (KSI_TAG_EQUALP), 0);
}

 /* End of code */
