/*
 * ksi_comp.c
 * compilation
 *
 * Copyright (C) 1997-2000,2001, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Tue Oct 28 16:06:31 1997
 * Last Update:   Sat Nov 10 00:39:30 2001
 *
 */

/* Comments:
 *
 */

#include "ksi_int.h"
#include "ksi_comp.h"
#include "ksi_mod.h"


static char rcsid[] = "$Id: ksi_comp.c,v 1.8 2001/11/25 14:30:01 demakov Exp $";


struct Ksi_RefInfo;
struct Ksi_VarInfo;
struct Ksi_EnvInfo;
struct Ksi_FrmInfo;

typedef struct Ksi_RefInfo	*ksi_refinfo;
typedef struct Ksi_DepInfo	*ksi_depinfo;
typedef struct Ksi_VarInfo	*ksi_varinfo;
typedef struct Ksi_EnvInfo	*ksi_envinfo;
typedef struct Ksi_FrmInfo	*ksi_frminfo;


struct Ksi_RefInfo
{
  ksi_refinfo	next;
  ksi_obj	var;
  ksi_obj	exp;
};

struct Ksi_DepInfo
{
  ksi_depinfo	next;
  ksi_varinfo	var;
  int		lev;
};

struct Ksi_VarInfo
{
  ksi_obj	name;
  ksi_obj	defs;
  ksi_refinfo	refs;
  ksi_depinfo	deps;
  int		idx;
  int		nloc;   /* non-local var -- ref'ed from internal frame */
};

struct Ksi_EnvInfo
{
  ksi_envinfo	next;
  ksi_envinfo	down;
  ksi_envinfo	link;
  ksi_frminfo	frm;
  int		nums;
  ksi_varinfo	vars;
  ksi_varinfo	curvar;
};

struct Ksi_FrmInfo
{
  ksi_frminfo	next;
  ksi_envinfo	envs;
  ksi_envinfo	*used;
  int		free;
};


static const char *syntax_s = "syntax";

static ksi_obj ksi_comp_sexp (ksi_obj form, ksi_envinfo env);
static ksi_obj ksi_comp_apply (ksi_obj form, ksi_envinfo env);


ksi_frame
ksi_new_frame (int n, ksi_frame old_frm)
{
  ksi_frame new_frm = NEW_TYPED_ARR (struct Ksi_Frame, n, next, vals);
  new_frm->num  = n;
  new_frm->next = old_frm;
  new_frm->env  = (old_frm ? old_frm->env : ksi_current_env ());

  return new_frm;
}

ksi_obj
ksi_new_code (int n, int tag)
{
  ksi_code code = NEW_TYPED_ARR (struct Ksi_Code, n, val, val);
  code->itag = tag;
  code->num  = n - 1;
  return (ksi_obj) code;
}

ksi_obj
ksi_new_quote (ksi_obj x)
{
  ksi_code q = (ksi_code) ksi_new_code (1, KSI_TAG_QUOTE);
  q -> val [0] = x;
  return (ksi_obj) q;
}

ksi_obj
ksi_new_freevar (ksi_obj sym)
{
  ksi_var var = (ksi_var) ksi_malloc_data (sizeof *var);
  var->itag = KSI_TAG_FREEVAR;
  var->var.sym = sym;
  var->var.val = 0;
  return (ksi_obj) var;
}

ksi_obj
ksi_new_varbox (int lev, int num)
{
  ksi_varbox var = (ksi_varbox) ksi_malloc_data (sizeof *var);
  var->lev = lev;
  var->num = num;
  var->itag = (lev == 0 ? KSI_TAG_VAR0 :
	       lev == 1 ? KSI_TAG_VAR1 :
	       lev == 2 ? KSI_TAG_VAR2 : KSI_TAG_VARN);

  return (ksi_obj) var;
}

ksi_obj
ksi_new_closure (int nums, int nary, int opts, ksi_frame frm,
		 ksi_obj body, ksi_obj doc)
{
  ksi_obj val;

  val = (ksi_obj) NEW_TYPED (struct Ksi_Closure, frm, body);
  val->itag = KSI_TAG_CLOSURE;
  CLOS_NUMS (val) = nums;
  CLOS_NARY (val) = nary;
  CLOS_OPTS (val) = opts;
  CLOS_FRM  (val) = frm ? frm : ksi_new_frame (0, 0);
  CLOS_DOC  (val) = doc ? doc : ksi_false;
  CLOS_BODY (val) = body;

  return val;
}


static ksi_obj
mk_const (ksi_obj x)
{
  int i;
  ksi_obj tmp;

  if (!KSI_OBJ_P (x))
    return x;

  switch (x->itag)
    {
    case KSI_TAG_STRING:
      tmp = ksi_str2string (KSI_STR_PTR (x), KSI_STR_LEN (x));
      tmp->itag = KSI_TAG_CONST_STRING;
      return tmp;

    case KSI_TAG_VECTOR:
      tmp = ksi_alloc_vector (KSI_VEC_LEN (x), KSI_TAG_CONST_VECTOR);
      for (i = 0; i < KSI_VEC_LEN (x); i++)
	KSI_VEC_REF (tmp, i) = mk_const (KSI_VEC_REF (x, i));
      return tmp;
    }

  return x;
}

static ksi_obj
ksi_comp_quote (ksi_obj x, ksi_envinfo env)
{
  x = mk_const (x);

  if (KSI_OBJ_P (x))
    {
      if (KSI_TAG_FIRST_CONST <= x->itag && x->itag <= KSI_TAG_LAST_CONST)
	return x;
    }

  return ksi_new_quote (x);
}

static ksi_obj
ksi_const_val (ksi_obj x)
{
  if (x->itag == KSI_TAG_QUOTE)
    return x;
  if (KSI_TAG_FIRST_CONST <= x->itag && x->itag <= KSI_TAG_LAST_CONST)
    return x;
  return 0;
}

static void
append_dep (ksi_varinfo var, int lev, ksi_depinfo *list)
{
  ksi_depinfo dep;

  for (dep = *list; dep; dep = dep->next)
    if (dep->var == var)
	return;

  dep = (ksi_depinfo) ksi_malloc (sizeof *dep);
  dep->next = *list;
  dep->var  = var;
  dep->lev  = lev;
  *list = dep;
}

static void
remove_dep (ksi_varinfo var, ksi_depinfo *list)
{
  for (/**/; *list; list = &(*list)->next)
    if ((*list)->var == var)
      {
	*list = (*list)->next;
	return;
      }
}

static int
weak_dep (ksi_varinfo def, ksi_varinfo ref)
{
  ksi_depinfo dep;

  if (def == ref)
    return 1;

  for (dep = ref->deps; dep; dep = dep->next)
    if (def == dep->var && dep->lev > 0)
      {
	remove_dep (def, &ref->deps);
	return 1;
      }

  return 0;
}

static ksi_obj
ksi_comp_sym (ksi_obj x, ksi_envinfo env, ksi_obj exp)
{
  ksi_frminfo	frm;
  ksi_refinfo	ref;
  ksi_obj	var;
  int		lev, i;

  for (frm = env->frm; env; env = env->next)
    for (i = 0; i < env->nums; i++)
      if (env->vars[i].name == x && env->vars[i].idx >= 0)
	{
	  for (lev = 0; frm != env->frm; frm = frm->next, lev++)
	    {
	      if (frm == 0)
		ksi_error ("compile: internal error");
	    }

	  var = ksi_new_varbox (lev, env->vars[i].idx);

	  ref = (ksi_refinfo) ksi_malloc (sizeof *ref);
	  ref->next = env->vars[i].refs;
	  ref->var  = var;
	  ref->exp  = exp;
	  env->vars[i].refs = ref;
	  if (lev > 0)
	    env->vars[i].nloc = 1;

	  /* for sorting letrec bindings */
	  if (env->curvar)
	    if (lev == 0 || !weak_dep (env->curvar, &env->vars[i]))
	      append_dep (&env->vars[i], lev, &env->curvar->deps); 

	  return var;
	}

  return ksi_new_freevar (x);
}

static void
ksi_del_ref (ksi_obj x, ksi_envinfo env)
{
  ksi_refinfo	*lst;
  int		i;

  for (/**/; env; env = env->next)
    for (i = 0; i < env->nums; i++)
      for (lst = &env->vars[i].refs; *lst; lst = &(*lst)->next)
	if ((*lst)->var == x)
	  {
	    *lst = (*lst)->next;
	    return;
	  }
}


static ksi_obj
ksi_comp_seq (ksi_obj seq, ksi_envinfo env)
{
  int		i, len;
  ksi_obj	x, res;

  for (res = ksi_nil, len = 0; seq != ksi_nil; seq = CDR (seq))
    {
      x = ksi_comp_sexp (CAR (seq), env);
      if (x->itag == KSI_TAG_BEGIN)
	{
	  for (i = 0; i <= CODE_NUM (x); i++)
	    {
	      if (CDR (seq) != ksi_nil && LOCVAR_P (CODE_VAL (x, i)))
		ksi_del_ref (CODE_VAL (x, i), env);
	      else if (CDR (seq) == ksi_nil
		       || !(ksi_const_val (CODE_VAL (x, i))
			    || FREEVAR_P (CODE_VAL (x, i))))
		{
		  res = CONS (CODE_VAL (x, i), res);
		  len += 1;
		}
	    }
	}
      else if (CDR (seq) != ksi_nil && LOCVAR_P (x))
	ksi_del_ref (x, env);
      else if (CDR (seq) == ksi_nil || !(ksi_const_val (x) || FREEVAR_P (x)))
	{
	  res = CONS (x, res);
	  len += 1;
	}
    }

  if (len == 0)
    return ksi_comp_quote (ksi_void, env);

  if (len == 1)
    return CAR (res);

  x = ksi_new_code (len, KSI_TAG_BEGIN);
  while (--len >= 0)
    {
      CODE_VAL (x, len) = CAR (res);
      res = CDR (res);
    }

  return x;
}


/* ksi_comp_define -- compile define expression
 *
 * <form>	::= (define <sym> <expr>)
 *		  | (define (<sym> <formals>) <body>)
 *
 * Note: definitions on top level(s) allowed only.
 */

static ksi_obj
ksi_comp_define (ksi_obj form, ksi_envinfo env)
{
  ksi_obj x, sym, r;

  x = CDR (form);
  KSI_EXN_CHECK (syntax_s, form, PAIR_P (x), "define: missing variable");
  KSI_EXN_CHECK (syntax_s, form, env->frm->next == 0,
		 "define: invalid context for definition");

  sym = CAR (x);
  x = CDR (x);
  if (x == ksi_nil)
    {
      KSI_EXN_CHECK (syntax_s, sym, SYM_P (sym), "define: invalid variable");
      x = ksi_void;
    }
  else
    {
      while (PAIR_P (sym))
	{
	  x = LIST1 (CONS (KSI->sym_lambda, CONS (CDR (sym), x)));
	  sym = CAR (sym);
	}
      KSI_EXN_CHECK (syntax_s, sym, SYM_P (sym), "define: invalid variable");
      KSI_EXN_CHECK (syntax_s, form, PAIR_P (x), "define: missing value");

      x = CAR (x);
    }

  r = ksi_new_code (2, KSI_TAG_DEFINE);
  CODE_VAL (r, 0) = sym;
  CODE_VAL (r, 1) = ksi_comp_sexp (x, env);
  return r;
}


/* ksi_comp_set -- compile set! expression
 *
 * form		::= (set! <sym> <expr>)
 */

static ksi_obj
ksi_comp_set (ksi_obj form, ksi_envinfo env)
{
  ksi_obj	val, x, sym;

  x = CDR (form);
  KSI_EXN_CHECK (syntax_s, form, PAIR_P (x), "set!: missing variable");

  sym = CAR (x);
  KSI_EXN_CHECK (syntax_s, form, SYM_P (sym), "set!: invalid variable");

  x = CDR (x);
  KSI_EXN_CHECK (syntax_s, form, PAIR_P (x), "set!: missing value");
  KSI_CHECK2 (form, CDR (x) == ksi_nil, "set!: extra args");

  val = ksi_new_code (2, KSI_TAG_SET);
  CODE_VAL (val, 0) = ksi_comp_sym (sym, env, val);
  CODE_VAL (val, 1) = ksi_comp_sexp (CAR (x), env);

  return val;
}


static int
alloc_vars (ksi_envinfo env, int free, int nloc)
{
  ksi_varinfo	var;
  ksi_refinfo	ref;
  int		i, used, size;

  if (nloc)
    {
      for (/**/; env; env = env->link)
	{
	  for (i = 0, var = env->vars; i < env->nums; i++, var++)
	    if (var->nloc && var->refs)
	      {
		for (ref = var->refs; ref; ref = ref->next)
		  VAR_NUM (ref->var) = free;
		free++;
	      }

	  free = alloc_vars (env->down, free, 1);
	}

      return free;
    }

  size = free;
  for (/**/; env; env = env->link)
    {
      used = free;
      for (i = 0, var = env->vars; i < env->nums; i++, var++)
	if (!var->nloc && var->refs)
	  {
	    for (ref = var->refs; ref; ref = ref->next)
	      VAR_NUM (ref->var) = used;
	    used++;
	  }

      used = alloc_vars (env->down, used, 0);

      if (size < used)
	size = used;
    }

  return size;
}


static int
realloc_vars (ksi_frminfo frm, int free)
{
  free = alloc_vars (frm->envs, free, 1);
  free = alloc_vars (frm->envs, free, 0);
  return free;
}


/* ksi_comp_lambda -- compile lambda expression
 *
 * <form>	::= (lambda <formals> <body>)
 */

static ksi_obj
ksi_comp_lambda (ksi_obj form, ksi_envinfo old_env)
{
  ksi_envinfo		new_env;
  struct Ksi_FrmInfo	new_frm;
  ksi_obj		formals, body, x, doc;
  int			len, nums, i, opts;

  len = LIST_LEN (form);
  KSI_EXN_CHECK (syntax_s, form, len >= 3, "lambda: invalid syntax");

  formals = CAR (CDR (form));
  body = CDR (CDR (form));

  opts = nums = 0;
  for (x = formals; x != ksi_nil; x = CDR (x), nums++)
    {
      if (SYM_P (x))
	{
	  nums += 1;
	  opts = 1;
	  break;
	}
      KSI_EXN_CHECK (syntax_s, form, PAIR_P (x) && SYM_P (CAR (x)),
		     "lambda: invalid formals");
    }

  new_env = (ksi_envinfo) ksi_malloc (sizeof *new_env);
  new_env->next = old_env;
  new_env->frm  = &new_frm;
  new_env->nums = nums;
  new_env->vars = (ksi_varinfo) ksi_malloc (nums * sizeof *new_env->vars);

  new_frm.next = old_env->frm;
  new_frm.envs = 0;
  new_frm.used = &new_frm.envs;
  new_frm.free = nums;

  for (i = 0, x = formals; i < nums; i++)
    {
      new_env->vars[i].idx  = i;
      new_env->vars[i].nloc = 1;

      if (SYM_P (x))
	{
	  new_env->vars[i].name = x;
	  break;
	}
      else
	{
	  new_env->vars[i].name = CAR (x);
	  x = CDR (x);
	}
    }

  doc = CAR (body);
  while (PAIR_P (doc) && CAR (doc) == KSI->sym_begin)
    doc = CAR (CDR (doc));
  if (PAIR_P (doc) && CAR (doc) == KSI->sym_quote)
    doc = CAR (CDR (doc));

  body = ksi_comp_seq (body, new_env);
  if (!KSI_STR_P (doc) || KSI_EQUAL_P (doc, body))
    doc = ksi_false;

  len = realloc_vars (&new_frm, nums);

  x = ksi_new_code (5, KSI_TAG_LAMBDA);
  CODE_VAL (x, 0) = KSI_MK_SINT (len);
  CODE_VAL (x, 1) = KSI_MK_SINT (nums);
  CODE_VAL (x, 2) = KSI_MK_SINT (opts);
  CODE_VAL (x, 3) = doc;
  CODE_VAL (x, 4) = body;

  return x;
}


static int
split_bindings (char *name, ksi_obj bind, ksi_obj *vars, ksi_obj *init)
{
  int		num;
  ksi_obj	x;

  *vars = ksi_nil;
  *init = ksi_nil;
  for (num = 0; bind != ksi_nil; num++, bind = CDR (bind))
    {
      if (!PAIR_P (bind))
	ksi_exn_error (syntax_s, bind, "%s: invalid binding", name);

      x = CAR (bind);
      if (LIST_LEN (x) != 2 || !SYM_P (CAR (x)))
	ksi_exn_error (syntax_s, x, "%s: invalid binding", name);

      *vars = CONS (CAR (x), ksi_nil);
      *init = CONS (CAR (CDR (x)), ksi_nil);
      vars = & CDR (*vars);
      init = & CDR (*init);
    }

  return num;
}

static int
var_asigned (ksi_refinfo refs)
{
  for (/**/; refs; refs = refs->next)
    if (refs->exp)
      return 1;

  return 0;
}

static ksi_varinfo
find_var (ksi_obj var, ksi_envinfo env)
{
  ksi_frminfo	frm;
  int		i;

  for (i = 0, frm = env->frm; i < VAR_LEV (var); i++, frm = frm->next)
    {
      if (frm == 0)
	ksi_error ("find_var: internal error");
    }

  for (/**/; env; env = env->next)
    if (env->frm == frm)
      break;

  for (/**/; env; env = env->next)
    for (i = 0; i < env->nums; i++)
      if (env->vars[i].idx == VAR_NUM (var))
	return &env->vars[i];

  ksi_error ("find_var: internal error");
  return 0;
}

static void
inline_var (ksi_varinfo var, ksi_envinfo env)
{
  if (var->defs && !var_asigned (var->refs))
    {
      if (var->defs->itag == KSI_TAG_QUOTE
	  && !KSI_OBJ_P (QUOTE_VAL (var->defs)))
	{
	  while (var->refs)
	    {
	      var->refs->var->itag = KSI_TAG_QUOTE;
	      CODE_NUM (var->refs->var) = 0;
	      CODE_VAL (var->refs->var, 0) = QUOTE_VAL (var->defs);
	      var->refs = var->refs->next;
	    }
	}
      else if (LOCVAR_P (var->defs))
	{
	  ksi_refinfo	tmp;
	  ksi_varinfo	def = find_var (var->defs, env);
	  int		lev;

	  if (!var_asigned (def->refs))
	    while (var->refs)
	      {
		lev = VAR_LEV (var->refs->var) + VAR_LEV (var->defs);
		var->refs->var->itag = (lev == 0 ? KSI_TAG_VAR0 :
					lev == 1 ? KSI_TAG_VAR1 :
					lev == 2 ? KSI_TAG_VAR2 :
					KSI_TAG_VARN);
		VAR_NUM (var->refs->var) = VAR_NUM (var->defs);
		VAR_LEV (var->refs->var) = lev;

		tmp = var->refs;
		var->refs = tmp->next;
		tmp->next = def->refs;
		def->refs = tmp;
		if (lev > 0)
		  def->nloc = 1;
	      }
	}
    }
}

static ksi_obj
gen_let_body (ksi_obj form, ksi_obj body, ksi_envinfo env)
{
  int		i, len, n, num, prev_num;
  ksi_obj	x, res;
  ksi_refinfo	ref;

  for (i = 0; i < env->nums; i++)
    inline_var (&env->vars[i], env);

  len = 0; res = ksi_nil;
  for (num = 0; num < env->nums; /**/)
    {
      prev_num = num;
      for (i = 0; i < env->nums; i++)
	{
	  if (!(x = env->vars[i].defs))
	    continue;
	    
	  if (env->vars[i].deps)
	    continue;

	  if (x->itag == KSI_TAG_LAMBDA || x->itag == KSI_TAG_DELAY)
	    {
	      if (env->vars[i].refs == 0)
	        goto next;
	      goto do_set;
	    }

	  if (env->vars[i].refs == 0)
	    {
	      //x = env->vars[i].defs;
	      if (x->itag == KSI_TAG_BEGIN)
		{
		  for (i = 0; i < CODE_NUM (x); i++)
		    {
		      res = CONS (CODE_VAL (x, i), res);
		      len += 1;
		    }

		  x = CODE_VAL (x, i);
		}

	      if (ksi_const_val (x) || LOCVAR_P (x) || FREEVAR_P (x))
		goto next;
	    }
	  else
	    {
	    do_set:
	      x = ksi_new_code (2, KSI_TAG_SET);
	      CODE_VAL (x, 0) = ksi_new_varbox (0, env->vars[i].idx);
	      CODE_VAL (x, 1) = env->vars[i].defs;

	      ref = (ksi_refinfo) ksi_malloc (sizeof *ref);
	      ref->next = env->vars[i].refs;
	      ref->var  = CODE_VAL (x, 0);
	      ref->exp  = 0;
	      env->vars[i].refs = ref;
	    }

	  res = CONS (x, res);
	  len += 1;

	next:
	  num += 1;
	  env->vars[i].defs = 0;
	  for (n = 0; n < env->nums; n++)
	    remove_dep (&env->vars[i], &env->vars[n].deps);
	}

      if (prev_num == num)
	ksi_exn_error (syntax_s, form, "letrec: invalid bindings");
    }

  if (body->itag != KSI_TAG_BEGIN)
    {
      res = CONS (body, res);
      len += 1;
    }
  else
    {
      for (i = 0; i <= CODE_NUM (body); i++)
	{
	  res = CONS (CODE_VAL (body, i), res);
	  len += 1;
	}
    }

  if (len == 1)
    return CAR (res);

  x = ksi_new_code (len, KSI_TAG_BEGIN);
  while (--len >= 0)
    {
      CODE_VAL (x, len) = CAR (res);
      res = CDR (res);
    }

  return x;
}


/* ksi_comp_letrec
 *
 * <form>	::= (letrec (<binding> ...) <expr> <expr> ...)
 * <binding>	::= (<var> <init>)
 */

static ksi_obj
ksi_comp_letrec (ksi_obj form, ksi_envinfo old_env)
{
  ksi_envinfo	new_env;
  ksi_obj	bind, body, vars, init;
  int		i, len, nums, free;

  len = LIST_LEN (form) - 2;
  KSI_EXN_CHECK (syntax_s, form, len >= 1, "letrec: invalid syntax");

  bind = CAR (CDR (form));
  body = CDR (CDR (form));
  nums = split_bindings ("letrec", bind, &vars, &init);

  if (nums == 0)
    return ksi_comp_seq (body, old_env);

  new_env = (ksi_envinfo) ksi_malloc (sizeof *new_env);
  new_env->next = old_env;
  new_env->frm  = old_env->frm;
  new_env->nums = nums;
  new_env->vars = (ksi_varinfo) ksi_malloc (nums * sizeof *new_env->vars);

  *new_env->frm->used = new_env;
  new_env->frm->used  = &new_env->down;

  free = new_env->frm->free;
  new_env->frm->free = free + nums;

  for (i = 0; i < nums; i++, vars = CDR (vars))
    {
      new_env->vars[i].name = CAR (vars);
      new_env->vars[i].idx  = i + free;
      new_env->vars[i].nloc = 0;
    }

  for (i = 0; i < nums; i++, init = CDR (init))
    {
      new_env->curvar = &new_env->vars[i];
      new_env->vars[i].defs = ksi_comp_sexp (CAR (init), new_env);
      new_env->curvar = 0;
    }

  body = ksi_comp_seq (body, new_env);

  new_env->frm->used  = &new_env->link;

  return gen_let_body (form, body, new_env);
}


/* ksi_comp_let
 *
 * <form>	::= (let (<binding> ...) <expr> <expr> ...)
 *		  | (let <symbol> (<binding> ...) <expr> <expr> ...)
 * <bindings>	::= (<var> <init>)
 */

static ksi_obj
ksi_comp_let (ksi_obj form, ksi_envinfo old_env)
{
  ksi_envinfo	new_env;
  ksi_obj	x, bind, body, vars, init;
  int		i, len, nums, free;

  len = LIST_LEN (form) - 2;
  KSI_EXN_CHECK (syntax_s, form, len >= 1, "let: invalid syntax");

  bind = CAR (CDR (form));
  body = CDR (CDR (form));

  if (SYM_P (bind))
    {
      ksi_obj proc = bind;
  
      KSI_EXN_CHECK (syntax_s, form, PAIR_P (body), "let: invalid syntax");
      bind = CAR (body);
      body = CDR (body);
      split_bindings ("let", bind, &vars, &init);

#ifdef __WATCOMC__
      x = CONS (KSI->sym_lambda, CONS (vars, body));
      x = LIST2 (proc, x);
      x = LIST3 (KSI->sym_letrec, LIST1 (x), proc);
      x = CONS (x, init);
#else
      x = CONS (LIST3 (KSI->sym_letrec,
		       LIST1 (LIST2 (proc,
				     CONS (KSI->sym_lambda,
					   CONS (vars, body)))),
		       proc),
		init);
#endif

      return ksi_comp_apply (x, old_env);
    }

  nums = split_bindings ("let", bind, &vars, &init);

  if (nums == 0)
    return ksi_comp_seq (body, old_env);

  new_env = (ksi_envinfo) ksi_malloc (sizeof *new_env);
  new_env->next = old_env;
  new_env->frm  = old_env->frm;
  new_env->nums = nums;
  new_env->vars = (ksi_varinfo) ksi_malloc (nums * sizeof *new_env->vars);

  *new_env->frm->used = new_env;
  new_env->frm->used  = &new_env->down;

  free = new_env->frm->free;
  new_env->frm->free = free + nums;

  for (i = 0; i < nums; i++, vars = CDR (vars))
    {
      new_env->vars[i].name = CAR (vars);
      new_env->vars[i].idx  = -1;
      new_env->vars[i].nloc = 0;
    }

  for (i = 0; i < nums; i++, init = CDR (init))
    new_env->vars[i].defs = ksi_comp_sexp (CAR (init), new_env);

  for (i = 0; i < nums; i++)
    new_env->vars[i].idx = i + free;

  body = ksi_comp_seq (body, new_env);

  new_env->frm->used = &new_env->link;

  return gen_let_body (form, body, new_env);
}


static ksi_obj
ksi_comp_apply (ksi_obj form, ksi_envinfo env)
{
  ksi_obj x;
  int i, len = LIST_LEN (form);
  KSI_EXN_CHECK (syntax_s, form, len >= 1, "@compile: invalid syntax");

  if (PAIR_P (CAR (form)) && CAR (CAR (form)) == KSI->sym_lambda)
    {
      ksi_obj formals, body, args, bind;

      /* ((lambda (f ...) body ...) e ...)
       *   ==>
       * (let ((f e) ...) body ...)
       */

      body = CAR (form);
      KSI_EXN_CHECK (syntax_s, body, LIST_LEN (body) >= 3,
		     "lambda: invalid syntax");

      formals = CAR (CDR (body));
      body = CDR (CDR (body));
      args = CDR (form);

      for (bind = ksi_nil; /**/; formals = CDR (formals), args = CDR (args))
	{
	  if (formals == ksi_nil)
	    {
	      WNA_CHECK (args == ksi_nil, 0, ksi_obj2str (CAR (form)));
	      break;
	    }

	  if (SYM_P (formals))
	    {
	      if (args == ksi_nil)
		bind = CONS (LIST2 (formals, LIST2 (KSI->sym_quote, ksi_nil)),
			     bind);
	      else
		bind = CONS (LIST2 (formals, CONS (KSI->sym_list, args)),
			     bind);
	      break;
	    }

	  KSI_EXN_CHECK (syntax_s, CAR (form),
			 PAIR_P (formals) && SYM_P (CAR (formals)),
			 "lambda: invalid syntax");
	  WNA_CHECK (PAIR_P (args), 0, ksi_obj2str (CAR (form)));

	  bind = CONS (LIST2 (CAR (formals), CAR (args)),
		       bind);
	}

      form = CONS (KSI->sym_let, CONS (bind, body));
      return ksi_comp_let (form, env);
    }

  x = ksi_new_code (len, KSI_TAG_CALL);
  for (i = 0; i < len; i++, form = CDR (form))
    CODE_VAL (x, i) = ksi_comp_sexp (CAR (form), env);

  return x;
}


/* ksi_comp_delay -- compile delay syntax.
 *
 * form		::= (delay <expr>)
 */

static ksi_obj
ksi_comp_delay (ksi_obj form, ksi_envinfo old_env)
{
  struct Ksi_FrmInfo	new_frm;
  struct Ksi_EnvInfo	new_env;
  ksi_obj		x, e;
  int			len;

  KSI_EXN_CHECK (syntax_s, form,
		 LIST_LEN (form) == 2,
		 "delay: invalid syntax");

  new_frm.next = old_env->frm;
  new_frm.envs = 0;
  new_frm.used = &new_frm.envs;
  new_frm.free = 0;

  new_env.next = old_env;
  new_env.down = 0;
  new_env.link = 0;
  new_env.frm  = &new_frm;
  new_env.nums = 0;
  new_env.vars = 0;

  e = ksi_comp_sexp (CAR (CDR (form)), &new_env);
  x = ksi_const_val (e);
  if (x == 0)
    {
      len = realloc_vars (&new_frm, 0);
      x = ksi_new_code (2, KSI_TAG_DELAY);
      CODE_VAL (x, 0) = KSI_MK_SINT (len);
      CODE_VAL (x, 1) = e;
    }

  return x;
}

/* ksi_comp_if -- compile if expression
 *
 * form		::= (if <test-expr> <then-expr>)
 *		  | (if <test-expr> <then-expr> <else-expr>)
 */

static ksi_obj
ksi_comp_if (ksi_obj form, ksi_envinfo env)
{
  int i, len;
  ksi_obj x, args = CDR (form);

  len = LIST_LEN (args);
  KSI_EXN_CHECK (syntax_s, form, len == 2 || len == 3, "if: invalid syntax");

  x = ksi_new_code (len, len == 2 ? KSI_TAG_IF2 : KSI_TAG_IF3);
  for (i = 0; i < len; i++, args = CDR (args))
    CODE_VAL (x, i) = ksi_comp_sexp (CAR (args), env);

  return x;
}


static ksi_obj
ksi_comp_and (ksi_obj form, ksi_envinfo env)
{
  ksi_obj	x, args;
  int		len, i;

  args = CDR (form);
  len = LIST_LEN (args);
  KSI_EXN_CHECK (syntax_s, form, len >= 0, "and: invalid syntax");

  if (len == 0)
    return ksi_comp_quote (ksi_true, env);
  if (len == 1)
    return ksi_comp_sexp (CAR (args), env);

  x = ksi_new_code (len, KSI_TAG_AND);
  for (i = 0; i < len; i++, args = CDR (args))
    CODE_VAL (x, i) = ksi_comp_sexp (CAR (args), env);

  return x;
}

static ksi_obj
ksi_comp_or (ksi_obj form, ksi_envinfo env)
{
  ksi_obj	x, args;
  int		len, i;

  args = CDR (form);
  len = LIST_LEN (args);
  KSI_EXN_CHECK (syntax_s, form, len >= 0, "or: invalid syntax");

  if (len == 0)
    return ksi_comp_quote (ksi_false, env);
  if (len == 1)
    return ksi_comp_sexp (CAR (args), env);

  x = ksi_new_code (len, KSI_TAG_OR);
  for (i = 0; i < len; i++, args = CDR (args))
    CODE_VAL (x, i) = ksi_comp_sexp (CAR (args), env);

  return x;
}


/* ksi_comp_cond -- compile cond expresion
 *
 * <form>	::= (cond <clause> <clause> ...)
 * <clause>	::= (<test> <expr> ...)
 *		  | (<test> => <recipient>)
 *		  | (else <expr> <expr> ...).
 */

static ksi_obj
ksi_comp_cond (ksi_obj form, ksi_envinfo env)
{
  ksi_obj x, cl, cls, sym;
  int len;

  cls = CDR (form);
  if (cls == ksi_nil)
    return ksi_comp_quote (ksi_void, env);

  KSI_EXN_CHECK (syntax_s, form, PAIR_P (cls), "cond: invalid syntax");
  cl = CAR (cls);
  cls = CDR (cls);

  len = LIST_LEN (cl);
  KSI_EXN_CHECK (syntax_s, form, len > 0, "cond: invalid syntax");

  if (CAR (cl) == KSI->sym_else)
    {
      KSI_EXN_CHECK (syntax_s, form,
		     len > 1 && cls == ksi_nil,
		     "cond: invalid syntax");

      return ksi_comp_seq (CDR (cl), env);
    }

  if (len == 1)
    {
      sym = ksi_gensym (0, 0);

#ifdef __WATCOMC__
      x = LIST2 (sym, CAR (cl));
      cls = CONS (KSI->sym_cond, cls);
      cls = LIST4 (KSI->sym_if, sym, sym, cls);
      x = LIST3 (KSI->sym_let, LIST1 (x), cls);
#else
      x = LIST3 (KSI->sym_let,
		 LIST1 (LIST2 (sym, CAR (cl))),
		 LIST4 (KSI->sym_if, sym, sym, CONS (KSI->sym_cond, cls)));
#endif

      return ksi_comp_let (x, env);
    }

  if (len == 3 && CAR (CDR (cl)) == KSI->sym_arrow)
    {
      sym = ksi_gensym (0, 0);

#ifdef __WATCOMC__
      x = LIST2 (sym, CAR (cl));
      cl = CAR (CDR (CDR (cl)));
      cls = CONS (KSI->sym_cond, cls);
      cls = LIST4 (KSI->sym_if, sym, LIST2 (cl, sym), cls);
      x = LIST3 (KSI->sym_let, LIST1 (x), cls);
#else
      x = LIST3 (KSI->sym_let,
		 LIST1 (LIST2 (sym, CAR (cl))),
		 LIST4 (KSI->sym_if,
			sym,
			LIST2 (CAR (CDR (CDR (cl))), sym),
			CONS (KSI->sym_cond, cls)));
#endif

      return ksi_comp_let (x, env);
    }

#ifdef __WATCOMC__
  x = CONS (KSI->sym_begin, CDR (cl));
  x = LIST4 (KSI->sym_if,
	     CAR (cl),
	     x,
	     CONS (KSI->sym_cond, cls));
#else
  x = LIST4 (KSI->sym_if,
	     CAR (cl),
	     CONS (KSI->sym_begin, CDR (cl)),
	     CONS (KSI->sym_cond, cls));
#endif

  return ksi_comp_if (x, env);
}


/* ksi_comp_case -- compile case expresion
 *
 * <case-expr>	::= (case <key> <clause> <clause> ...)
 * <clause>	::= ((<datum> <datum> ...) <body>) | (else <body>)
 * <body>	::= <expr> <expr> ...
 *
 */

static ksi_obj
ksi_comp_case (ksi_obj form, ksi_envinfo env)
{
  ksi_obj x, key, cl, cls, sym;

  cls = CDR (form);
  KSI_EXN_CHECK (syntax_s, form, PAIR_P (cls), "case: invalid syntax");

  key = CAR (cls);
  cls = CDR (cls);

  if (PAIR_P (key))
    {
      sym = ksi_gensym (0, 0);

#ifdef __WATCOMC__
      x = LIST2 (sym, key);
      cls = CONS (KSI->sym_case, CONS (sym, cls));
      x = LIST3 (KSI->sym_let, LIST1 (x), cls);
#else
      x = LIST3 (KSI->sym_let,
		 LIST1 (LIST2 (sym, key)),
		 CONS (KSI->sym_case, CONS (sym, cls)));
#endif

      return ksi_comp_let (x, env);
    }

  if (cls == ksi_nil)
    return ksi_comp_quote (ksi_void, env);

  KSI_EXN_CHECK (syntax_s, form, PAIR_P (cls), "case: invalid syntax");
  cl = CAR (cls);
  cls = CDR (cls);

  KSI_EXN_CHECK (syntax_s, form, LIST_LEN (cl) > 1, "case: invalid syntax");

  if (CAR (cl) == KSI->sym_else)
    {
      KSI_EXN_CHECK (syntax_s, form, cls == ksi_nil, "case: invalid syntax");
      return ksi_comp_seq (CDR (cl), env);
    }

  x = LIST2 (KSI->sym_quote, ksi_glob_val (KSI->sym_memv, ksi_false));
  x = LIST4 (KSI->sym_if,
	     LIST3 (x, key, LIST2 (KSI->sym_quote, CAR (cl))),
	     CONS (KSI->sym_begin, CDR (cl)),
	     CONS (KSI->sym_case, CONS (key, cls)));

  return ksi_comp_if (x, env);
}


static ksi_obj
ksi_comp_letstar (ksi_obj form, ksi_envinfo env)
{
  ksi_obj bind, body, bx;
  int len = LIST_LEN (form) - 2;
  KSI_EXN_CHECK (syntax_s, form, len >= 1, "let*: invalid syntax");

  bind = CAR (CDR (form));
  body = CDR (CDR (form));
  if (bind == ksi_nil)
    return ksi_comp_seq (body, env);

  KSI_EXN_CHECK (syntax_s, form, PAIR_P (bind), "let*: invalid syntax");
  if (CDR (bind) == ksi_nil)
    return ksi_comp_let (form, env);

  bx = CAR (bind);
  bind = CDR (bind);

  form = LIST3 (KSI->sym_let, LIST1 (bx),
		CONS (KSI->sym_letstar, CONS (bind, body)));

  return ksi_comp_let (form, env);
}


/* ksi_comp_do
 *
 * (do ((<var> <init> <step>) ...) (<test> <expr> ...) <cmd> ...)
 *   =>
 * (letrec ((loop (lambda (<var> ...)
 *                   (cond (<test> <expr> ...) 
 *                         (else <cmd> ...
 *                               (loop <step> ...))))))
 *    (loop <init> ...))
 */

static ksi_obj
ksi_comp_do (ksi_obj form, ksi_envinfo env)
{
  ksi_obj body, binds, test, cmds, loop;
  ksi_obj vars = ksi_nil,  *locvar = &vars;
  ksi_obj inits = ksi_nil, *locinit = &inits;
  ksi_obj steps = ksi_nil, *locstep = &steps;

  binds = CDR (form);
  KSI_EXN_CHECK (syntax_s, form, PAIR_P (binds), "do: invalid syntax");

  test = CDR (binds);
  KSI_EXN_CHECK (syntax_s, form, PAIR_P (test), "do: invalid syntax");

  cmds = CDR (test);
  KSI_EXN_CHECK (syntax_s, form, LIST_LEN (cmds) >= 0, "do: invalid syntax");

  test = CAR (test);
  KSI_EXN_CHECK (syntax_s, form, PAIR_P (test), "do: invalid syntax");

  body = CDR (test);
  KSI_EXN_CHECK (syntax_s, form, LIST_LEN (body) >= 0, "do: invalid syntax");

  for (binds = CAR (binds); binds != ksi_nil; binds = CDR (binds))
    {
      ksi_obj x, v, i, s;
      KSI_EXN_CHECK (syntax_s, form, PAIR_P (binds), "do: invalid syntax");

      x = CAR (binds);
      KSI_EXN_CHECK (syntax_s, form, PAIR_P (x), "do: invalid syntax");

      v = CAR (x);
      KSI_EXN_CHECK (syntax_s, form, SYM_P (v), "do: invalid syntax");

      x = CDR (x);
      KSI_EXN_CHECK (syntax_s, form, PAIR_P (x), "do: invalid syntax");

      i = CAR (x);
      x = CDR (x);
      if (x == ksi_nil)
	s = v;
      else
	{
	  KSI_EXN_CHECK (syntax_s, form, PAIR_P (x), "do: invalid syntax");
	  KSI_EXN_CHECK (syntax_s, form,
			 CDR (x) == ksi_nil,
			 "do: invalid syntax");
	  s = CAR (x);
	}

      *locvar = CONS (v, ksi_nil);
      locvar = & CDR (*locvar);

      *locinit = CONS (i, ksi_nil);
      locinit = & CDR (*locinit);

      *locstep = CONS (s, ksi_nil);
      locstep = & CDR (*locstep);
    }

  loop = ksi_gensym (0, 0);
  if (cmds == ksi_nil)
    cmds = LIST1 (CONS (loop, steps));
  else
    {
      for (body = cmds; CDR (body) != ksi_nil; body = CDR (body))
	{/* empty */}

      CDR (body) = LIST1 (CONS (loop, steps));
    }

  /*
   * (letrec ((loop (lambda (<var> ...)
   *                   (cond (<test> <expr> ...) 
   *                         (else <cmd> ...
   *                               (loop <step> ...))))))
   *    (loop <init> ...))
   */

#ifdef __WATCOMC__
  body = CONS (KSI->sym_else, cmds);
  body = LIST3 (KSI->sym_cond, test, body);
  body = LIST3 (KSI->sym_lambda, vars, body);
  body = LIST2 (loop, body);
  body = LIST3 (KSI->sym_letrec, LIST1 (body), CONS (loop, inits));
#else
  body = LIST3 (KSI->sym_letrec,
		LIST1 (LIST2 (loop, LIST3 (KSI->sym_lambda,
					   vars,
					   LIST3 (KSI->sym_cond,
						  test,
						  CONS (KSI->sym_else,
							cmds))))),
		CONS (loop, inits));
#endif

  return ksi_comp_letrec (body, env);
}


/* ksi_comp_pair -- compile pair
 *
 */

static ksi_obj
ksi_comp_pair (ksi_obj form, ksi_envinfo env)
{
  ksi_obj x = CAR (form);
  if (x == KSI->sym_quote)
    {
      KSI_EXN_CHECK (syntax_s, form,
		     LIST_LEN (form) == 2,
		     "quote: invalid syntax");
      return ksi_comp_quote (CAR (CDR (form)), env);
    }
  if (x == KSI->sym_define)
    return ksi_comp_define (form, env);
  if (x == KSI->sym_lambda)
    return ksi_comp_lambda (form, env);
  if (x == KSI->sym_if)
    return ksi_comp_if (form, env);
  if (x == KSI->sym_begin)
    {
      KSI_EXN_CHECK (syntax_s, form,
		     LIST_LEN (form) > 1,
		     "begin: invalid syntax");
      return ksi_comp_seq (CDR (form), env);
    }
  if (x == KSI->sym_let)
    return ksi_comp_let (form, env);
  if (x == KSI->sym_letrec)
    return ksi_comp_letrec (form, env);
  if (x == KSI->sym_set)
    return ksi_comp_set (form, env);
  if (x == KSI->sym_and)
    return ksi_comp_and (form, env);
  if (x == KSI->sym_or)
    return ksi_comp_or (form, env);
  if (x == KSI->sym_delay)
    return ksi_comp_delay (form, env);

  /* next syntax'es are expanded by syntax-case macro package
   * and is not appeares in most programs
   */
  if (x == KSI->sym_cond)
    return ksi_comp_cond (form, env);
  if (x == KSI->sym_case)
    return ksi_comp_case (form, env);
  if (x == KSI->sym_letstar)
    return ksi_comp_letstar (form, env);
  if (x == KSI->sym_do)
    return ksi_comp_do (form, env);

  return ksi_comp_apply (form, env);
}

static ksi_obj
ksi_comp_sexp (ksi_obj form, ksi_envinfo env)
{
#if STRICT_R5RS
  if (form == ksi_nil)
    ksi_exn_error ("misc", form, "@compile: no subexpresion in procedure call");
#endif

  if (!KSI_OBJ_P (form))
    return ksi_comp_quote (form, env);

  switch (form -> itag)
    {
    case KSI_TAG_STRING:
    case KSI_TAG_CONST_STRING:
    case KSI_TAG_BIGNUM:
    case KSI_TAG_FLONUM:
    case KSI_TAG_RATIONAL:
    case KSI_TAG_COMPLEX:
    case KSI_TAG_KEYWORD:
      return ksi_comp_quote (form, env);

    case KSI_TAG_SYMBOL:
      return ksi_comp_sym (form, env, 0);

    case KSI_TAG_PAIR:
    case KSI_TAG_CONST_PAIR:
      return ksi_comp_pair (form, env);
    }

  KSI_EXN_ERR (syntax_s, form, "@compile: invalid syntax");
  return 0;
}

ksi_obj
ksi_comp (ksi_obj form, ksi_obj env)
{
  struct Ksi_EnvInfo	envinfo;
  struct Ksi_FrmInfo	frminfo;
  ksi_obj		code, frm;
  int			len;

  frminfo.next = 0;
  frminfo.envs = 0;
  frminfo.used = &frminfo.envs;
  frminfo.free = 0;

  envinfo.next = 0;
  envinfo.down = 0;
  envinfo.link = 0;
  envinfo.frm  = &frminfo;
  envinfo.nums = 0;
  envinfo.vars = 0;

  code = ksi_comp_sexp (form, &envinfo);
  len = realloc_vars (&frminfo, 0);
  if (len == 0)
    return code;

  frm = ksi_new_code (2, KSI_TAG_FRAME);
  CODE_VAL (frm, 0) = KSI_MK_SINT (len);
  CODE_VAL (frm, 1) = code;
  return frm;
}


static ksi_obj
ksi_clos_env (ksi_obj clos)
{
  TYPE_CHECK (clos, KSI_OBJ_IS (clos, KSI_TAG_CLOSURE),
	      "@closure-environment: invalid closure in arg1");
  return CLOS_FRM (clos) ? CLOS_FRM (clos) -> env : ksi_false;
}

static ksi_obj
ksi_clos_doc (ksi_obj clos)
{
  TYPE_CHECK (clos, KSI_OBJ_IS (clos, KSI_TAG_CLOSURE),
	      "@closure-documentation: invalid closure in arg1");
  return CLOS_DOC (clos);
}

static ksi_obj
ksi_clos_body (ksi_obj clos)
{
  TYPE_CHECK (clos, KSI_OBJ_IS (clos, KSI_TAG_CLOSURE),
	      "@closure-body: invalid closure in arg1");
  return CLOS_BODY (clos);
}

static ksi_obj
ksi_clos_p (ksi_obj clos)
{
  return KSI_OBJ_IS (clos, KSI_TAG_CLOSURE) ? ksi_true : ksi_false;
}

static struct Ksi_Prim_Def defs [] =
{
  { "@compile",			ksi_comp,	KSI_CALL_ARG2, 2 },
  { "@closure-environment",	ksi_clos_env,	KSI_CALL_ARG1, 1 },
  { "@closure-documentation",	ksi_clos_doc,	KSI_CALL_ARG1, 1 },
  { "@closure-body",		ksi_clos_body,	KSI_CALL_ARG1, 1 },
  { "@closure?",		ksi_clos_p,	KSI_CALL_ARG1, 1 },

  { 0 }
};

void
ksi_init_comp (void)
{
  ksi_ignore (rcsid);
  ksi_reg_unit (defs);
}

 /* End of code */
