/*
 * ksi_err.c
 * error handling
 *
 * Copyright (C) 1997-2000,2001, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Thu Jul 31 23:10:41 1997
 * Last Update:   Fri Nov  9 23:50:56 2001
 *
 */

/* Comments:
 *
 */

#include "ksi_int.h"

#include <time.h>

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif

#ifdef HAVE_FCNTL_H
#  include <fcntl.h>
#endif

#ifdef _MSC_VER
#  include <io.h>
#endif

static char rcsid[] = "$Id: ksi_err.c,v 1.4 2001/11/25 14:30:01 demakov Exp $";


struct Errlog_Desc
{
  char	*str;
  char	*msg;
};

static struct Errlog_Desc errlog [] =
{
  { "emerg",	"EMERG!   "	},
  { "alert",	"ALERT!   "	},
  { "crit",	"CRIT!    "	},
  { "error",	"ERROR!   "	},
  { "warning",	"WARNING! "	},
  { "notice",	"NOTICE!  "	},
  { "info",	"INFO!    "	},
  { "debug",	"DEBUG!   "	}
};


static char*
ksi_errlog_hdr (int pri)
{
  time_t curtime;
  struct tm *loctime;
  char *tm_str;
  int len;

  curtime = time (0);
  loctime = localtime (&curtime);
  tm_str = asctime (loctime);
  len = strlen (tm_str);
  if (tm_str [len - 1] == '\n')
    tm_str [len - 1] = '\0';

  return ksi_aprintf ("%s (%s): ", tm_str, errlog[pri].str);
}

void
ksi_errlog_msg (int pri, const char *msg)
{
  ksi_obj dest;

  dest = KSI ? KSI->errlog_port : 0;
  if (dest && KSI_OUTPUT_PORT_P (dest))
    {
      char *hdr = ksi_errlog_hdr (pri);
      ksi_port_write ((ksi_port) dest, hdr, strlen (hdr));
      ksi_port_write ((ksi_port) dest, msg, strlen (msg));
      ksi_port_putc  ((ksi_port) dest, '\n');
      ksi_flush_port (dest);
    }

  dest = KSI ? (ksi_obj) KSI->error_port : 0;
  if (KSI && dest && (KSI_INITED || dest != (ksi_obj) KSI->null_port))
    {
      if (KSI->output_port && KSI_OUTPUT_PORT_P (KSI->output_port))
	{
	  if (KSI->output_port -> last_write_char != '\n')
	    ksi_port_putc  (KSI->output_port, '\n');
	  ksi_flush_port ((ksi_obj) KSI->output_port);
	}

      ksi_port_write ((ksi_port) dest, errlog[pri].msg, strlen (errlog[pri].msg));
      if (KSI->interactive && pri <= ERRLOG_WARNING)
	ksi_port_putc ((ksi_port) dest, '\a');
      ksi_port_write ((ksi_port) dest, msg, strlen (msg));
      ksi_port_putc  ((ksi_port) dest, '\n');
      ksi_flush_port (dest);
    }
  else
    {
#ifdef MSWIN32
      MessageBox (NULL, msg, errlog[pri].msg,
		  MB_TASKMODAL | MB_ICONEXCLAMATION | MB_OK);
#else
      write (2, errlog[pri].msg, strlen (errlog[pri].msg));
      write (2, msg, strlen (msg));
      write (2, "\n", 1);
#endif
    }
}


static inline ksi_obj
mk_exn (ksi_obj type, ksi_obj errobj, ksi_obj msg, ksi_obj pri)
{
  ksi_obj exn;
  exn = ksi_alloc_vector (4, KSI_TAG_EXN);
  KSI_EXN_TYPE (exn) = type;
  KSI_EXN_MSG  (exn) = msg;
  KSI_EXN_VAL  (exn) = errobj;
  KSI_EXN_PRI  (exn) = pri;
  return exn;
}

ksi_obj
ksi_make_exn (const char *type, ksi_obj errobj, const char *msg, ksi_obj pri)
{
  int priority;

  if (!pri)
    priority = ERRLOG_ERROR;
  else
    {
      TYPE_CHECK (pri, KSI_SINT_P (pri), "make-exn: invalid priority");
      priority = KSI_SINT_COD (pri);
      RANGE_CHECK (pri, ERRLOG_EMERG <= priority && priority <= ERRLOG_ALL,
		   "make-exn: priority out of range");
    }

  return mk_exn (ksi_str02sym (type ? type : "misc"),
		 errobj ? errobj : ksi_void,
		 ksi_str02string (msg ? msg : "unspecified error"),
		 KSI_MK_SINT (priority));
}

static ksi_obj
ksi_mk_exn (ksi_obj type, ksi_obj errobj, ksi_obj msg, ksi_obj pri)
{
  int priority;

  TYPE_CHECK (type, KSI_SYM_P (type), "make-exn: invalid symbol in arg1");
  TYPE_CHECK (msg,  KSI_STR_P (msg),  "make-exn: invalid string in arg3");
  if (!pri)
    priority = ERRLOG_ERROR;
  else
    {
      TYPE_CHECK (pri, KSI_SINT_P (pri),  "make-exn: invalid priority in arg4");
      priority = KSI_SINT_COD (pri);
      RANGE_CHECK (pri, ERRLOG_EMERG <= priority && priority <= ERRLOG_ALL,
		   "make-exn: priority out of range");
    }

  return mk_exn (type, errobj, msg, KSI_MK_SINT (priority));
}

ksi_obj
ksi_exn_p (ksi_obj x)
{
  return KSI_EXN_P (x) ? ksi_true : ksi_false;
}

ksi_obj
ksi_exn_type (ksi_obj x)
{
  TYPE_CHECK (x, KSI_EXN_P (x), "exn-type: invalid exception in arg1");
  return KSI_EXN_TYPE (x);
}

ksi_obj
ksi_exn_message (ksi_obj x)
{
  TYPE_CHECK (x, KSI_EXN_P (x), "exn-message: invalid exception in arg1");
  return KSI_EXN_MSG (x);
}

ksi_obj
ksi_exn_value (ksi_obj x)
{
  TYPE_CHECK (x, KSI_EXN_P (x), "exn-value: invalid exception in arg1");
  return KSI_EXN_VAL (x);
}

ksi_obj
ksi_exn_priority (ksi_obj x)
{
  TYPE_CHECK (x, KSI_EXN_P (x), "exn-priority: invalid exception in arg1");
  return KSI_EXN_PRI (x);
}

int
ksi_exn_error (const char *type, ksi_obj errobj, const char *fmt, ...)
{
  va_list args;
  ksi_obj msg;
  int pri;

  va_start (args, fmt);
  msg = ksi_vprintf (ksi_false, fmt, args);
  va_end (args);

  pri = (KSI_INITED ? ERRLOG_ERROR : ERRLOG_EMERG);
  ksi_throw_error (mk_exn (ksi_str02sym (type ? type : "misc"),
			   errobj ? errobj : ksi_void,
			   msg,
			   KSI_MK_SINT (pri)));
  return 0;
}

int
ksi_warn (const char* msg, ...)
{
  ksi_obj str;
  va_list args;

  if (!KSI || ERRLOG_WARNING > KSI->errlog_priority)
    return 0;

  va_start (args, msg);
  str = ksi_vprintf (ksi_false, msg, args);
  va_end (args);

  ksi_errlog_msg (ERRLOG_WARNING, KSI_STR_PTR (str));
  return 0;
}

int
ksi_error (const char* fmt, ...)
{
  va_list args;
  ksi_obj msg;
  int pri;

  va_start (args, fmt);
  msg = ksi_vprintf (ksi_false, fmt, args);
  va_end (args);

  pri = (KSI_INITED ? ERRLOG_ERROR : ERRLOG_EMERG);
  ksi_throw_error (mk_exn (ksi_str02sym ("misc"),
			   ksi_void,
			   msg,
			   KSI_MK_SINT (pri)));
  return 0;
}

int
ksi_crit_error (ksi_obj priority, const char* fmt, ...)
{
  va_list args;
  ksi_obj msg;
  int pri;

  TYPE_CHECK (priority, KSI_SINT_P (priority), "crit-error: invalid priority");
  pri = KSI_SINT_COD (priority);
  RANGE_CHECK (priority, ERRLOG_EMERG <= pri && pri <= ERRLOG_ALL,
	       "crit-error: priority out of range");

  va_start (args, fmt);
  msg = ksi_vprintf (ksi_false, fmt, args);
  va_end (args);

  ksi_throw_error (mk_exn (ksi_str02sym ("misc"),
			   ksi_void,
			   msg,
			   priority));
  return 0;
}

static ksi_obj
error_handler (ksi_obj exn)
{
  int         pri = KSI_SINT_COD (KSI_EXN_PRI (exn));
  const char *msg = KSI_STR_PTR (KSI_EXN_MSG (exn));
  ksi_obj     val = KSI_EXN_VAL (exn);
 
  if (val != ksi_void)
    msg = ksi_aprintf ("%s: %s", msg, ksi_obj2str (val));

  ksi_errlog_msg (pri, msg);
  return ksi_void;
}


ksi_obj
ksi_open_errlog (ksi_obj priority, ksi_obj dest)
{
  int     new_port = 0, pri;
  ksi_obj port = 0;

  TYPE_CHECK (priority, KSI_SINT_P (priority),
	      "open-errlog: invalid priority in arg1");
  pri = KSI_SINT_COD (priority);
  RANGE_CHECK (priority, ERRLOG_EMERG <= pri && pri <= ERRLOG_ALL,
	       "open-errlog: priority out of range");

  if (!dest)
    {
      KSI->errlog_priority = pri;
      return ksi_void;
    }

  if (KSI_OUTPUT_PORT_P (dest))
    port = dest;
  else if (KSI_STR_P (dest))
    {
      port = ksi_open_fd_port_int (KSI_STR_PTR (dest), "a", "open-errlog");
      new_port = 1;
    }
  else
    ksi_exn_error ("type", dest, "open-errlog: invalid destination in arg2");


  if (KSI->errlog_port && KSI_OUTPUT_PORT_P (KSI->errlog_port))
    {
      if (KSI->close_errlog_port)
	ksi_close_port (KSI->errlog_port);
    }

  KSI->errlog_port = port;
  KSI->close_errlog_port = new_port;
  KSI->errlog_priority = pri;

  return ksi_void;
}

ksi_obj
ksi_errlog_priority (ksi_obj module, ksi_obj priority)
{
  int pri;

  TYPE_CHECK (priority, KSI_SINT_P (priority),
	      "errlog-priority: invalid integer in arg2");
  pri = KSI_SINT_COD (priority);
  RANGE_CHECK (priority, ERRLOG_EMERG <= pri && pri <= ERRLOG_ALL,
	       "errlog-priority: priority out of range");

  KSI->module_priority = ksi_assq_set_x (KSI->module_priority, module, priority);
  return ksi_void;
}

ksi_obj
ksi_errlog (ksi_obj module, ksi_obj priority, const char *fmt, ...)
{
  va_list args;
  int pri, log_pri;
  ksi_obj str, x;

  TYPE_CHECK (priority, KSI_SINT_P (priority), "errlog: invalid priority");
  pri = KSI_SINT_COD (priority);
  RANGE_CHECK (priority, ERRLOG_EMERG <= pri && pri <= ERRLOG_ALL,
	       "errlog: priority out of range");

  if (!KSI)
    return ksi_void;

  x = (module ? ksi_assq_ref (KSI->module_priority, module) : ksi_false);
  if (x == ksi_false)
    x = ksi_assq_ref (KSI->module_priority, KSI->ksi_app_sym);

  if (x == ksi_false)
    log_pri = KSI->errlog_priority;
  else
    log_pri = KSI_SINT_COD (x);

  if (pri > log_pri)
    return ksi_void;

  va_start (args, fmt);
  str = ksi_vprintf (ksi_false, fmt, args);
  va_end (args);

  ksi_errlog_msg (pri, KSI_STR_PTR (str));
  return ksi_void;
}


static ksi_obj
ksi_errlog_format (ksi_obj mod, ksi_obj priority, ksi_obj fmt,
		   int argc, ksi_obj* argv)
{
  int pri, log_pri;
  ksi_obj str, x;

  TYPE_CHECK (priority, KSI_SINT_P (priority),
	      "errlog: invalid priority in arg2");
  TYPE_CHECK (fmt, KSI_STR_P (fmt),
	      "errlog: invalid string in arg3");

  pri = KSI_SINT_COD (priority);
  RANGE_CHECK (priority, ERRLOG_EMERG <= pri && pri <= ERRLOG_ALL,
	       "errlog: priority out of range");

  x = ksi_assq_ref (KSI->module_priority, mod);
  if (x == ksi_false && mod != KSI->ksi_app_sym)
    x = ksi_assq_ref (KSI->module_priority, KSI->ksi_app_sym);

  if (x == ksi_false)
    log_pri = KSI->errlog_priority;
  else
    log_pri = KSI_SINT_COD (x);

  if (pri > log_pri)
    return ksi_void;

  str = ksi_format (ksi_false, KSI_STR_PTR (fmt), argc, argv);
  ksi_errlog_msg (pri, KSI_STR_PTR (str));
  return ksi_void;
}


static struct Ksi_Prim_Def defs[] =
{
  { "make-exn",		ksi_mk_exn,		KSI_CALL_ARG4,  3 },
  { "exn?",		ksi_exn_p,		KSI_CALL_ARG1,  1 },
  { "exn-type",		ksi_exn_type,		KSI_CALL_ARG1,  1 },
  { "exn-message",	ksi_exn_message,	KSI_CALL_ARG1,  1 },
  { "exn-value",	ksi_exn_value,		KSI_CALL_ARG1,  1 },
  { "exn-priority",	ksi_exn_priority,	KSI_CALL_ARG1,  1 },

  { "open-errlog",	ksi_open_errlog,	KSI_CALL_ARG2,  1 },
  { "errlog-priority",	ksi_errlog_priority,	KSI_CALL_ARG2,  2 },
  { "errlog",		ksi_errlog_format,	KSI_CALL_REST3, 3 },

  { 0 }
};

void
ksi_init_error (void)
{
  ksi_ignore (rcsid);
  ksi_reg_unit (defs);

  ksi_defsym ("errlog/emerg",   errlog_emerg,   0);
  ksi_defsym ("errlog/alert",   errlog_alert,   0);
  ksi_defsym ("errlog/crit",    errlog_crit,    0);
  ksi_defsym ("errlog/error",   errlog_error,   0);
  ksi_defsym ("errlog/warning", errlog_warning, 0);
  ksi_defsym ("errlog/notice",  errlog_notice,  0);
  ksi_defsym ("errlog/info",    errlog_info,    0);
  ksi_defsym ("errlog/debug",   errlog_debug,   0);
  ksi_defsym ("errlog/all",     errlog_all,     0);

  KSI->no_mem_exn = ksi_make_exn ("fatal",
				  ksi_void,
				  "cannot allocate memory",
				  errlog_alert);

  KSI->user_quit_exn = ksi_make_exn ("quit",
				     ksi_void,
				     "interrupted by user",
				     errlog_error);

  ksi_add_error_handler (ksi_defun (0, error_handler, KSI_CALL_ARG1, 1));
}

 /* End of code */
