/*
 * ksi_evt.c
 * async events
 *
 * Copyright (C) 2000-2001, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written
 * agreement is hereby granted, provided that the above copyright notice
 * and this paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Tue Mar 28 02:51:26 2000
 * Last Update:   Sun Apr 29 18:03:17 2001
 *
 */

#include "ksi_int.h"

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif

#ifdef HAVE_FCNTL_H
#  include <fcntl.h>
#endif

#ifdef TIME_WITH_SYS_TIME
#  include <sys/time.h>
#  include <time.h>
#else
#  ifdef HAVE_SYS_TIME_H
#    include <sys/time.h>
#  else
#    include <time.h>
#  endif
#endif

static char rcsid[] = "$Id: ksi_evt.c,v 1.17 2001/04/29 11:22:06 demakov Exp $";


static int	events_in_progress = 0;
static int	events_disabled = 1;
static int	events_blocked = 0;

#define BLOCK_WAIT					\
  do {							\
    if (events_blocked == 0)				\
      if (KSI->event_mgr->block_wait)			\
	KSI->event_mgr->block_wait (KSI->event_mgr);	\
    events_blocked++;					\
  } while (0)

#define UNBLOCK_WAIT					\
  do {							\
    if (--events_blocked == 0)				\
      if (KSI->event_mgr->unblock_wait)			\
	KSI->event_mgr->unblock_wait (KSI->event_mgr);	\
  } while (0)


/* append_event, remove_event -- /   
 */

static void
append_event (ksi_event x, ksi_event *list)
{
  if (*list)
    (*list)->prev = x;

  x->prev = 0;
  x->next = *list;
  *list = x;
}

static void
remove_event (ksi_event x, ksi_event *list)
{
  if (x->prev)
    x->prev->next = x->next;
  else
    *list = x->next;
  if (x->next)
    x->next->prev = x->prev;

  x->prev = x->next = 0;
}


ksi_obj
ksi_event_p (ksi_obj x)
{
  return (KSI_EVT_P (x) ? ksi_true : ksi_false);
}

ksi_obj
ksi_event_state (ksi_obj x)
{
  TYPE_CHECK (x, KSI_EVT_P (x), "event-state: invalid event in arg1");
  return KSI_EVT_STATE (x);
}

ksi_obj
ksi_event_result (ksi_obj x)
{
  TYPE_CHECK (x, KSI_EVT_P (x), "event-result: invalid event in arg1");
  return KSI_EVT_RESULT (x);
}

ksi_obj
ksi_event_procedure (ksi_obj x)
{
  TYPE_CHECK (x, KSI_EVT_P (x), "event-procedure: invalid event in arg1");
  return KSI_EVT_ACTION (x);
}


ksi_obj
ksi_start_event (ksi_obj x)
{
  TYPE_CHECK (x, KSI_EVT_P (x), "start-event: invalid event in arg1");
  if (KSI->event_mgr == 0)
    ksi_exn_error ("system", 0, "start-event: not supported");

  BLOCK_WAIT;
  if (((ksi_event) x) -> waiting || ((ksi_event) x) -> pending)
    {
      ((ksi_event) x) -> start = 1;
      ((ksi_event) x) -> stop  = 0;
      ((ksi_event) x) -> ops -> init ((ksi_event) x);
    }
  else
    {
      if (((ksi_event) x) -> active)
	{
	  ((ksi_event) x) -> active = 0;
	  remove_event ((ksi_event) x, &KSI->active_events);
	}

      ((ksi_event) x) -> start = 1;
      ((ksi_event) x) -> stop  = 0;
      ((ksi_event) x) -> ops -> init ((ksi_event) x);

      if (!((ksi_event) x) -> waiting && !((ksi_event) x) -> pending)
	{
	  ((ksi_event) x) -> waiting = 1;
	  append_event ((ksi_event) x, &KSI->waiting_events);
	}
    }
  UNBLOCK_WAIT;

  if (!events_in_progress && !events_disabled)
    ksi_run_pending_events ();
  else
    KSI->have_event = 1;

  return x;
}

ksi_obj
ksi_stop_event (ksi_obj x)
{
  TYPE_CHECK (x, KSI_EVT_P (x), "stop-event: invalid event in arg1");
  if (KSI->event_mgr == 0)
    ksi_exn_error ("system", 0, "stop-event: not supported");

  BLOCK_WAIT;
  if (((ksi_event) x) -> waiting || ((ksi_event) x) -> pending)
    {
      ((ksi_event) x) -> start = 0;
      ((ksi_event) x) -> stop  = 1;
    }
  else
    {
      if (((ksi_event) x) -> active)
	{
	  ((ksi_event) x) -> active  = 0;
	  ((ksi_event) x) -> start   = 0;
	  ((ksi_event) x) -> stop    = 1;
	  ((ksi_event) x) -> waiting = 1;
	  remove_event ((ksi_event) x, &KSI->active_events);
	  append_event ((ksi_event) x, &KSI->waiting_events);
	}
    }
  UNBLOCK_WAIT;

  if (!events_in_progress && !events_disabled)
    ksi_run_pending_events ();
  else
    KSI->have_event = 1;

  return x;
}

void
ksi_run_event (ksi_event evt, void *data, int invoke)
{
  if (evt)
    {
      BLOCK_WAIT;
      evt->data = data;

      if (evt->active)
	{
	  evt->active = 0;
	  remove_event (evt, &KSI->active_events);
	}

      evt->ready = 1;
      if (!evt->waiting && !evt->pending)
	{
	  evt->waiting = 1;
	  append_event (evt, &KSI->waiting_events);
	}
      UNBLOCK_WAIT;
    }

  if (invoke)
    ksi_run_pending_events ();
  else
    KSI->have_event = 1;
}

int
ksi_run_pending_events (void)
{
  volatile ksi_wind old_exit, new_exit;

  if (events_in_progress)
    return 0;

  events_in_progress++;
  ksi_disable_evt ();

  old_exit = KSI->exit_catch;
  new_exit = ksi_add_catch (ksi_true, 0, 0);

  KSI->exit_catch = new_exit;
  FLUSH_REGISTER_WINDOWS;
  if (setjmp (new_exit->the_catch->jmp.j_buf) == 0)
    {
      do
	{
	  while (KSI->pending_events)
	    {
	      ksi_event curr = KSI->pending_events;

	      BLOCK_WAIT;
	      curr->pending = 0;
	      curr->waiting = 1;
	      curr->ready   = 0;
	      curr->state   = KSI->sym_ready;
	      remove_event (curr, &KSI->pending_events);
	      append_event (curr, &KSI->waiting_events);
	      UNBLOCK_WAIT;

	      if (curr->ops->invoke (curr, curr->data))
		{
		  if (!curr->start)
		    curr->stop = 1;
		}
	    }

	  while (KSI->waiting_events)
	    {
	      ksi_event curr = KSI->waiting_events;

	      BLOCK_WAIT;
	      curr->waiting = 0;
	      remove_event (curr, &KSI->waiting_events);
	      UNBLOCK_WAIT;

	      if (curr->start)
		{
		  if (curr->inited)
		    {
		      curr->inited = 0;
		      curr->ops->cancel (curr);
		    }

		  curr->inited = 1;
		  curr->start  = 0;
		  curr->stop   = 0;
		  curr->ready  = 0;
		  curr->state  = KSI->sym_wait;
		  curr->ops->setup (curr);
		}

	      if (curr->stop)
		{
		  if (curr->inited)
		    {
		      curr->inited = 0;
		      curr->ops->cancel (curr);
		    }

		  curr->start  = 0;
		  curr->stop   = 0;
		  curr->ready  = 0;
		  curr->state  = KSI->sym_inactive;
		}

	      if (curr->ready)
		{
		  BLOCK_WAIT;
		  if (!curr->waiting && !curr->pending)
		    {
		      curr->pending = 1;
		      append_event (curr, &KSI->pending_events);
		    }
		  UNBLOCK_WAIT;
		}

	      if (curr->inited)
		{
		  BLOCK_WAIT;
		  if (!curr->waiting && !curr->pending)
		    {
		      curr->active = 1;
		      append_event (curr, &KSI->active_events);
		    }
		  UNBLOCK_WAIT;
		}
	    }
	} while (KSI->pending_events);

      ksi_del_catch (new_exit);
      new_exit = 0;
    }

  KSI->exit_catch = old_exit;
  KSI->have_event = 0;

  ksi_enable_evt ();
  events_in_progress--;

  if (new_exit)
    ksi_rethrow (new_exit);
  return 0;
}

int
ksi_has_pending_events (void)
{
  if (KSI->pending_events || KSI->waiting_events)
    return 1;
  return KSI->have_event;
}

int
ksi_do_events (void)
{
  if (!events_in_progress && !events_disabled && (KSI->pending_events || KSI->waiting_events))
    ksi_run_pending_events ();

  KSI->have_event = 0;
  return 0;
}

ksi_obj
ksi_wait_event (ksi_obj tm)
{
  volatile ksi_wind old_exit, new_exit;
  volatile ksi_obj retval;
  double timeout, wait_time;

  if (KSI->event_mgr == 0 || KSI->event_mgr->wait_event == 0)
    ksi_exn_error ("system", 0, "wait-event: not supported");

  if (tm == ksi_true)
    tm = 0;

  if (tm == 0 || tm == ksi_false || tm == ksi_zero)
    timeout = 0.0;
  else
    {
      TYPE_CHECK (tm, KSI_REAL_P (tm), "wait-event: invalid real in arg1");
      timeout = ksi_num2double (tm);
      RANGE_CHECK (tm, timeout >= 0.0, "wait-event: negative real in arg1");
    }

  wait_time = ksi_real_time () + timeout;
  retval    = ksi_false;

  ksi_disable_evt ();

  old_exit = KSI->exit_catch;
  new_exit = ksi_add_catch (ksi_true, 0, 0);

  KSI->exit_catch = new_exit;
  FLUSH_REGISTER_WINDOWS;
  if (setjmp (new_exit->the_catch->jmp.j_buf) == 0)
    {
      do
	{
	  KSI->event_mgr->wait_event (KSI->event_mgr, tm ? timeout : -1.0);
	  if (KSI->pending_events || KSI->waiting_events)
	    {
	      retval = ksi_true;
	      break;
	    }
	} while (tm == 0 || (timeout = wait_time - ksi_real_time ()) > 0.0);

      ksi_del_catch (new_exit);
      new_exit = 0;
    }
  KSI->exit_catch = old_exit;

  if (ksi_enable_evt () == ksi_false)
    ksi_run_pending_events ();

  if (new_exit)
    ksi_rethrow (new_exit);
  return retval;
}

ksi_obj
ksi_sleep (ksi_obj tm)
{
  double timeout;

  TYPE_CHECK (tm, KSI_REAL_P (tm), "sleep: invalid real in arg1");
  timeout = ksi_num2double (tm);
  RANGE_CHECK (tm, timeout >= 0.0, "sleep: negative real in arg1");

  if (KSI->event_mgr && KSI->event_mgr->wait_event)
    {
      double wait_time = ksi_real_time () + timeout;
      while ((timeout = wait_time - ksi_real_time ()) > 0.0)
	ksi_wait_event (ksi_double2num (timeout));
    }
  else
    {
      double wait_time = ksi_real_time () + timeout;
      while ((timeout = wait_time - ksi_real_time ()) > 0.0)
	{
#if defined(HAVE_NANOSLEEP)
	  struct timespec tv;
	  tv.tv_sec  = (time_t) timeout;
	  tv.tv_nsec = (long) ((timeout - tv.tv_sec) * 1000000000);
	  nanosleep (&tv, 0);
#elif defined(HAVE_SLEEP) && defined(HAVE_USLEEP)
	  unsigned long sec  = (unsigned long) timeout;
	  unsigned long usec = (unsigned long) ((timeout - sec) * 1000000);
	  if (sec)  sleep (sec);
	  if (usec) usleep (usec);
#elif defined(HAVE_SLEEP)
	  unsigned long sec  = (unsigned long) (timeout + 0.5);
	  if (sec) sleep (sec);
#endif
	}
    }

  return ksi_void;
}


ksi_obj
ksi_disable_evt (void)
{
  if (KSI->event_mgr == 0)
    ksi_exn_error ("system", 0, "disable-async-event: not supported");

  if (events_disabled++ == 0)
    {
      if (KSI->event_mgr->disable_async_wait)
	KSI->event_mgr->disable_async_wait (KSI->event_mgr);
    }

  return ksi_false;
}

ksi_obj
ksi_enable_evt (void)
{
  if (KSI->event_mgr == 0)
    ksi_exn_error ("system", 0, "enable-async-event: not supported");

  if (events_disabled == 0)
    ksi_exn_error ("misc", 0, "enable-async-event: async events already enabled");

  if (--events_disabled == 0)
    {
      if (KSI->event_mgr->enable_async_wait)
	KSI->event_mgr->enable_async_wait (KSI->event_mgr);

      if (!events_in_progress && (KSI->pending_events || KSI->waiting_events))
	ksi_run_pending_events ();
      return ksi_true;
    }

  return ksi_false;
}


ksi_event_mgr
ksi_current_event_mgr (void)
{
  return KSI->event_mgr;
}

ksi_event_mgr
ksi_register_event_mgr (ksi_event_mgr mgr)
{
  ksi_event_mgr old_mgr = KSI->event_mgr;

  if (KSI->event_mgr)
    {
      BLOCK_WAIT;
      ksi_disable_evt ();
      ksi_run_pending_events ();

      while (KSI->active_events)
	{
	  ksi_event curr = KSI->active_events;

	  curr->ops->cancel (curr);
	  
	  curr->active  = 0;
	  curr->inited  = 0;
	  curr->waiting = 1;
	  curr->start   = 1;
	  curr->ready   = 0;
	  remove_event (curr, &KSI->active_events);
	  append_event (curr, &KSI->waiting_events);
	}

      UNBLOCK_WAIT;

      if (KSI->event_mgr->term)
	KSI->event_mgr->term (mgr);
    }

  KSI->event_mgr = mgr;

  if (KSI->event_mgr)
    {
      if (KSI->event_mgr->init)
	KSI->event_mgr->init (mgr);

      if (ksi_enable_evt () == ksi_false)
	ksi_run_pending_events ();
    }

  return old_mgr;
}

void*
ksi_wait_timer (ksi_event evt, double tm, int restart)
{
  if (KSI->event_mgr == 0 || KSI->event_mgr->wait_timer == 0)
    ksi_exn_error ("system", 0, "ksi_wait_timer: not supported");
  return KSI->event_mgr->wait_timer (KSI->event_mgr, evt, tm, restart);
}

void
ksi_cancel_timer (ksi_event evt, void *data)
{
  if (KSI->event_mgr == 0 || KSI->event_mgr->cancel_timer == 0)
    ksi_exn_error ("system", 0, "ksi_cancel_timer: not supported");
  KSI->event_mgr->cancel_timer (KSI->event_mgr, evt, data);
}

void*
ksi_wait_input (ksi_event evt, int fd, int restart)
{
  if (KSI->event_mgr == 0 || KSI->event_mgr->wait_input == 0)
    ksi_exn_error ("system", 0, "ksi_wait_input: not supported");
  return KSI->event_mgr->wait_input (KSI->event_mgr, evt, fd, restart);
}

void
ksi_cancel_input (ksi_event evt, void *data)
{
  if (KSI->event_mgr == 0 || KSI->event_mgr->cancel_input == 0)
    ksi_exn_error ("system", 0, "ksi_cancel_input: not supported");
  KSI->event_mgr->cancel_input (KSI->event_mgr, evt, data);
}

void*
ksi_wait_output (ksi_event evt, int fd, int restart)
{
  if (KSI->event_mgr == 0 || KSI->event_mgr->wait_output == 0)
    ksi_exn_error ("system", 0, "ksi_wait_output: not supported");
  return KSI->event_mgr->wait_output (KSI->event_mgr, evt, fd, restart);
}

void
ksi_cancel_output (ksi_event evt, void *data)
{
  if (KSI->event_mgr == 0 || KSI->event_mgr->cancel_output == 0)
    ksi_exn_error ("system", 0, "ksi_cancel_output: not supported");
  KSI->event_mgr->cancel_output (KSI->event_mgr, evt, data);
}

void*
ksi_wait_signal (ksi_event evt, int sig, int restart)
{
  if (KSI->event_mgr == 0 || KSI->event_mgr->wait_signal == 0)
    ksi_exn_error ("system", 0, "ksi_wait_signal: not supported");
  return KSI->event_mgr->wait_signal (KSI->event_mgr, evt, sig, restart);
}

void
ksi_cancel_signal (ksi_event evt, void *data)
{
  if (KSI->event_mgr == 0 || KSI->event_mgr->cancel_signal == 0)
    ksi_exn_error ("system", 0, "ksi_cancel_signal: not supported");
  KSI->event_mgr->cancel_signal (KSI->event_mgr, evt, data);
}

void*
ksi_wait_idle (ksi_event evt, int restart)
{
  if (KSI->event_mgr == 0 || KSI->event_mgr->wait_idle == 0)
    ksi_exn_error ("system", 0, "ksi_wait_idle: not supported");
  return KSI->event_mgr->wait_idle (KSI->event_mgr, evt, restart);
}

void
ksi_cancel_idle (ksi_event evt, void *data)
{
  if (KSI->event_mgr == 0 || KSI->event_mgr->cancel_idle == 0)
    ksi_exn_error ("system", 0, "ksi_cancel_idle: not supported");
  KSI->event_mgr->cancel_idle (KSI->event_mgr, evt, data);
}


/*------------------------------------------------------------------
 * ready
 */

static const char*
ready_name (ksi_event x)
{
  return "ready";
}

static void
ready_init (ksi_event x)
{
}

static void
ready_setup (ksi_event x)
{
  ksi_run_event (x, 0, 0);
}

static void
ready_cancel (ksi_event x)
{
}

static int
ready_invoke (ksi_event x, void *data)
{
  x->result = ksi_apply_1_with_catch (x->action, (ksi_obj) x);
  return 1;
}

static struct Ksi_Event_Tag ready_ops =
{
  ready_name,
  ready_init,
  ready_setup,
  ready_cancel,
  ready_invoke
};

ksi_obj
ksi_ready_event (ksi_obj proc)
{
  ksi_event x;

  TYPE_CHECK (proc, PROC_P (proc), "make-ready-event: invalid procedure in arg1");

  x = (ksi_event) ksi_malloc (sizeof *x);
  x->itag   = KSI_TAG_EVENT;
  x->ops    = &ready_ops;
  x->state  = KSI->sym_inactive;
  x->action = proc;
  x->result = ksi_void;

  return (ksi_obj) x;
}


/* -----------------------------------------------------------------
 * timer
 */

typedef struct Ksi_Timer
{
  struct Ksi_Event	evt;
  void			*timer_data;
  double		time;               /*    */
  double		timeout;            /*   */
} *ksi_timer;


static const char*
timer_name (ksi_event x)
{
  double tm;

  if (x->state != KSI->sym_wait)
    return ksi_aprintf ("timer %y", x->state);

  tm = ((ksi_timer) x) -> time - ksi_real_time ();
  return ksi_aprintf ("timer %f", tm);
}

static void
timer_init (ksi_event x)
{
  ((ksi_timer) x) -> time = ksi_real_time () + ((ksi_timer) x) -> timeout;
}

static void
timer_setup (ksi_event x)
{
  double tm = ((ksi_timer) x) -> time - ksi_real_time ();
  if (tm <= 0.0)
    ksi_run_event (x, 0, 0);
  else
    ((ksi_timer) x) -> timer_data = ksi_wait_timer (x, tm, 0);
}

static void
timer_cancel (ksi_event x)
{
  if (((ksi_timer) x) -> timer_data)
    {
      ksi_cancel_timer (x, ((ksi_timer) x) -> timer_data);
      ((ksi_timer) x) -> timer_data = 0;
    }
}

static int
timer_invoke (ksi_event x, void *data)
{
  if (((ksi_timer) x) -> timer_data == data)
    ((ksi_timer) x) -> timer_data = 0;

  x->state  = KSI->sym_timeout;
  x->result = ksi_apply_1_with_catch (x->action, (ksi_obj) x);
  return 1;
}

static struct Ksi_Event_Tag timer_ops =
{
  timer_name,
  timer_init,
  timer_setup,
  timer_cancel,
  timer_invoke
};

ksi_obj
ksi_timer_event (ksi_obj tm, ksi_obj proc)
{
  ksi_timer x;
  double timeout;

  TYPE_CHECK (tm, KSI_REAL_P (tm), "make-timer-event: invalid real in arg1");
  timeout = ksi_num2double (tm);
  RANGE_CHECK (tm, timeout >= 0.0, "make-timer-event: negative in arg1");
  TYPE_CHECK (proc, PROC_P (proc), "make-timer-event: invalid procedure in arg2");

  x = (ksi_timer) ksi_malloc (sizeof *x);
  x->evt.itag   = KSI_TAG_EVENT;
  x->evt.ops    = &timer_ops;
  x->evt.state  = KSI->sym_inactive;
  x->evt.action = proc;
  x->evt.result = ksi_void;
  x->timeout    = timeout;

  return (ksi_obj) x;
}


/* -----------------------------------------------------------------
 * port events
 */

typedef struct Ksi_PortReady
{
  struct Ksi_Event	evt;
  ksi_port		port;
  void			*timer_data;
  void			*r_data;
  void			*w_data;
  double		tick;               /*    */
  double		timeout;            /*   */
  double		time;               /*    */
  unsigned		r_wait : 1;         /* wait for read */
  unsigned		w_wait : 1;         /* wait for write */
  unsigned		r_ready : 1;        /* read ready */
  unsigned		w_ready : 1;        /* write ready */
  unsigned		no_timeout : 1;
} *ksi_port_ready;


static const char*
port_name (ksi_event x)
{
  ksi_port_ready e = ((ksi_port_ready) x);

  if (e->r_wait)
    return ksi_aprintf ("input-ready %w", e->port);
  if (e->w_wait)
    return ksi_aprintf ("output-ready %w", e->port);

  return "port-ready";
}

static void
port_init (ksi_event e)
{
  ((ksi_port_ready) e) -> time = ksi_real_time () + ((ksi_port_ready) e) -> timeout;
}

static void
port_setup (ksi_event x)
{
  ksi_port_ready e = ((ksi_port_ready) x);
  double tm;

  if (e->port->closed)
    {
      ksi_run_event ((ksi_event) e, 0, 0);
      return;
    }

  if (e->r_wait)
    {
      int fd = e->port->ops->input_fd (e->port);
      e->r_data = ksi_wait_input ((ksi_event) e, fd, 0);
    }
  if (e->w_wait)
    {
      int fd = e->port->ops->output_fd (e->port);
      e->w_data = ksi_wait_output ((ksi_event) e, fd, 0);
    }

  if ((e->r_data || !e->r_wait) && (e->w_data || !e->w_wait))
    {
      if (!e->no_timeout)
	{
	  tm = e->time - ksi_real_time ();
	  if (tm <= 0.0)
	    ksi_run_event ((ksi_event) e, 0, 0);
	  else
	    e->timer_data = ksi_wait_timer ((ksi_event) e, tm, 0);
	}
    }
  else
    {
      if (e->r_wait && (e->r_ready || e->port->ops->input_ready (e->port)))
	{
	  e->r_ready = 1;
	  ksi_run_event ((ksi_event) e, 0, 0);
	  return;
	}

      if (e->w_wait && (e->w_ready || e->port->ops->output_ready (e->port)))
	{
	  e->w_ready = 1;
	  ksi_run_event ((ksi_event) e, 0, 0);
	  return;
	}

      tm = ksi_real_time () + e->tick;
      if (!e->no_timeout && e->time < tm)
	tm = e->time;

      tm = tm - ksi_real_time ();
      if (tm <= 0.0)
	ksi_run_event ((ksi_event) e, 0, 0);
      else
	e->timer_data = ksi_wait_timer ((ksi_event) e, tm, 0);
    }
}

static void
port_cancel (ksi_event x)
{
  ksi_port_ready e = ((ksi_port_ready) x);

  if (e->timer_data)
    {
      ksi_cancel_timer ((ksi_event) e, e->timer_data);
      e->timer_data = 0;
    }
  if (e->r_data)
    {
      ksi_cancel_input ((ksi_event) e, e->r_data);
      e->r_data = 0;
    }
  if (e->w_data)
    {
      ksi_cancel_output ((ksi_event) e, e->w_data);
      e->w_data = 0;
    }
}

static int
port_invoke (ksi_event x, void *data)
{
  ksi_port_ready e = ((ksi_port_ready) x);

  if (e->port->closed)
    return 1;

  if (data)
    {
      if (data == e->r_data)
	{
	  e->r_ready = 1;
	  e->r_data  = 0;
	}
      if (data == e->w_data)
	{
	  e->w_ready = 1;
	  e->w_data  = 0;
	}
      if (data == e->timer_data)
	{
	  e->timer_data = 0;
	}
    }

  if (e->r_data)
    {
      ksi_cancel_input ((ksi_event) e, e->r_data);
      e->r_data = 0;
    }
  if (e->w_data)
    {
      ksi_cancel_output ((ksi_event) e, e->w_data);
      e->w_data = 0;
    }
  if (e->timer_data)
    {
      ksi_cancel_timer ((ksi_event) e, e->timer_data);
      e->timer_data = 0;
    }

  if (e->r_wait && (e->r_ready || e->port->ops->input_ready (e->port)))
    {
      e->r_ready = 0;
      x->result  = ksi_apply_1_with_catch (x->action, (ksi_obj) e);
    }
  else if (e->w_wait && (e->w_ready || e->port->ops->output_ready (e->port)))
    {
      e->w_ready = 0;
      x->result  = ksi_apply_1_with_catch (x->action, (ksi_obj) e);
    }
  else if (!e->no_timeout && ksi_real_time () <= e->time)
    {
      x->state  = KSI->sym_timeout;
      x->result = ksi_apply_1_with_catch (x->action, (ksi_obj) x);
      return 1;
    }

  if (KSI_EXN_P (x->result))
    return 1;

  e->time  = ksi_real_time () + e->timeout;
  port_setup ((ksi_event) e);
  return 0;
}


static struct Ksi_Event_Tag port_ops =
{
  port_name,
  port_init,
  port_setup,
  port_cancel,
  port_invoke
};


ksi_obj
ksi_input_event (ksi_obj tm, ksi_obj port, ksi_obj proc)
{
  double time;
  ksi_port_ready x;

  if (tm == ksi_true || tm == ksi_false)
    time = 0.0;
  else
    {
      TYPE_CHECK (tm, KSI_REAL_P (tm), "make-input-event: invalid real in arg1");
      time = ksi_num2double (tm);
      RANGE_CHECK (tm, time >= 0.0, "make-input-event: negative in arg1");
    }

  TYPE_CHECK (port, KSI_INPUT_PORT_P (port), "make-input-event: invalid input port in arg2");
  TYPE_CHECK (proc, PROC_P (proc), "make-input-event: invalid procedure in arg3");

  x = (ksi_port_ready) ksi_malloc (sizeof *x);
  x->evt.itag   = KSI_TAG_EVENT;
  x->evt.ops    = &port_ops;
  x->evt.state  = KSI->sym_inactive;
  x->evt.action = proc;
  x->evt.result = ksi_void;
  x->port       = (ksi_port) port;
  x->r_wait     = 1;
  x->tick       = 0.1;
  x->timeout    = time;
  x->no_timeout = (tm == ksi_true ? 1 : 0);

  return (ksi_obj) x;
}

ksi_obj
ksi_output_event (ksi_obj tm, ksi_obj port, ksi_obj proc)
{
  double time;
  ksi_port_ready x;

  if (tm == ksi_true || tm == ksi_false)
    time = 0.0;
  else
    {
      TYPE_CHECK (tm, KSI_REAL_P (tm), "make-output-event: invalid real in arg1");
      time = ksi_num2double (tm);
      RANGE_CHECK (tm, time >= 0.0, "make-output-event: negative in arg1");
    }
  TYPE_CHECK (port, KSI_OUTPUT_PORT_P (port), "make-output-event: invalid output port in arg2");
  TYPE_CHECK (proc, PROC_P (proc), "make-output-event: invalid procedure in arg3");

  x = (ksi_port_ready) ksi_malloc (sizeof *x);
  x->evt.itag   = KSI_TAG_EVENT;
  x->evt.ops    = &port_ops;
  x->evt.state  = KSI->sym_inactive;
  x->evt.action = proc;
  x->evt.result = ksi_void;
  x->port       = (ksi_port) port;
  x->w_wait     = 1;
  x->tick       = 0.1;
  x->timeout    = time;
  x->no_timeout = (tm == ksi_true ? 1 : 0);

  return (ksi_obj) x;
}

static ksi_obj
ksi_event_port (ksi_obj x)
{
  TYPE_CHECK (x, KSI_EVT_P (x), "event-port: invalid event in arg1");
  TYPE_CHECK (x, ((ksi_event) x) -> ops == &port_ops, "event-port: invalid event in arg1");
  return (ksi_obj) ((ksi_port_ready) x) -> port;
}

static ksi_obj
ksi_event_port_tick (ksi_obj x, ksi_obj tm)
{
  TYPE_CHECK (x, KSI_EVT_P (x), "event-port-tick: invalid event in arg1");
  TYPE_CHECK (x, ((ksi_event) x) -> ops == &port_ops, "event-port-tick: invalid event in arg1");

  if (tm)
    {
      double tick;
      TYPE_CHECK (tm, KSI_REAL_P (tm), "event-port-tick: invalid event in arg2");
      tick = ksi_num2double (tm);
      TYPE_CHECK (tm, tick > 0.0, "event-port-tick: nonpositive in arg2");
      ((ksi_port_ready) x) -> tick = tick;
    }

  return ksi_double2num (((ksi_port_ready) x) -> tick);
}

void
ksi_cancel_port_events (ksi_port port)
{
  ksi_event curr, next;

  for (curr = KSI->active_events; curr; curr = next)
    {
      next = curr->next;
      if (curr->ops == &port_ops && ((ksi_port_ready) curr) -> port == port)
	ksi_stop_event ((ksi_obj) curr);
    }

  ksi_run_pending_events ();
}


/* -----------------------------------------------------------------
 * idle events
 */

typedef struct Ksi_Idle
{
  struct Ksi_Event	evt;
  void			*data;
} *ksi_idle;

static const char*
idle_name (ksi_event x)
{
  return "idle";
}

static void
idle_init (ksi_event x)
{
}

static void
idle_setup (ksi_event x)
{
  ((ksi_idle) x) -> data = ksi_wait_idle (x, 1);
}

static void
idle_cancel (ksi_event x)
{
  if (((ksi_idle) x) -> data)
    {
      ksi_cancel_idle (x, ((ksi_idle) x) -> data);
      ((ksi_idle) x) -> data = 0;
    }
}

static int
idle_invoke (ksi_event x, void *data)
{
  x->result = ksi_apply_1_with_catch (x->action, (ksi_obj) x);
  if (KSI_EXN_P (x->result))
    return 1;
  return 0;
}

static struct Ksi_Event_Tag idle_ops =
{
  idle_name,
  idle_init,
  idle_setup,
  idle_cancel,
  idle_invoke
};

ksi_obj
ksi_idle_event (ksi_obj proc)
{
  ksi_idle x;
  TYPE_CHECK (proc, PROC_P (proc), "make-idle-event: invalid procedure in arg1");

  x = (ksi_idle) ksi_malloc (sizeof *x);
  x->evt.itag   = KSI_TAG_EVENT;
  x->evt.ops    = &idle_ops;
  x->evt.state  = KSI->sym_inactive;
  x->evt.action = proc;
  x->evt.result = ksi_void;

  return (ksi_obj) x;
}

/* -----------------------------------------------------------------
 * signal events
 */

typedef struct Ksi_Signal
{
  struct Ksi_Event	evt;
  void			*data;
  int			sig;
} *ksi_signal;

static const char*
signal_name (ksi_event x)
{
  return ksi_aprintf ("signal %d", ((ksi_signal) x) -> sig);
}

static void
signal_init (ksi_event x)
{
}

static void
signal_setup (ksi_event x)
{
  ((ksi_signal) x) -> data = ksi_wait_signal (x, ((ksi_signal) x) -> sig, 1);
}

static void
signal_cancel (ksi_event x)
{
  if (((ksi_signal) x) -> data)
    {
      ksi_cancel_signal (x, ((ksi_signal) x) -> data);
      ((ksi_idle) x) -> data = 0;
    }
}

static int
signal_invoke (ksi_event x, void *data)
{
  x->result = ksi_apply_1_with_catch (x->action, (ksi_obj) x);
  if (KSI_EXN_P (x->result))
    return 1;
  return 0;
}

static struct Ksi_Event_Tag signal_ops =
{
  signal_name,
  signal_init,
  signal_setup,
  signal_cancel,
  signal_invoke
};

ksi_obj
ksi_signal_event (ksi_obj sig, ksi_obj proc)
{
  ksi_signal x;
  TYPE_CHECK (sig, KSI_SINT_P (sig), "make-signal-event: invalid procedure in arg2");
  TYPE_CHECK (proc, PROC_P (proc), "make-signal-event: invalid procedure in arg2");

  x = (ksi_signal) ksi_malloc (sizeof *x);
  x->evt.itag   = KSI_TAG_EVENT;
  x->evt.ops    = &signal_ops;
  x->evt.state  = KSI->sym_inactive;
  x->evt.action = proc;
  x->evt.result = ksi_void;
  x->sig        = KSI_SINT_COD (sig);

  return (ksi_obj) x;
}


static struct Ksi_Prim_Def defs [] =
{
  { "event?",			ksi_event_p,		KSI_CALL_ARG1, 1 },
  { "event-state",		ksi_event_state,	KSI_CALL_ARG1, 1 },
  { "event-result",		ksi_event_result,	KSI_CALL_ARG1, 1 },
  { "event-procedure",		ksi_event_procedure,	KSI_CALL_ARG1, 1 },

  { "start-event",		ksi_start_event,	KSI_CALL_ARG1, 1 },
  { "stop-event",		ksi_stop_event,		KSI_CALL_ARG1, 1 },
  { "wait-event",		ksi_wait_event,		KSI_CALL_ARG1, 0 },
  { "sleep",			ksi_sleep,		KSI_CALL_ARG1, 1 },
  { "disable-async-events",	ksi_disable_evt,	KSI_CALL_ARG0, 0 },
  { "enable-async-events",	ksi_enable_evt,		KSI_CALL_ARG0, 0 },

  { "make-ready-event",		ksi_ready_event,	KSI_CALL_ARG1, 1 },
  { "make-timer-event",		ksi_timer_event,	KSI_CALL_ARG2, 2 },
  { "make-idle-event",		ksi_idle_event,		KSI_CALL_ARG1, 1 },
  { "make-signal-event",	ksi_signal_event,	KSI_CALL_ARG2, 2 },

  { "make-input-event",		ksi_input_event,	KSI_CALL_ARG3, 3 },
  { "make-output-event",	ksi_output_event,	KSI_CALL_ARG3, 3 },
  { "event-port",		ksi_event_port,		KSI_CALL_ARG1, 1 },
  { "event-port-tick",		ksi_event_port_tick,	KSI_CALL_ARG2, 1 },

  { 0 }
};


void
ksi_init_events (void)
{
  ksi_ignore (rcsid);
  ksi_reg_unit (defs);
}

void
ksi_term_events (void)
{
  ksi_register_event_mgr (0);
}

 /* End of code */
