/*
 * ksi_gc.c
 * gc function
 *
 * Copyright (C) 1997-2000,2001, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Wed Nov 26 23:14:46 1997
 * Last Update:   Mon Aug 27 15:11:39 2001
 *
 */

/* Comments:
 *
 */

#include "ksi_int.h"

#include <gc/gc.h>
#include <gc/gc_typed.h>

static char rcsid[] = "$Id: ksi_gc.c,v 1.7.6.1 2009/03/01 10:32:34 ksion Exp $";


#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif

#ifdef HAVE_FCNTL_H
#  include <fcntl.h>
#endif

#define TYPE_LEN	16
#define NUMS_LEN	16

static GC_descr typed_descr [TYPE_LEN] [NUMS_LEN];


/* ksi_malloc
 *	 SZ  .
 */

void*
ksi_malloc (size_t sz)
{
  void* ptr;
  ptr = GC_malloc (sz);
  if (!ptr)
    ksi_throw_fatal (KSI->no_mem_exn);

  return ptr;
}

void*
ksi_malloc_data (size_t sz)
{
  void* ptr;

  /* see "gc/gc.h" about 256 */
  if (sz >= 256)
    ptr = GC_malloc_atomic_ignore_off_page (sz);
  else
    ptr = GC_malloc_atomic (sz);

  if (!ptr)
    ksi_throw_fatal (KSI->no_mem_exn);

  return ptr;
}

void*
ksi_malloc_eternal (size_t sz)
{
  void* ptr = GC_malloc_uncollectable (sz);
  if (!ptr)
    ksi_throw_fatal (KSI->no_mem_exn);

  return ptr;
}

void*
ksi_malloc_typed (size_t sz, size_t fp, size_t lp)
{
  void* ptr;
  size_t tp, nm;

  if (fp > lp) tp = fp, fp = lp, lp = tp;
  tp = fp / sizeof (GC_word);
  nm = (lp - fp) / sizeof (GC_word);

  if (tp >= TYPE_LEN || nm >= NUMS_LEN)
    ptr = GC_malloc (sz);
  else
    ptr = GC_malloc_explicitly_typed (sz, typed_descr [tp] [nm]);

  if (!ptr)
    ksi_throw_fatal (KSI->no_mem_exn);

  return ptr;
}

void*
ksi_realloc (void* ptr, size_t sz)
{
  ptr = GC_realloc (ptr, sz);
  if (!ptr)
    ksi_throw_fatal (KSI->no_mem_exn);

  return ptr;
}

void
ksi_free (void* ptr)
{
  GC_free (ptr);
}

void
ksi_register_finalizer (void* obj, ksi_finalizer_function proc, void* data)
{
  GC_finalization_proc old_proc;
  void* old_data;

  GC_register_finalizer_ignore_self (obj, proc, data,
				     &old_proc, &old_data);
}

int
ksi_expand_heap (unsigned sz)
{
  return GC_expand_hp (sz);
}

void
ksi_set_max_heap (unsigned sz)
{
  GC_set_max_heap_size (sz);
}

unsigned
ksi_heap_size (void)
{
  return GC_get_heap_size ();
}

unsigned
ksi_heap_free (void)
{
  return GC_get_free_bytes ();
}

unsigned
ksi_gcollections (void)
{
  return GC_gc_no;
}

int
ksi_gcollect (int full)
{
  if (full)
    {
      while (GC_collect_a_little ());
      GC_gcollect ();
      return 0;
    }

  return GC_collect_a_little ();
}

void
ksi_enable_gc (void)
{
  GC_dont_gc = 0;
}

void
ksi_disable_gc (void)
{
  GC_dont_gc = 1;
}

void
ksi_gc_protect (ksi_obj x)
{
  if (x && KSI_PTR_P (x))
    KSI->protected_objs = CONS (x, KSI->protected_objs);
}

void
ksi_gc_unprotect (ksi_obj x)
{
  ksi_obj cur, *loc;

  loc = & KSI->protected_objs;
  for (cur = *loc; cur != ksi_nil; loc = & CDR (cur), cur = CDR (cur))
    if (cur == x)
      {
	*loc = CDR (cur);
	break;
      }
}


/*
 * ksi_gc_warn
 *	warning procedure that called from GC
 */

static void
ksi_gc_warn (char* msg, GC_word arg)
{
  ksi_warn (msg, arg);
}


#ifdef MSWIN32

extern void GC_init ();
extern void GC_register_root_section (void* static_root);

void
ksi_register_dll (void* ptr)
{
  static int inited = 0;

  if (!inited)
    GC_set_warn_proc (ksi_gc_warn);

  if (ksi_win32s ())
    {
      if (!inited)
	{
	  GC_init ();
#if defined(MAKE_si_LIB)
	  GC_register_root_section (&inited);
#endif
	}

      if (ptr)
	GC_register_root_section (ptr);
    }

  inited = 1;
}

#endif


/*
 * ksi_init_gc
 *	gc initializations -- called from ksi_init_interp
 */

#if defined(MSWIN32)
extern void GC_win32_free_heap (void);
#endif

void
ksi_init_gc (void)
{
  int i, j, n;

  ksi_ignore (rcsid);
  GC_set_warn_proc (ksi_gc_warn);

#if defined(MSWIN32) && defined(KSIDLL)
  ksi_register_dll (0);
#endif

#if GC_INCREMENTAL
  GC_enable_incremental ();
#endif

  for (i = 0; i < TYPE_LEN; i++)
    for (j = 0; j < NUMS_LEN; j++)
      {
	GC_word bm[1];
	for (n = 0, bm [0] = 0; n <= j; n++)
	  bm [0] |= 1 << (n+i);

	typed_descr [i] [j] = GC_make_descriptor (bm, n+i);
      }
}


 /* End of code */
