/*
 * ksi_hash.c
 * hash
 *
 * Copyright (C) 1998-2000,2001, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written
 * agreement is hereby granted, provided that the above copyright notice
 * and this paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Mon Feb 16 02:36:03 1998
 * Last Update:   Mon Apr 16 02:11:41 2001
 *
 */

/* Comments:
 *
 */

#include "ksi_int.h"
#include "ksi_hash.h"
#include "ksi_klos.h"

static char rcsid[] = "$Id: ksi_hash.c,v 1.5 2001/04/29 11:22:06 demakov Exp $";


typedef struct _hashrec
{
  ksi_obj	key;
  ksi_obj	val;
} hashrec;


static unsigned
hasher (ksi_obj x, unsigned n, unsigned d)
{
  if (KSI_OBJ_P (x))
    switch (x->itag)
      {
      case KSI_TAG_SYMBOL:
	return ksi_hash_str (KSI_SYM_PTR (x), KSI_SYM_LEN (x), n);

      case KSI_TAG_STRING:
      case KSI_TAG_CONST_STRING:
	return ksi_hash_str (KSI_STR_PTR (x), KSI_STR_LEN (x), n);

      case KSI_TAG_KEYWORD:
	return ksi_hash_str (KSI_KEY_PTR (x), KSI_KEY_LEN (x), n);

      case KSI_TAG_BIGNUM:
	x = ksi_mod (x, ksi_uint2num (n));
	return (unsigned) ksi_num2uint (x) % n;

      case KSI_TAG_FLONUM:
      case KSI_TAG_RATIONAL:
      case KSI_TAG_COMPLEX:
	{
	  char* ptr = ksi_num2str (x, 10);
	  return ksi_hash_str (ptr, strlen (ptr), n);
	}

      case KSI_TAG_POINTER:
	return (((unsigned) KSI_CPTR_VAL (x)) >> 2) % n;

      case KSI_TAG_PAIR:
      case KSI_TAG_CONST_PAIR:
	if (d)
	  return (hasher (CAR (x), n, d/2) + hasher (CDR (x), n, d/2)) % n;
	break;

      case KSI_TAG_VECTOR:
      case KSI_TAG_CONST_VECTOR:
	{
	  int len = KSI_VEC_LEN (x), i;
	  unsigned h;
	  if (len > 5)
	    {
	      i = d/2; h = 1;
	      while (--i >= 0) 
		h = ((h<<8) + hasher (KSI_VEC_REF (x, h%len), n, 2)) % n;
	      return h;
	    }
	  else
	    {
	      i = len; h = n-1;
	      while (--i >= 0)
		h = ((h<<8) + (hasher (KSI_VEC_REF (x, i), n, d/len))) % n;
	      return h;
	    }
	}

      case KSI_TAG_INSTANCE:
	return ksi_hash_inst (x, n);
      }

  return ((unsigned) x) % n;
}


unsigned
ksi_hash_obj (ksi_obj x, unsigned n)
{
  return hasher (x, n, 10);
}


ksi_obj
ksi_hash (ksi_obj x, ksi_obj n)
{
  TYPE_CHECK (n, KSI_SINT_P (n) && KSI_SINT_COD (n) >= 0,
	      "hash: invalid integer in arg2");

  if (KSI_SINT_COD (n) == 0)
    return ksi_zero;

  return ksi_uint2num (hasher (x, KSI_SINT_COD (n), 10));
}

ksi_obj
ksi_hashv (ksi_obj x, ksi_obj n)
{
  TYPE_CHECK (n, KSI_SINT_P (n) && KSI_SINT_COD (n) >= 0,
	      "hashv: invalid integer in arg2");

  if (KSI_SINT_COD (n) == 0)
    return ksi_zero;

  if (KSI_NUM_P (x))
    return ksi_uint2num (hasher (x, KSI_SINT_COD (n), 10));
  return KSI_MK_SINT (KSI_COD (x) % KSI_SINT_COD (n));
}

ksi_obj
ksi_hashq (ksi_obj x, ksi_obj n)
{
  TYPE_CHECK (n, KSI_SINT_P (n) && KSI_SINT_COD (n) >= 0,
	     "hashq: invalid integer in arg2");

  if (KSI_SINT_COD (n) == 0)
    return ksi_zero;

  return KSI_MK_SINT (KSI_COD (x) % KSI_SINT_COD (n));
}


static unsigned
hash_rec (void* val, unsigned size, void* data)
{
  ksi_obj hash = KSI_HASHTAB_HASH (data);
  if (!hash)
    return hasher (((hashrec*) val) -> key, size, 10);

  return (unsigned) ksi_num2uint (ksi_apply_2 (hash,
					       ((hashrec*) val) -> key,
					       ksi_uint2num (size)));
}

static int
cmp_rec (void* v1, void* v2, void* data)
{
  ksi_obj cmp = KSI_HASHTAB_CMP (data);
  if (!cmp)
    return ! (KSI_EQUAL_P (((hashrec*) v1) -> key,
			   ((hashrec*) v2) -> key));

  return ! (KSI_TRUE_P (ksi_apply_2 (cmp,
				     ((hashrec*) v1) -> key,
				     ((hashrec*) v2) -> key)));
}

ksi_obj
ksi_new_hashtab (int size, ksi_obj hash, ksi_obj cmp)
{
  ksi_hashtab tab = NEW_TYPED (struct Ksi_Hashtab, table, cmp);
  tab->itag = KSI_TAG_HASHTAB;
  tab->table = ksi_new_valtab (size, hash_rec, cmp_rec, tab);
  tab->hash = hash;
  tab->cmp = cmp;
  return (ksi_obj) tab;
}

ksi_obj
ksi_hashtab_p (ksi_obj x)
{
  return KSI_HASHTAB_P (x) ? ksi_true : ksi_false;
}

ksi_obj
ksi_make_hashtab (ksi_obj size, ksi_obj hash, ksi_obj cmp)
{
  if (!size)
    size = ksi_zero;
  else
    TYPE_CHECK (size, KSI_SINT_P (size) && KSI_SINT_COD (size) >= 0,
		"make-hash-table: invalid integer in arg1");

  TYPE_CHECK (hash, !hash || PROC_P (hash),
	      "make-hash-table: invalid procedure in arg2");
  TYPE_CHECK (cmp, !cmp || PROC_P (cmp),
	      "make-hash-table: invalid procedure in arg3");

  return ksi_new_hashtab (KSI_SINT_COD (size), hash, cmp);
}

ksi_obj
ksi_hash_ref (ksi_obj tab, ksi_obj key, ksi_obj def)
{
  hashrec r, *p;
  TYPE_CHECK (tab, KSI_HASHTAB_P (tab),
	      "hash-ref: invalid hash-table in arg1");

  r.key = key;
  p = (hashrec*) ksi_lookup_vtab (KSI_HASHTAB_TAB (tab), &r);
  if (p)
    return p -> val;

  return (def ? def : ksi_void);
}

ksi_obj
ksi_hash_set_x (ksi_obj tab, ksi_obj key, ksi_obj val)
{
  hashrec r, *p;
  TYPE_CHECK (tab, KSI_HASHTAB_P (tab),
	      "hash-set!: invalid hash-table in arg1");

  r.key = key;
  p = (hashrec*) ksi_lookup_vtab (KSI_HASHTAB_TAB (tab), &r);
  if (p)
    p -> val = val;
  else
    {
      p = (hashrec*) ksi_malloc (sizeof *p);
      p -> key = key;
      p -> val = val;
      ksi_append_vtab (KSI_HASHTAB_TAB (tab), p);
    }

  return ksi_void;
}

ksi_obj
ksi_hash_remove_x (ksi_obj tab, ksi_obj key)
{
  hashrec r;
  TYPE_CHECK (tab, KSI_HASHTAB_P (tab),
	      "hash-remove!: INVALID hash-table in arg1");

  r.key = key;
  ksi_remove_vtab (KSI_HASHTAB_TAB (tab), &r);

  return ksi_void;
}


/*
 * ksi_hash_map, ksi_hash_for_each
 *  applies PROC for each elem in hash-table TAB.
 *
 * Return:
 *  ksi_hash_map -- list of results
 *  ksi_hash_for_each -- ksi_void
 *
 * Note:
 *  functions use direct access in internal structure of hash-tables
 */

ksi_obj
ksi_hash_map (ksi_obj proc, ksi_obj tab)
{
  unsigned i;
  struct Ksi_Tabrec *rec;
  ksi_obj res = ksi_nil;

  TYPE_CHECK (proc, PROC_P (proc), "hash-map: invalid procedure in arg1");
  TYPE_CHECK (tab, KSI_HASHTAB_P (tab), "hash-map: invalid hash-table in arg2");

  for (i = 0; i < KSI_HASHTAB_TAB (tab) -> size; i++)
    {
      for (rec = KSI_HASHTAB_TAB (tab) -> table [i]; rec; rec = rec -> next)
	{
	  ksi_obj key = ((hashrec*) rec -> val) -> key;
	  ksi_obj val = ((hashrec*) rec -> val) -> val;
	  res = CONS (ksi_apply_2 (proc, key, val), res);
	  CHECK_EVENTS;
	}
    }

  return res;
}

ksi_obj
ksi_hash_for_each (ksi_obj proc, ksi_obj tab)
{
  unsigned i;
  struct Ksi_Tabrec *rec;

  TYPE_CHECK (proc, PROC_P (proc), "hash-for-each: invalid procedure in arg1");
  TYPE_CHECK (tab, KSI_HASHTAB_P (tab),
	      "hash-for-each: invalid hash-table in arg2");

  for (i = 0; i < KSI_HASHTAB_TAB (tab) -> size; i++)
    {
      for (rec = KSI_HASHTAB_TAB (tab) -> table [i]; rec; rec = rec -> next)
	{
	  ksi_obj key = ((hashrec*) rec -> val) -> key;
	  ksi_obj val = ((hashrec*) rec -> val) -> val;
	  ksi_apply_2 (proc, key, val);
	  CHECK_EVENTS;
	}
    }

  return ksi_void;
}


static struct Ksi_Prim_Def defs [] =
{
  { "hash",		ksi_hash,		KSI_CALL_ARG2, 2 },
  { "hashv",		ksi_hashv,		KSI_CALL_ARG2, 2 },
  { "hashq",		ksi_hashq,		KSI_CALL_ARG2, 2 },

  { "hash-table?",	ksi_hashtab_p,		KSI_CALL_ARG1, 1 },
  { "make-hash-table",	ksi_make_hashtab,	KSI_CALL_ARG3, 0 },
  { "hash-ref",		ksi_hash_ref,		KSI_CALL_ARG3, 2 },
  { "hash-set!",	ksi_hash_set_x,		KSI_CALL_ARG3, 3 },
  { "hash-remove!",	ksi_hash_remove_x,	KSI_CALL_ARG2, 2 },
  { "hash-map",		ksi_hash_map,		KSI_CALL_ARG2, 2 },
  { "hash-for-each",	ksi_hash_for_each,	KSI_CALL_ARG2, 2 },

  {0}
};

void
ksi_init_hash ()
{
  ksi_ignore (rcsid);
  ksi_reg_unit (defs);
}

 /* End of code */
