/*
 * ksi_jump.c
 * long jump & continuation
 *
 * Copyright (C) 1997-2000,2001, Ivan Demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Fri Aug  8 00:01:49 1997
 * Last Update:   Tue Mar 27 19:58:56 2001
 *
 */

/* Comments:
 *
 */

#include "ksi_int.h"
#include "ksi_jump.h"


static char rcsid[] = "$Id: ksi_jump.c,v 1.3 2001/03/27 16:42:19 demakov Exp $";


static void copy_stack (struct Ksi_Jump* buf, ksi_stack start, ksi_stack end);
static void uncopy_stack (ksi_stack a, struct Ksi_Jump* buf);
static void restore_stack (ksi_stack a, struct Ksi_Jump* buf);
static void dowind (ksi_wind w);
static void unwind (ksi_wind wind);


/*
 * copy_stack
 *	      
 *	   .
 */

static void
copy_stack (struct Ksi_Jump* buf, ksi_stack start, ksi_stack here)
{
  int size;

#ifdef KSI_STACK_GROWS_UP
  size = here - start;
  buf->from = start;
#else
  size = start - here;
  buf->from = here;
#endif

  if (size > 0)
    {
      buf->stack = (ksi_stack) ksi_malloc (size * sizeof (*here));
      buf->size  = size;
    }
  else
    {
      buf->stack = 0;
      buf->size  = 0;
    }

  for (here = buf->stack, start = buf->from;
       --size >= 0;
       *here++ = *start++);
}


/*
 * restore_stack
 *	   Ksi_Jump::uncopy_stack    
 *	    
 *	.
 */

static void
restore_stack (ksi_stack a, struct Ksi_Jump* buf)
{
  KSI_STACK_ELEM junk[100];

#if KSI_STACK_GROWS_UP
  junk[0]  = (KSI_STACK_ELEM) STK_CMP(&junk[0], (buf->from + buf->size));
#else
  junk[0]  = (KSI_STACK_ELEM) STK_CMP(&junk[0], buf->from);
#endif
  junk[1]  = (KSI_STACK_ELEM) a[1];
  junk[99] = (KSI_STACK_ELEM) junk;

  uncopy_stack (junk, buf);
}


/*
 * uncopy_stack
 *	        ,
 *	  Ksi_Jump::restore_stack, ,  
 *	   ,   uncopy_stack.
 */

static void
uncopy_stack (ksi_stack a, struct Ksi_Jump* buf)
{
  ksi_stack from;
  ksi_stack copy;
  int size;

  if (!a[0])
    restore_stack (a, buf);

  FLUSH_REGISTER_WINDOWS;
  from = buf->from;
  copy = buf->stack;
  size = buf->size;
  while (--size >= 0)
    *from++ = *copy++;

  if (buf->wind != KSI->wind)
    dowind (buf->wind);

  longjmp (buf->j_buf, 1);
}


static void
dowind (ksi_wind w)
{
  if (w != KSI->wind)
    {
      dowind (w->cont);
      if (w->pre) ksi_apply_0 (w->pre);
      KSI->wind = w;
    }
}


static void
unwind (ksi_wind wind)
{
  int len1 = 0, len2 = 0;
  ksi_wind w1, w2;

  for (w1 = wind; w1; w1 = w1->cont) ++len1;
  for (w2 = KSI->wind; w2; w2 = w2->cont) ++len2;

  for (w1 = wind; len1 > len2; w1 = w1->cont) --len1;
  for (w2 = KSI->wind; len2 > len1; w2 = w2->cont) ++len1;

  while (w1 != w2)
    {
      w1 = w1->cont;
      w2 = w2->cont;
    }

  /* w1 == w2 */
  while (KSI->wind != w1)
    {
      ksi_obj val = KSI->wind->post;
      KSI->wind = KSI->wind->cont;
      if (val) ksi_apply_0 (val);
    }
}

void
ksi_init_jump (ksi_jump b, ksi_stack start, ksi_stack here)
{
  ksi_ignore (rcsid);

  /*
   * Store all needed values in b before copiing stack, since b may be placed
   * on stack and when we will restore stack this values will not corrupted.
   */

  b->module = KSI->curr_module;
  b->wind   = KSI->wind;
  if (start)
    copy_stack (b, start, here);
  else
    {
      b->stack = 0;
      b->size = 0;
    }
}

ksi_obj
ksi_set_jump (ksi_jump b, ksi_stack start)
{
  KSI_STACK_ELEM here;
  ksi_ignore (rcsid);

  ksi_init_jump (b, start, &here);

  FLUSH_REGISTER_WINDOWS;
  if (setjmp (b->j_buf))
    return KSI->jump_val;

  return 0;
}

void
ksi_long_jump (ksi_jump buf, ksi_obj val)
{
  KSI_STACK_ELEM z[2];

  if (buf->wind != KSI->wind)
    unwind (buf->wind);

  KSI->curr_module = buf->module;
  KSI->jump_val    = val;

  if (buf->stack == 0)
    longjmp (buf->j_buf, 1);

#if KSI_STACK_GROWS_UP
  z[0] = (KSI_STACK_ELEM) STK_CMP(&z[0], (buf->from + buf->size));
#else
  z[0] = (KSI_STACK_ELEM) STK_CMP(&z[0], buf->from);
#endif
  z[1] = (KSI_STACK_ELEM) z;

  uncopy_stack (z, buf);
}


 /* End of code */
