/*
 * ksi_load.c
 * loading
 *
 * Copyright (C) 1997-2000,2001, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Sun Aug 10 21:42:05 1997
 * Last Update:   Fri Apr 13 23:36:01 2001
 *
 */

/* Comments:
 *
 */

#include "ksi_int.h"

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#ifdef HAVE_FCNTL_H
#  include <fcntl.h>
#endif

#ifdef HAVE_SYS_STAT_H
#  include <sys/stat.h>
#endif

#ifdef HAVE_SYS_MMAN_H
#  include <sys/mman.h>
#  ifndef MAP_FAILED
#    define MAP_FAILED ((caddr_t) -1)
#  endif
#endif

#ifdef _MSC_VER
#  include <io.h>
#  define F_OK 0
#endif

static char rcsid[] = "$Id: ksi_load.c,v 1.11 2001/04/13 16:56:21 demakov Exp $";

#ifndef O_BINARY
#  define O_BINARY 0
#endif


enum { LOAD_OK, LOAD_ERR, NO_FILE };

static struct
{
  char major, minor, level;
} ko_version[] =
  {
    { 3, 4, 0 },
    { KSI_MAJOR_VERSION, KSI_MINOR_VERSION, KSI_PATCH_LEVEL }
  };


static int
correct_version (int major, int minor, int level)
{
  if (major < ko_version[0].major)
    return 0;
  if (major == ko_version[0].major)
    {
      if (minor < ko_version[0].minor)
	return 0;
      if (minor == ko_version[0].minor)
	if (level < ko_version[0].level)
	  return 0;
    }

  if (major > ko_version[1].major)
    return 0;
  if (major == ko_version[1].major)
    {
      if (minor > ko_version[1].minor)
	return 0;
      if (minor == ko_version[1].minor)
	if (level > ko_version[1].level)
	  return 0;
    }

  return 1;
}

static int
ksi_load_msg (const char* fname)
{
  ksi_obj verbose;
  int i;

  if (!KSI || KSI->inited <= 2 || !KSI->interactive)
    return 0;

  verbose = ksi_spec_val ("*load-verbose*", 0, 0);
  if (verbose != ksi_false)
    {
      ksi_obj err = ksi_current_error_port ();
      if (err && KSI_OUTPUT_PORT_P (err))
	{
	  ksi_port_write ((ksi_port) err, "; ", 2);
	  for (i = 0; i < KSI->load_indent; i++)
	    ksi_port_putc ((ksi_port) err, ' ');
	  ksi_port_write ((ksi_port) err, "Loading ", 8);
	  ksi_port_write ((ksi_port) err, fname, strlen (fname));
	  ksi_port_putc ((ksi_port) err, '\n');
	}

      KSI->load_indent += 2;
      return 1;
    }

  return 0;
}

static int
ksi_done_msg (const char* msg, int error, int verbose)
{
  int i;
  ksi_obj err;

  if (!verbose)
    return error;

  KSI->load_indent -= 2;
  if (error != LOAD_OK)
    return error;

  err = ksi_current_error_port ();
  if (err && KSI_OUTPUT_PORT_P (err))
    {
      ksi_port_write ((ksi_port) err, "; ", 2);
      for (i = 0; i < KSI->load_indent; i++)
	ksi_port_putc ((ksi_port) err, ' ');
      ksi_port_write ((ksi_port) err, "Done ", 5);
      ksi_port_write ((ksi_port) err, msg, strlen (msg));
      ksi_port_putc ((ksi_port) err, '\n');
    }

  return error;
}

static int
ksi_load_dumped (const char* fname)
{
  volatile int fd;
  volatile ksi_wind wind;
  volatile ksi_obj port, val;
  volatile char *loading_file;
  int major, minor, level, tmp;

  if ((fd = open (fname, O_RDONLY | O_BINARY)) < 0)
    return (errno == ENOENT ? NO_FILE : LOAD_ERR);

  port = ksi_make_fd_port (fd, fname, "r");
  val  = ksi_read (port);
  while (1)
    {
      tmp = ksi_port_getc ((ksi_port) port);
      if (tmp <= 0)
	break;
    }

  major = ksi_port_getc ((ksi_port) port);
  minor = ksi_port_getc ((ksi_port) port);
  level = ksi_port_getc ((ksi_port) port);
  if (val == ksi_eof || major < 0 || minor < 0 || level < 0 || tmp < 0)
    {
      ksi_close_port (port);
      KSI->load_err_msg = "invalid file header";
      return LOAD_ERR;
    }
  if (!correct_version (major, minor, level))
    {
      ksi_close_port (port);
      KSI->load_err_msg = "invalid version, file need recompile";
      return LOAD_ERR;
    }

  loading_file = KSI->loading_file;
  KSI->loading_file = (char*) fname;

  wind = ksi_add_catch (ksi_true, 0, 0);
  FLUSH_REGISTER_WINDOWS;
  if (setjmp (wind->the_catch->jmp.j_buf) == 0)
    {
      for (;;)
        {
	  if ((val = ksi_read_dump (port)) == ksi_eof)
	    break;
          ksi_eval_sexp (val, 0);
        }

      ksi_del_catch (wind);
      KSI->loading_file = (char*) loading_file;
      ksi_close_port (port);
      return LOAD_OK;
    }

  ksi_handle_error (wind->the_catch->thrown, wind->the_catch->value);
  KSI->load_err_msg = "error while loading";
  KSI->loading_file = (char*) loading_file;
  ksi_close_port (port);
  return LOAD_ERR;
}

static int
ksi_load_file (const char* fname, int need_expand)
{
  volatile int fd;
  volatile ksi_port port, prev_port;
  volatile ksi_wind wind;
  volatile char *loading_file;
  volatile int load_err = 0;

  if (access (fname, F_OK) != 0)
    return NO_FILE;
  if ((fd = open (fname, O_RDONLY)) < 0)
    return LOAD_ERR;

  port = (ksi_port) ksi_make_fd_port (fd, fname, "r");

  prev_port = KSI->curr_port;
  KSI->curr_port = port;

  loading_file = KSI->loading_file;
  KSI->loading_file = (char*) fname;

  wind = ksi_add_catch (ksi_true, 0, 0);
  FLUSH_REGISTER_WINDOWS;
  if (setjmp (wind -> the_catch -> jmp . j_buf) == 0)
    {
      for (;;)
        {
	  ksi_obj val = ksi_read_const ((ksi_obj) port);
          if (val == ksi_eof)
            break;

	  if (need_expand)
	    ksi_eval (val, 0);
	  else
	    ksi_eval_sexp (val, 0);
        }

      ksi_del_catch (wind);
    }
  else
    {
      ksi_obj tag = wind -> the_catch -> thrown;
      ksi_obj val = wind -> the_catch -> value;
      ksi_handle_error (tag, val);
      KSI->load_err_msg = "error while loading";
      load_err = 1;
    }

  KSI->loading_file = (char*) loading_file;
  KSI->curr_port = prev_port;
  ksi_close_port ((ksi_obj) port);

  return load_err ? LOAD_ERR : LOAD_OK;
}

static int
ksi_load_lib (const char *fname)
{
  struct Ksi_Stat_Lib *curr;

  for (curr = KSI->stat_libs; curr; curr = curr->next)
    if (strpathcmp (fname, curr->name) == 0)
      {
	curr->proc (curr->data);
	return LOAD_OK;
      }

  return NO_FILE;
}

void
ksi_load_init_file (const char* fn)
{
  int err, len;
  char *tmp;

  if (ABSOLUTE_PATH (fn) || fn [0] == '~')
    fn = ksi_expand_file_name (fn);
  else
    {
      len = strlen (KSI->ksi_init_path) + strlen (DIR_SEP) + strlen (fn);
      tmp = (char*) alloca (len + 1);

      strcpy (tmp, KSI->ksi_init_path);
      strcat (tmp, DIR_SEP);
      strcat (tmp, fn);
      fn = tmp;
    }

  if (ksi_load_lib (fn) == LOAD_OK)
    return;

  if (ksi_has_suffix (fn, KSI_SCM_SUFFIX))
    err = ksi_load_file (fn, 1);
  else if (ksi_has_suffix (fn, KSI_KO_SUFFIX))
    err = ksi_load_dumped (fn);
  else
    {
      len = strlen (KSI_KO_SUFFIX) + strlen (KSI_SCM_SUFFIX) + strlen (fn);
      tmp = (char*) alloca (len + 1);

      strcpy (tmp, fn); strcat (tmp, KSI_KO_SUFFIX);
      err = ksi_load_dumped (tmp);

      if (err == NO_FILE)
	{
	  strcpy (tmp, fn); strcat (tmp, KSI_SCM_SUFFIX);
	  err = ksi_load_file (tmp, 1);
	}
    }

  if (err == LOAD_OK)
    return;

  ksi_error ("Cannot load \"%s\".\n\t%s\n\tExecution aborted.",
	     fn, KSI->load_err_msg ? KSI->load_err_msg : strerror (errno));
}


static int
ksi_try_load_file (const char* prefix, const char* fname, int try_suffix)
{
  char *full_name;

  full_name = ksi_aprintf ("%s%s%s", prefix, *prefix ? DIR_SEP : "", fname);
  if (ksi_load_lib (full_name) == LOAD_OK)
    return LOAD_OK;

  if (access (full_name, F_OK) == 0)
    {
      int verb = ksi_load_msg (full_name);

#if defined(DYNAMIC_LINKING) && defined(KSI_DL_SUFFIX)
      if (ksi_has_suffix (full_name, KSI_DL_SUFFIX))
	{
	  if ((KSI->load_err_msg = ksi_dynload_file (full_name)) != 0)
	    return ksi_done_msg (KSI->load_err_msg, LOAD_ERR, verb);

	  return ksi_done_msg (full_name, LOAD_OK, verb);
	}
#endif

      if (ksi_has_suffix (full_name, KSI_KO_SUFFIX))
	return ksi_done_msg (full_name, ksi_load_dumped (full_name), verb);

      return ksi_done_msg (full_name, ksi_load_file (full_name, 1), verb);
    }

  if (!try_suffix
#if defined(DYNAMIC_LINKING) && defined(KSI_DL_SUFFIX)
      || ksi_has_suffix (full_name, KSI_DL_SUFFIX)
#endif
      || ksi_has_suffix (full_name, KSI_KO_SUFFIX)
      || ksi_has_suffix (full_name, KSI_SCM_SUFFIX))
    return NO_FILE;

#if defined(DYNAMIC_LINKING) && defined(KSI_DL_SUFFIX)
  full_name = ksi_aprintf ("%s%s%s%s", prefix, *prefix ? DIR_SEP : "", fname,
			   KSI_DL_SUFFIX);

  if (access (full_name, F_OK) == 0)
    {
      int verb = ksi_load_msg (full_name);
      if ((KSI->load_err_msg = ksi_dynload_file (full_name)) != 0)
        return ksi_done_msg (full_name, LOAD_ERR, verb);

      return ksi_done_msg (full_name, LOAD_OK, verb);
    }
#endif

  full_name = ksi_aprintf ("%s%s%s%s", prefix, *prefix ? DIR_SEP : "", fname,
			   KSI_KO_SUFFIX);
  if (access (full_name, F_OK) == 0)
    {
      int verb = ksi_load_msg (full_name);
      return ksi_done_msg (full_name, ksi_load_dumped (full_name), verb);
    }

  full_name = ksi_aprintf ("%s%s%s%s", prefix, *prefix ? DIR_SEP : "", fname,
			   KSI_SCM_SUFFIX);
  if (access (full_name, F_OK) == 0)
    {
      int verb = ksi_load_msg (full_name);
      return ksi_done_msg (full_name, ksi_load_file (full_name, 1), verb);
    }

  /* this is kludge to load preprocessed `psyntax.ks' */
  full_name = ksi_aprintf ("%s%s%s%s", prefix, *prefix ? DIR_SEP : "", fname,
			   ".ks");
  if (access (full_name, F_OK) == 0)
    {
      int verb = ksi_load_msg (full_name);
      return ksi_done_msg (full_name, ksi_load_file (full_name, 0), verb);
    }

  return NO_FILE;
}


ksi_obj
ksi_load (ksi_obj str)
{
  int err;
  const char* fname = ksi_mk_filename (str, "load");

  KSI->load_err_msg = 0;
  if (ABSOLUTE_PATH (fname) || fname [0] == '~')
    {
      fname = ksi_expand_file_name (fname);
      err = ksi_try_load_file ("", fname, 0);
      if (err == LOAD_OK)
	return ksi_void;
    }
  else
    {
      ksi_obj load_path = ksi_spec_val ("*load-path*", 0, 0);
      while (KSI_PAIR_P (load_path))
	{
	  const char* pname = ksi_mk_filename (KSI_CAR (load_path), "load");

	  if ((err = ksi_try_load_file (pname, fname, 1)) == LOAD_OK)
	    return ksi_void;
	  if (err != NO_FILE)
	    break;

	  load_path = KSI_CDR (load_path);
	}
    }

  ksi_exn_error ("system", ksi_str02string (fname),
		 "load: %s",
		 KSI->load_err_msg ? KSI->load_err_msg : strerror (errno));
  return ksi_void;
}

ksi_obj
ksi_try_load (ksi_obj str, ksi_obj init)
{
  int err;
  const char* fname = ksi_mk_filename (str, "@try-load");
  fname = ksi_expand_file_name (fname);

  err = ksi_try_load_file ("", fname, 1);
  if (err == LOAD_OK)
    return ksi_true;

  if (err == NO_FILE)
    return ksi_false;

  if (init == ksi_true)
    ksi_error ("Cannot load \"%s\".\n\t%s\n\tExecution aborted.",
	       fname,
	       KSI->load_err_msg ? KSI->load_err_msg : strerror (errno));

  return ksi_false;
}


static ksi_obj
ksi_load_dump (ksi_obj str)
{
  int verb;
  const char* fname = ksi_mk_filename (str, "@load-dump");
  fname = ksi_expand_file_name (fname);

  KSI->load_err_msg = 0;
  verb = ksi_load_msg (fname);

  ksi_done_msg (fname, ksi_load_dumped (fname), verb);
  return ksi_void;
}

static struct Ksi_Prim_Def defs [] =
{
  { "load",			ksi_load,		KSI_CALL_ARG1, 1 },
  { "@try-load",		ksi_try_load,		KSI_CALL_ARG2, 1 },
  { "@load-dump",		ksi_load_dump,		KSI_CALL_ARG1, 1 },

  { 0 }
};

void
ksi_init_load ()
{
  ksi_ignore (rcsid);
  ksi_reg_unit (defs);
}

 /* End of code */
