/*
 * ksi_misc.c
 * misc utils
 *
 * Copyright (C) 1997-2000,2001, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Tue Jun 24 03:33:49 1997
 * Last Update:Sat Nov 10 00:19:41 2001
 *
 */


#include "ksi_int.h"
#include "ksi_vtab.h"

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif

#include <math.h>
#include <float.h>


static char rcsid[] = "$Id: ksi_misc.c,v 1.16 2001/11/25 14:30:01 demakov Exp $";


ksi_obj
ksi_new_imm (int tag)
{
  ksi_obj x = (ksi_obj) ksi_malloc_data (sizeof *x);
  x -> itag = tag;
  return x;
}

ksi_obj
ksi_make_const_pointer (const void *val, size_t size)
{
  ksi_pointer ptr = NEW_TYPED (struct Ksi_Pointer, val, val);
  ptr -> itag = KSI_TAG_POINTER;
  ptr -> size = size;
  ptr -> val = (void*) val;
  return (ksi_obj) ptr;
}

ksi_obj
ksi_make_pointer (void *val, size_t size)
{
  ksi_pointer ptr = NEW_TYPED (struct Ksi_Pointer, val, val);
  ptr -> itag = KSI_TAG_POINTER;
  ptr -> size = size;
  ptr -> val = val;
  return (ksi_obj) ptr;
}

static ksi_obj
ksi_get_size (ksi_obj pnt)
{
  TYPE_CHECK (pnt, KSI_CPTR_P (pnt), "@get-size: invalid pointer in arg1");
  return ksi_uint2num (KSI_CPTR_SZ (pnt));
}

static ksi_obj
ksi_peek_byte (ksi_obj pnt, ksi_obj offset)
{
  int of, sz;
  unsigned char *start;
  TYPE_CHECK (pnt, KSI_CPTR_P (pnt), "@peek-byte: invalid pointer in arg1");
  TYPE_CHECK (offset, KSI_SINT_P (offset), "@peek-byte: invalid offset in arg2");

  of = ksi_num2int (offset);
  sz = sizeof (char);
  RANGE_CHECK (offset, sz > 0 && 0 <= of && of+sz <= KSI_CPTR_SZ (pnt),
	       "@peek-byte: out of range");

  start = ((unsigned char*) KSI_CPTR_VAL (pnt)) + of;
  return ksi_int2num (*start);
}

static ksi_obj
ksi_peek_int (ksi_obj pnt, ksi_obj offset)
{
  int of, sz;
  int *start;
  TYPE_CHECK (pnt, KSI_CPTR_P (pnt), "@peek-int: invalid pointer in arg1");
  TYPE_CHECK (offset, KSI_SINT_P (offset), "@peek-int: invalid offset in arg2");

  of = ksi_num2int (offset);
  sz = sizeof (int);
  RANGE_CHECK (offset, sz > 0 && 0 <= of && of+sz <= KSI_CPTR_SZ (pnt),
	       "@peek-int: out of range");

  start = (int*) (((char*) KSI_CPTR_VAL (pnt)) + of);
  return ksi_int2num (*start);
}

static ksi_obj
ksi_peek_double (ksi_obj pnt, ksi_obj offset)
{
  int of, sz;
  double *start;
  TYPE_CHECK (pnt, KSI_CPTR_P (pnt), "@peek-double: invalid pointer in arg1");
  TYPE_CHECK (offset, KSI_SINT_P (offset), "@peek-double: invalid offset in arg2");

  of = ksi_num2int (offset);
  sz = sizeof (double);
  RANGE_CHECK (offset, sz > 0 && 0 <= of && of+sz <= KSI_CPTR_SZ (pnt),
	       "@peek-double: out of range");

  start = (double*) (((char*) KSI_CPTR_VAL (pnt)) + of);
  return ksi_double2num (*start);
}

static ksi_obj
ksi_poke_byte (ksi_obj pnt, ksi_obj offset, ksi_obj val)
{
  int of, sz, v;
  unsigned char *start;
  TYPE_CHECK (pnt, KSI_M_CPTR_P (pnt), "@poke-byte: invalid or const pointer in arg1");
  TYPE_CHECK (offset, KSI_SINT_P (offset), "@poke-byte: invalid offset in arg2");
  TYPE_CHECK (val, KSI_EINT_P (val), "@poke-byte: invalid value in arg3");

  of = ksi_num2int (offset);
  sz = sizeof (char);
  v = ksi_num2int (val);
  RANGE_CHECK (offset, sz > 0 && 0 <= of && of+sz <= KSI_CPTR_SZ (pnt),
	       "@poke-byte: out of range");

  start = ((unsigned char*) KSI_CPTR_VAL (pnt)) + of;
  *start = (unsigned char) v;
  return ksi_int2num (of + sz);
}

static ksi_obj
ksi_poke_int (ksi_obj pnt, ksi_obj offset, ksi_obj val)
{
  int of, sz, v;
  int *start;
  TYPE_CHECK (pnt, KSI_M_CPTR_P (pnt), "@poke-int: invalid or const pointer in arg1");
  TYPE_CHECK (offset, KSI_SINT_P (offset), "@poke-int: invalid offset in arg2");
  TYPE_CHECK (val, KSI_EINT_P (val), "@poke-int: invalid value in arg3");

  of = ksi_num2int (offset);
  sz = sizeof (int);
  v = ksi_num2int (val);
  RANGE_CHECK (offset, sz > 0 && 0 <= of && of+sz <= KSI_CPTR_SZ (pnt),
	       "@poke-int: out of range");

  start = (int*) (((unsigned char*) KSI_CPTR_VAL (pnt)) + of);
  *start = v;
  return ksi_int2num (of + sz);
}

static ksi_obj
ksi_poke_double (ksi_obj pnt, ksi_obj offset, ksi_obj val)
{
  int of, sz;
  double v, *start;
  TYPE_CHECK (pnt, KSI_M_CPTR_P (pnt), "@poke-double: invalid or const pointer in arg1");
  TYPE_CHECK (offset, KSI_SINT_P (offset), "@poke-double: invalid offset in arg2");
  TYPE_CHECK (val, KSI_REAL_P (val), "@poke-double: invalid value in arg3");

  of = ksi_num2int (offset);
  sz = sizeof (double);
  v = ksi_num2double (val);
  RANGE_CHECK (offset, sz > 0 && 0 <= of && of+sz <= KSI_CPTR_SZ (pnt),
	       "@poke-double: out of range");

  start = (double*) (((unsigned char*) KSI_CPTR_VAL (pnt)) + of);
  *start = v;
  return ksi_int2num (of + sz);
}


/*--------- format --------------------------------------------------*/

static ksi_obj
coerce2num (ksi_obj x)
{
  if (KSI_NUM_P (x))
    return x;
  if (KSI_CHAR_P (x))
    return ksi_uint2num (KSI_CHAR_COD (x));
  if (x == ksi_false)
    return ksi_zero;
  if (x == ksi_true)
    return ksi_one;
  return ksi_uint2num ((unsigned long) x);
}

static int
ksi_internal_format (ksi_port port, const char* fmt,
		     int argc, ksi_obj* argv, char* nm)
{
  int nums;
  ksi_obj tmp;
  char *buf;

  for (nums = 0; *fmt; fmt++)
    {
      if (*fmt == '~' && *(fmt+1) != '\0')
	{
	  fmt++;
	  switch (*fmt)
	    {
	    case 'a': case 'A':
	      if (nums >= argc)
		{
		too_many:
		  ksi_error ("%s: too many ~ spec's in format string", nm);
		}
	      ksi_display (argv [nums++], (ksi_obj) port);
	      break;

	    case 's': case 'S':
	    case 'c': case 'C':
	      if (nums >= argc)
		goto too_many;
	      ksi_write (argv [nums++], (ksi_obj) port);
	      break;

	    case 'd': case 'D':
	      if (nums >= argc)
		goto too_many;
	      tmp = coerce2num (argv [nums++]);
	      buf = ksi_num2str (tmp, 10);
	      ksi_port_write (port, buf, strlen (buf));
	      break;

	    case 'b': case 'B':
	      if (nums >= argc)
		goto too_many;
	      tmp = coerce2num (argv [nums++]);
	      buf = ksi_num2str (tmp, 2);
	      ksi_port_write (port, buf, strlen (buf));
	      break;

	    case 'o': case 'O':
	      if (nums >= argc)
		goto too_many;
	      tmp = coerce2num (argv [nums++]);
	      buf = ksi_num2str (tmp, 8);
	      ksi_port_write (port, buf, strlen (buf));
	      break;

	    case 'x': case 'X':
	      if (nums >= argc)
		goto too_many;
	      tmp = coerce2num (argv [nums++]);
	      buf = ksi_num2str (tmp, 16);
	      ksi_port_write (port, buf, strlen (buf));
	      break;

	    case '?':
	      if (nums+1 >= argc)
		goto too_many;
	      {
		ksi_obj f = argv [nums++];
		ksi_obj x = argv [nums++];
		ksi_obj *av;
		int     ac, i;

		if (!KSI_STR_P (f))
		  ksi_error ("%s: invalid format string for ~? spec: %s",
			     nm, ksi_obj2str (f));
		if (KSI_LIST_P (x))
		  x = ksi_list2vector (x);
		else if (!KSI_VEC_P (x))
		  ksi_error ("%s: invalid list for ~? spec: %s",
			     nm, ksi_obj2str (x));

		ac = KSI_VEC_LEN (x);
		av = KSI_VEC_ARR (x);
		while (ac > 0)
		  {
		    i = ksi_internal_format (port, KSI_STR_PTR (f),
					     ac, av, nm);
		    if (i == 0)
		      break;
		    ac -= i;
		    av += i;
		  }
	      }
	      break;

	    case '%':
	      ksi_port_putc (port, '\n');
	      break;

	    case '&':
	      if (port->last_write_char != '\n')
		ksi_port_putc (port, '\n');
	      break;

	    case '_':
	      ksi_port_putc (port, ' ');
	      break;

	    case '/':
	      ksi_port_putc (port, '\t');
	      break;

	    case '|':
	      ksi_port_putc (port, '\f');
	      break;

	    case '!':
	      ksi_flush_port ((ksi_obj) port);
	      break;

	    default:
	      ksi_port_putc (port, '~');
	      ksi_port_putc (port, *fmt);
	      break;
	    }
	}
      else
	ksi_port_putc (port, *fmt);
    }

  return nums;
}


ksi_obj
ksi_format (ksi_obj port, const char* fmt, int argc, ksi_obj* argv)
{
  ksi_obj retval = ksi_void;

  if (port == ksi_true)
    port = ksi_current_output_port ();
  else if (port == ksi_false)
    {
      retval = ksi_make_string (0, 0);
      port = (ksi_obj) ksi_new_str_port (retval);
      ((ksi_port) port) -> output = 1;
    }

  TYPE_CHECK (port, KSI_OUTPUT_PORT_P (port), "format: invalid port in arg1");

  ksi_internal_format ((ksi_port) port, fmt, argc, argv, "format");
  return retval;
}


/*--------- printf --------------------------------------------------*/

struct Ksi_Spec
{
  int		conv;
  int		type;
  int		width;
  int		prec;
  int		base;
  unsigned	width_param : 1;
  unsigned	prec_param : 1;
  unsigned	minus : 1;
  unsigned	plus : 1;
  unsigned	space : 1;
  unsigned	sharp : 1;
  unsigned	zero : 1;
  unsigned	quote : 1;
  unsigned	upcase : 1;
  unsigned	efmt : 1;
  unsigned	g_flag : 1;
};

static const char*
parse_spec (const unsigned char *fmt, struct Ksi_Spec *spec)
{
  bzero (spec, sizeof *spec);

  while (*fmt)
    switch (*fmt)
      {
      case '-':  spec->minus = 1; fmt++; break;
      case '+':  spec->plus  = 1; fmt++; break;
      case ' ':  spec->space = 1; fmt++; break;
      case '#':  spec->sharp = 1; fmt++; break;
      case '0':  spec->zero  = 1; fmt++; break;
      case '\'': spec->quote = 1; fmt++; break;
      default:   goto parse_width;
      }

parse_width:
  if (*fmt == '*')
    {
      spec->width_param = 1;
      fmt++;
    }
  else
    while (isdigit (*fmt))
      {
	spec->width *= 10;
	spec->width += *fmt - '0';
	fmt++;
      }

  if (*fmt != '.')
    spec->prec = -1;
  else
    {
      fmt++;
      if (*fmt == '*')
	{
	  spec->prec_param = 1;
	  fmt++;
	}
      else
	while (isdigit (*fmt))
	  {
	    spec->prec *= 10;
	    spec->prec += *fmt - '0';
	    fmt++;
	  }
    }

  if (*fmt == 'h' || *fmt == 'l' || *fmt ==  'L' || *fmt == 'q')
    spec->type = *fmt++;

  if (*fmt)
    spec->conv = *fmt++;

  return fmt;
}

static int
printf_int (ksi_port port, const char *str, struct Ksi_Spec *spec)
{
  int len, size, sign = 0;

  if (*str == '-' || *str == '+')
    sign = *str++;
  else if (spec->plus)
    sign = '+';
  else if (spec->space)
    sign = ' ';

  if (spec->prec >= 0)
    while (*str == '0')
      str++;

  size = len = strlen (str);
  if (sign)
    size += 1;
  if (spec->sharp)
    {
      if (spec->base == 2 || spec->base == 16)
	size += 2;
      else if (spec->base == 8)
	size += 1;
    }

  spec->width -= size;
  if (spec->prec >= 0)
    {
      if ((spec->prec -= size) < 0)
	spec->prec = 0;
      spec->width -= spec->prec;

      if (!spec->minus)
	while (--spec->width >= 0)
	  ksi_port_putc (port, ' ');
    }
  else if (spec->minus)
    spec->prec = 0;
  else if (spec->zero)
    {
      spec->prec  = spec->width;
      spec->width = 0;
    }
  else
    {
      while (--spec->width >= 0)
	ksi_port_putc (port, ' ');
    }

  if (sign)
    ksi_port_putc (port, sign);

  if (spec->sharp)
    {
      if (spec->base == 2)
	ksi_port_write (port, spec->upcase ? "0B" : "0b", 2);
      else if (spec->base == 16)
	ksi_port_write (port, spec->upcase ? "0X" : "0x", 2);
      else if (spec->base == 8)
	ksi_port_putc (port, '0');
    }

  while (--spec->prec >= 0)
    ksi_port_putc (port, '0');

  while (*str)
    {
      ksi_port_putc (port, spec->upcase ? toupper (*str) : tolower (*str));
      str++;
    }

  while (--spec->width >= 0)
    ksi_port_putc (port, ' ');

  return 0;
}

static int
printf_char (ksi_port port, int c, struct Ksi_Spec *spec)
{
  if (!spec->minus)
    {
      while (--spec->width > 0)
	ksi_port_putc (port, ' ');
    }

  ksi_port_putc (port, c);

  while (--spec->width > 0)
    ksi_port_putc (port, ' ');

  return 0;
}

static int
printf_str (ksi_port port, const char *str, struct Ksi_Spec *spec)
{
  int len, width;

  if (str == 0)
    str = "(nil)";

  len = strlen (str);
  if (0 < spec->prec && spec->prec < len)
    len = spec->prec;

  width = spec->width - len;

  if (!spec->minus)
    {
      while (--width >= 0)
	ksi_port_putc (port, ' ');
    }

  ksi_port_write (port, str, len);

  while (--width >= 0)
    ksi_port_putc (port, ' ');

  return 0;
}

static int
printf_double (ksi_port port, double num, struct Ksi_Spec *spec)
{
  int i, len, wid, sign, exp, exp_sign, int_digs, fra_digs, exp_digs, has_dig;
  char *num_buf, *exp_buf;

  if (num != num)
    return printf_str (port, spec->upcase ? "NAN" : "nan", spec);
  if (num == num/2)
    {
      if (num < 0.0)
	return printf_str (port, spec->upcase ? "-INF" : "-inf", spec);
      if (num > 0.0)
	return printf_str (port, spec->upcase ? "+INF" : "+inf", spec);
    }

  if (spec->prec < 0)
    spec->prec = FLT_DIG;

  sign = 0;
  if (copysign (1.0, num) < 0.0)
    {
      sign = '-';
      num  = -num;
    }
  else if (spec->plus)
    sign = '+';
  else if (spec->space)
    sign = ' ';

  if (spec->g_flag)
    {
      if (num < 1.0e-4 || num >= pow (10.0, (double) spec->prec))
	spec->efmt = 1;
      if (spec->prec == 0)
	spec->prec = 1;
    }

  exp      = 0;
  exp_sign = '+';
  if (spec->efmt)
    {
      if (num == 0.0)
	{
	  if (spec->g_flag)
	    spec->efmt = 0;
	}
      else
	{
	  exp = (int) floor (log10 (num));
	  num /= pow (10.0, (double) exp);
	  if (exp < 0)
	    {
	      exp_sign = '-';
	      exp      = -exp;
	    }
	}
    }

  int_digs = (num > 1.0 ? 1 + (int) floor (log10 (num)) : 1);
  fra_digs = (spec->g_flag ? spec->prec - int_digs : spec->prec);
  num = floor (num * pow (10.0, (double) fra_digs) + 0.5);
  if ((int) log10 (num) + 1 > int_digs + fra_digs)
    int_digs++;

  len = int_digs + fra_digs + 4;
  num_buf = (char*) alloca (len);
  num_buf = num_buf + len;

  *--num_buf = '\0';
  has_dig = 0;
  for (i = 0; i < fra_digs; i++)
    {
      *--num_buf = (char) (((int) fmod (num, 10.0)) + '0');
      num = floor (num / 10.0);

      if (!has_dig && *num_buf == '0' && spec->g_flag)
	num_buf++;
      else
	has_dig = 1;
    }
  if (has_dig || spec->sharp)
    *--num_buf = '.';
  for (i = 0; i < int_digs; i++)
    {
      *--num_buf = (char) (((int) fmod (num, 10.0)) + '0');
      num = floor (num / 10.0);
      if (num < DBL_EPSILON)
	break;
    }

  exp_digs = 0;
  exp_buf  = 0;
  if (spec->efmt)
    {
      exp_buf = (char*) alloca (KSI_INT_DIGS);
      exp_buf += KSI_INT_DIGS;
      *--exp_buf = '\0';
      while (exp || exp_digs < 2)
	{
	  *--exp_buf = (char) (exp % 10 + '0');
	  exp /= 10;
	  exp_digs++;
	}
    }

  wid = len = strlen (num_buf);
  if (sign)
    wid += 1;
  if (exp_digs > 0)
    wid += exp_digs + 2;

  spec->width -= wid;

  if (spec->minus)
    {
      if (sign)
	ksi_port_putc (port, sign);
    }
  else if (spec->zero)
    {
      if (sign)
	ksi_port_putc (port, sign);
      while (--spec->width >= 0)
	ksi_port_putc (port, '0');
    }
  else
    {
      while (--spec->width >= 0)
	ksi_port_putc (port, ' ');
      if (sign)
	ksi_port_putc (port, sign);
    }

  ksi_port_write (port, num_buf, len);
  if (exp_digs > 0)
    {
      ksi_port_putc (port, spec->upcase ? 'E' : 'e');
      ksi_port_putc (port, exp_sign);
      ksi_port_write (port, exp_buf, exp_digs);
    }

  while (--spec->width >= 0)
    ksi_port_putc (port, ' ');

  return 0;
}

static ksi_obj
ksi_printf_int (ksi_port port, const char *fmt, int argc, ksi_obj *argv)
{
  while (*fmt)
    if (*fmt != '%')
      ksi_port_putc (port, *fmt++);
    else
      {
	struct Ksi_Spec spec;
	ksi_obj val;
	const char *str, *ptr;

	ptr = fmt;
	fmt = parse_spec (fmt+1, &spec);
	if (spec.width_param)
	  {
	    if (argc <= 0)
	      {
	      too_many:
		ksi_error ("printf: too many %% spec's in format string");
	      }

	    val = *argv++; --argc;
	    TYPE_CHECK (val, KSI_SINT_P (val), "printf: invalid integer");
	    spec.width = KSI_SINT_COD (val);
	    if (spec.width < 0)
	      {
		spec.minus = 1;
		spec.width = -spec.width;
	      }
	  }

	if (spec.prec_param)
	  {
	    if (argc <= 0)
	      goto too_many;

	    val = *argv++; --argc;
	    TYPE_CHECK (val, KSI_SINT_P (val), "printf: invalid integer");
	    spec.prec = KSI_SINT_COD (val);
	  }

	switch (spec.conv)
	  {
	  case '%':
	    ksi_port_putc (port, '%');
	    break;

	  case 'd': case 'i': case 'u':
	    spec.base = 10;
	  print_int:
	    if (argc <= 0)
	      goto too_many;
	    val = *argv++; --argc;
	    if (KSI_CHAR_P (val))
	      val = KSI_MK_SINT (KSI_CHAR_COD (val));
	    TYPE_CHECK (val, KSI_EINT_P (val), "printf: invalid integer");
	    str = ksi_num2str (val, spec.base);
	    printf_int (port, str, &spec);
	    break;

	  case 'B':
	    spec.upcase = 1;
	  case 'b':
	    spec.base = 2;
	    goto print_int;

	  case 'o':
	    spec.base = 8;
	    goto print_int;

	  case 'X':
	    spec.upcase = 1;
	  case 'x':
	    spec.base = 16;
	    goto print_int;

	  case 'p':
	    if (argc <= 0)
	      goto too_many;
	    val = *argv++; --argc;
	    spec.base  = 16;
	    spec.sharp = 1;
	    spec.plus  = 0;
	    spec.space = 0;
	    str = ksi_num2str (ksi_uint2num (KSI_COD (val)), spec.base);
	    printf_int (port, str, &spec);
	    break;

	  case 'c':
	    if (argc <= 0)
	      goto too_many;
	    val = *argv++; --argc;
	    TYPE_CHECK (val, KSI_CHAR_P (val), "printf: invalid char");
	    printf_char (port, KSI_CHAR_COD (val), &spec);
	    break;

	  case 's':
	    if (argc <= 0)
	      goto too_many;
	    val = *argv++; --argc;
	    TYPE_CHECK (val, KSI_STR_P (val), "printf: invalid string");
	    if (spec.quote)
	      str = ksi_string2str (val);
	    else
	      str = KSI_STR_PTR (val);
	    printf_str (port, str, &spec);
	    break;

	  case 'm':
	    printf_str (port, strerror (errno), &spec);
	    break;

	  case 'E':
	    spec.upcase = 1;
	  case 'e':
	    spec.efmt   = 1;
	    goto print_float;

	  case 'G':
	    spec.upcase = 1;
	  case 'g':
	    spec.g_flag = 1;
	    goto print_float;

	  case 'f':
	  print_float:
	    if (argc <= 0)
	      goto too_many;
	    val = *argv++; --argc;
	    TYPE_CHECK (val, KSI_REAL_P (val), "printf: invalid real");
	    printf_double (port, ksi_num2double (val), &spec);
	    break;

	  case 'w':
	    if (argc <= 0)
	      goto too_many;
	    val = *argv++; --argc;
	    if (spec.prec < 0 && spec.width <= 0)
	      ksi_write (val, (ksi_obj) port);
	    else
	      printf_str (port, ksi_obj2str (val), &spec);
	    break;

	  case 'y':
	    if (argc <= 0)
	      goto too_many;
	    val = *argv++; --argc;
	    if (spec.prec < 0 && spec.width <= 0)
	      ksi_display (val, (ksi_obj) port);
	    else
	      printf_str (port, ksi_aprintf ("%y", val), &spec);
	    break;

	  default:
	    while (ptr < fmt)
	      ksi_port_putc (port, *ptr++);
	  }
      }

  return ksi_void;
}

ksi_obj
ksi_vprintf (ksi_obj port, const char* fmt, va_list args)
{
  ksi_obj res = ksi_void;

  if (!port || port == ksi_true)
    port = ksi_current_output_port ();
  else if (port == ksi_false)
    {
      res = ksi_make_string (0, 0);
      port = (ksi_obj) ksi_new_str_port (res);
      ((ksi_port) port) -> output = 1;
      ksi_vprintf (port, fmt, args);
      return res;
    }
  else if (KSI_STR_P (port))
    {
      port = ksi_open_fd_port_int (KSI_STR_PTR (port), "a", "ksi_vprintf");
      res = ksi_vprintf (port, fmt, args);
      ksi_close_port (port);
      return res;
    }

  if (!port || !KSI_OUTPUT_PORT_P (port))
    return ksi_void;

  while (*fmt)
    if (*fmt != '%')
      ksi_port_putc ((ksi_port) port, *fmt++);
    else
      {
	struct Ksi_Spec spec;
	const char *str, *ptr;
	ksi_obj val;
	union
	{
#ifdef LONG_LONG
	  LONG_LONG           L;
#endif
	  long                l;
	  int                 i;
	  void                *p;
	  double              d;
	} num;

	ptr = fmt;
	fmt = parse_spec (fmt+1, &spec);
	if (spec.width_param)
	  {
	    spec.width = va_arg (args, int);
	    if (spec.width < 0)
	      {
		spec.minus = 1;
		spec.width = -spec.width;
	      }
	  }

	if (spec.prec_param)
	  {
	    spec.prec = va_arg (args, int);
	  }

	switch (spec.conv)
	  {
	  case '%':
	    ksi_port_putc ((ksi_port) port, '%');
	    break;

	  case 'd': case 'i':
	    spec.base = 10;
#ifdef LONG_LONG
	    if (spec.type == 'L' || spec.type == 'q')
	      {
		num.L = va_arg (args, LONG_LONG);
		val   = ksi_long2num ((LONG_LONG) num.L);
	      }
	    else if (spec.type == 'l')
#else
	    if (spec.type == 'l' || spec.type == 'L' || spec.type == 'q')
#endif
	      {
		num.l = va_arg (args, long);
		val   = ksi_int2num ((long) num.l);
	      }
	    else if (spec.type == 'h')
	      {
		num.i = va_arg (args, int);
		val   = ksi_int2num ((short) num.i);
	      }
	    else
	      {
		num.i = va_arg (args, int);
		val   = ksi_int2num ((int) num.i);
	      }

	    str = ksi_num2str (val, spec.base);
	    printf_int ((ksi_port) port, str, &spec);
	    break;

	  case 'u':
	    spec.base = 10;
	  print_uint:
#ifdef LONG_LONG
	    if (spec.type == 'L' || spec.type == 'q')
	      {
		num.L = va_arg (args, LONG_LONG);
		val   = ksi_ulong2num ((unsigned LONG_LONG) num.L);
	      }
	    else if (spec.type == 'l')
#else
	    if (spec.type == 'l' || spec.type == 'L' || spec.type == 'q')
#endif
	      {
		num.l = va_arg (args, long);
		val   = ksi_uint2num ((unsigned long) num.l);
	      }
	    else if (spec.type == 'h')
	      {
		num.i = va_arg (args, int);
		val   = ksi_uint2num ((unsigned short) num.i);
	      }
	    else
	      {
		num.i = va_arg (args, int);
		val   = ksi_uint2num ((unsigned int) num.i);
	      }

	    str = ksi_num2str (val, spec.base);
	    printf_int ((ksi_port) port, str, &spec);
	    break;

	  case 'B':
	    spec.upcase = 1;
	  case 'b':
	    spec.base = 2;
	    goto print_uint;

	  case 'o':
	    spec.base = 8;
	    goto print_uint;

	  case 'X':
	    spec.upcase = 1;
	  case 'x':
	    spec.base = 16;
	    goto print_uint;

	  case 'p':
	    spec.base  = 16;
	    spec.sharp = 1;
	    spec.plus  = 0;
	    spec.space = 0;
	    num.p = va_arg (args, void*);
	    str = ksi_num2str (ksi_uint2num ((unsigned) num.p), spec.base);
	    printf_int ((ksi_port) port, str, &spec);
	    break;

	  case 'c':
	    num.i = va_arg (args, int);
	    printf_char ((ksi_port) port, num.i, &spec);
	    break;

	  case 's':
	    num.p = va_arg (args, char*);
	    if (spec.quote)
	      str = ksi_string2str (ksi_ptr02string ((char*) num.p));
	    else
	      str = (char*) num.p;
	    printf_str ((ksi_port) port, str, &spec);
	    break;

	  case 'm':
	    printf_str ((ksi_port) port, strerror (errno), &spec);
	    break;

	  case 'E':
	    spec.upcase = 1;
	  case 'e':
	    spec.efmt   = 1;
	    goto print_float;

	  case 'G':
	    spec.upcase = 1;
	  case 'g':
	    spec.g_flag = 1;
	    goto print_float;

	  case 'f':
	  print_float:
	    if (spec.type == 'l' || spec.type == 'L')
	      num.d = (double) va_arg (args, long double);
	    else
	      num.d = va_arg (args, double);
	    printf_double ((ksi_port) port, num.d, &spec);
	    break;

	  case 'w':
	    val = va_arg (args, ksi_obj);
	    if (spec.prec < 0 && spec.width <= 0)
	      ksi_write (val, port);
	    else
	      printf_str ((ksi_port) port, ksi_obj2str (val), &spec);
	    break;

	  case 'y':
	    val = va_arg (args, ksi_obj);
	    if (spec.prec < 0 && spec.width <= 0)
	      ksi_display (val, port);
	    else
	      printf_str ((ksi_port) port, ksi_aprintf ("%y", val), &spec);
	    break;

	  default:
	    while (ptr < fmt)
	      ksi_port_putc ((ksi_port) port, *ptr++);
	  }
      }

  return ksi_void;
}

ksi_obj
ksi_printf (ksi_obj port, const char* fmt, ...)
{
  ksi_obj val;
  va_list args;

  va_start (args, fmt);
  val = ksi_vprintf (port, fmt, args);
  va_end (args);

  return val;
}

char*
ksi_aprintf (const char* fmt, ...)
{
  char *ptr;
  ksi_obj val;
  va_list args;

  va_start (args, fmt);
  val = ksi_vprintf (ksi_false, fmt, args);
  va_end (args);

  ptr = (char*) ksi_malloc_data (KSI_STR_LEN (val) + 1);
  strcpy (ptr, KSI_STR_PTR (val));
  return ptr;
}


static ksi_obj
format_proc (ksi_obj port, ksi_obj str, int argc, ksi_obj* argv)
{
  TYPE_CHECK (str, KSI_STR_P (str), "format: invalid string in arg2");
  return ksi_format (port, KSI_STR_PTR (str), argc, argv);
}

static ksi_obj
error_proc (int argc, ksi_obj* argv)
{
  ksi_obj str;
  ksi_port port;
  int i;

  if (argc <= 0)
    ksi_error ("unspecified error");

  str  = ksi_make_string (0, 0);
  port = ksi_new_str_port (str);
  port->output = 1;

  if (argc > 0 && KSI_SYM_P (argv[0]))
    {
      ksi_write (argv[0], (ksi_obj) port);
      ksi_port_write (port, ": ", 2);
      --argc; ++argv;
    }

  if (argc > 0 && KSI_STR_P (argv[0]))
    {
      i = ksi_internal_format (port, KSI_STR_PTR (argv[0]),
			       argc-1, argv+1, "error");
      argc -= i + 1;
      argv += i + 1;
    }

  for (i = 0; i < argc; i++)
    {
      ksi_port_putc (port, ' ');
      ksi_write (argv[i], (ksi_obj) port);
    }

  ksi_error (KSI_STR_PTR (str));
  return ksi_void;
}

static ksi_obj
printf_proc (ksi_obj port, ksi_obj str, int argc, ksi_obj* argv)
{
  ksi_obj res = ksi_void;

  if (port == ksi_true)
    port = ksi_current_output_port ();
  else if (port == ksi_false)
    {
      res  = ksi_make_string (0, 0);
      port = (ksi_obj) ksi_new_str_port (res);
      ((ksi_port) port) -> output = 1;
    }

  TYPE_CHECK (port, KSI_OUTPUT_PORT_P (port), "printf: invalid port in arg1");
  TYPE_CHECK (str, KSI_STR_P (str), "printf: invalid string in arg2");

  ksi_printf_int ((ksi_port) port, KSI_STR_PTR (str), argc, argv);
  return res;
}

static ksi_obj
gc_expand_heap (ksi_obj x)
{
  TYPE_CHECK (x, KSI_SINT_P (x) && KSI_SINT_COD (x) > 0,
	      "ksi:gc-expand: invalid integer");

  return ksi_expand_heap (KSI_SINT_COD (x)) ? ksi_true : ksi_false;
}

static ksi_obj
gc_set_heap_size (ksi_obj x)
{
  TYPE_CHECK (x, KSI_SINT_P (x) && KSI_SINT_COD (x) > 0,
	      "ksi:gc-set-size!: invalid integer");

  ksi_set_max_heap (KSI_SINT_COD (x));
  return ksi_void;
}

static ksi_obj
gc_heap_size ()
{
  return ksi_uint2num (ksi_heap_size ());
}

static ksi_obj
gc_free_size (void)
{
  return ksi_uint2num (ksi_heap_free ());
}

static ksi_obj
gc_collections (void)
{
  return ksi_uint2num (ksi_gcollections ());
}

static ksi_obj
gc_collect ()
{
  ksi_gcollect (1);
  return ksi_void;
}

ksi_obj
ksi_abbrev (char* str, int len)
{
  if (!KSI || !KSI_INITED)
    return ksi_nil;

  if (str[0] == '#' && str[1] == '\\')
    {
      int i;
      ksi_obj res = ksi_nil;

      for (i = 0; ksi_char_names[i]; i++)
	{
	  const char* name = ksi_char_names[i];
	  if (memcmp (name, str, len) == 0)
	    res = CONS (ksi_str2char (name, strlen (name)), res);
	}

      return res;
    }

  if (str[0] == '#' && str[1] == ':')
    return ksi_abbrev_sym (KSI->keyarray, str+2, len-2);

  return ksi_abbrev_sym (KSI->obarray, str, len);
}

static ksi_obj
abbrev (ksi_obj str)
{
  TYPE_CHECK (str, KSI_STR_P (str), "abbrev: invalid string in arg1");

  return ksi_abbrev (KSI_STR_PTR (str), KSI_STR_LEN (str));
}



static struct Ksi_Prim_Def defs [] =
{
  { "@get-size",	ksi_get_size,		KSI_CALL_ARG1, 1 },
  { "@peek-byte",	ksi_peek_byte,		KSI_CALL_ARG2, 2 },
  { "@peek-int",	ksi_peek_int,		KSI_CALL_ARG2, 2 },
  { "@peek-double",	ksi_peek_double,	KSI_CALL_ARG2, 2 },
  { "@poke-byte",	ksi_poke_byte,		KSI_CALL_ARG3, 3 },
  { "@poke-int",	ksi_poke_int,		KSI_CALL_ARG3, 3 },
  { "@poke-double",	ksi_poke_double,	KSI_CALL_ARG3, 3 },

  { "ksi:gc-expand",	gc_expand_heap,		KSI_CALL_ARG1, 1 },
  { "ksi:gc-set-size!",	gc_set_heap_size,	KSI_CALL_ARG1, 1 },
  { "ksi:gc-size",	gc_heap_size,		KSI_CALL_ARG0, 0 },
  { "ksi:gc-free",	gc_free_size,		KSI_CALL_ARG0, 0 },
  { "ksi:gcollections",	gc_collections,		KSI_CALL_ARG0, 0 },
  { "ksi:gcollect",	gc_collect,		KSI_CALL_ARG0, 0 },

  { "format",		format_proc,		KSI_CALL_REST2, 2 },
  { "error",		error_proc,		KSI_CALL_REST0, 0 },
  { "printf",		printf_proc,		KSI_CALL_REST2, 2 },

  { "abbrev",		abbrev,			KSI_CALL_ARG1, 1 },

  { 0 }
};


void
ksi_init_misc (void)
{
  ksi_ignore (rcsid);
  ksi_reg_unit (defs);
}


 /* End of code */
