/*
 * ksi_mod.c
 * ksi modules
 *
 * Copyright (C) 1998-2000,2001, Ivan Demakov.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written
 * agreement is hereby granted, provided that the above copyright notice
 * and this paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Mon Apr 13 16:35:45 1998
 * Last Update:   Wed Mar 28 04:56:24 2001
 *
 */

/* Comments:
 *
 */

#include "ksi_int.h"
#include "ksi_vtab.h"
#include "ksi_mod.h"
#include "ksi_comp.h"


static char rcsid[] = "$Id: ksi_mod.c,v 1.6 2001/04/02 19:49:05 demakov Exp $";


static ksi_obj
ksi_new_module (ksi_obj name, ksi_obj env, ksi_obj export, ksi_obj opts)
{
  ksi_module mod;
  mod = NEW_TYPED (struct Ksi_Module, env, opts);
  mod -> itag = KSI_TAG_MODULE;
  mod -> env  = env;
  mod -> name = name;
  mod -> export = export;
  mod -> uses = ksi_nil;
  mod -> opts = opts;

  if (env)
    ENV_MOD (env) = (ksi_obj) mod;

  return (ksi_obj) mod;
}


ksi_obj
ksi_module_p (ksi_obj x)
{
  return MOD_P (x) ? ksi_true : ksi_false;
}


ksi_obj
ksi_make_module (ksi_obj name, ksi_obj export, ksi_obj opts, ksi_obj env)
{
  ksi_obj x, mod;

  if (name != ksi_false)
    {
      TYPE_CHECK (name, SYM_P (name), "@make-module: invalid modules name");
      KSI_EXN_CHECK ("module", name,
		     ksi_assq_ref (KSI->modules, name) == ksi_false,
		     "@make-module: module already exists");
    }

  if (export != ksi_true)
    for (x = export; x != ksi_nil; x = CDR (x))
      {
	TYPE_CHECK (export, PAIR_P (x) && SYM_P (CAR (x)),
		    "@make-module: invalid module export");
      }

  if (env)
    TYPE_CHECK (env, ENV_P (env), "@make-module: invalid enveronment");
  else
    env = (ksi_obj) ksi_new_env (0);

  mod = ksi_new_module (name, env, export, opts ? opts : ksi_nil);

  if (name != ksi_false)
    KSI->modules = ksi_assq_set_x (KSI->modules, name, mod);

  return mod;
}


ksi_obj
ksi_find_module (ksi_obj name)
{
  TYPE_CHECK (name, SYM_P (name), "@find-module: invalid module name");
  return ksi_assq_ref (KSI->modules, name);
}


ksi_obj
ksi_clear_module (ksi_obj mod)
{
  TYPE_CHECK (mod, MOD_P (mod), "@clear-module: invalid module in arg1");
  ksi_clear_vtab (((ksi_env) (MOD_ENV (mod))) -> valtab);
  MOD_USES (mod) = ksi_nil;
  return ksi_void;
}


ksi_obj
ksi_current_module ()
{
  return KSI->curr_module;
}


ksi_obj
ksi_set_current_module (ksi_obj mod)
{
  ksi_obj old_mod;

  TYPE_CHECK (mod, mod == ksi_false || MOD_P (mod),
	      "@set-current-module: invalid module in arg1");

  old_mod = KSI->curr_module;
  if (old_mod != mod)
    KSI->curr_module = mod;

  return old_mod;
}


static ksi_obj
ksi_module_env (ksi_obj mod)
{
  TYPE_CHECK (mod, MOD_P (mod), "@module-environment: invalid module in arg1");
  return MOD_ENV (mod);
}


static ksi_obj
ksi_module_name (ksi_obj mod)
{
  TYPE_CHECK (mod, MOD_P (mod), "@module-name: invalid module in arg1");
  return MOD_NAME (mod);
}


static ksi_obj
ksi_module_export (ksi_obj mod)
{
  TYPE_CHECK (mod, MOD_P (mod), "@module-export: invalid module in arg1");
  return MOD_EXPORT (mod);
}

static ksi_obj
ksi_set_module_export (ksi_obj mod, ksi_obj export)
{
  ksi_obj x;

  TYPE_CHECK (mod, MOD_P (mod), "@set-module-export!: invalid module in arg1");

  if (export != ksi_true)
    for (x = export; x != ksi_nil; x = CDR (x))
      {
	TYPE_CHECK (export, PAIR_P (x) && SYM_P (CAR (x)),
		    "@set-module-export!: invalid module export");
      }

  MOD_EXPORT (mod) = export;
  return ksi_void;
}


static ksi_obj
ksi_module_uses (ksi_obj mod)
{
  TYPE_CHECK (mod, MOD_P (mod), "@module-uses: invalid module in arg1");
  return MOD_USES (mod);
}


static ksi_obj
ksi_add_module_use (ksi_obj mod, ksi_obj use, ksi_obj unused)
{
  ksi_obj x;

  TYPE_CHECK (mod, MOD_P (mod), "@add-module-use: invalid module in arg1");
  TYPE_CHECK (use, MOD_P (use), "@add-module-use: invalid module in arg2");

  for (x = MOD_USES (mod); x != ksi_nil; x = CDR (x))
    {
      if (CAR (x) == use)
	return ksi_void;
    }

  MOD_USES (mod) = CONS (use, MOD_USES (mod));

  return ksi_void;
}


static ksi_obj
ksi_module_opts (ksi_obj mod)
{
  TYPE_CHECK (mod, MOD_P (mod), "@module-options: invalid module in arg1");
  return MOD_OPTS (mod);
}


ksi_obj
ksi_set_module_opt (ksi_obj mod, ksi_obj key, ksi_obj val)
{
  TYPE_CHECK (mod, MOD_P (mod), "@set-module-option!: invalid module in arg1");
  MOD_OPTS (mod) = ksi_assq_set_x (MOD_OPTS (mod), key, val);
  return ksi_void;
}


ksi_obj
ksi_get_module_opt (ksi_obj mod, ksi_obj key)
{
  ksi_obj x;

  TYPE_CHECK (mod, MOD_P (mod), "@get-module-option: invalid module in arg1");
  x = ksi_assq (key, MOD_OPTS (mod));
  return (x == ksi_false ? ksi_void : CDR (x));
}


static struct Ksi_Prim_Def defs [] =
{
  { "@current-module",		ksi_current_module,	KSI_CALL_ARG0, 0 },
  { "@set-current-module!",	ksi_set_current_module,	KSI_CALL_ARG1, 1 },
  { "@module-environment",	ksi_module_env,		KSI_CALL_ARG1, 1 },
  { "@module-name",		ksi_module_name,	KSI_CALL_ARG1, 1 },
  { "@module-export",		ksi_module_export,	KSI_CALL_ARG1, 1 },
  { "@set-module-export!",	ksi_set_module_export,	KSI_CALL_ARG2, 2 },
  { "@module-uses",		ksi_module_uses,	KSI_CALL_ARG1, 1 },
  { "@add-module-use",		ksi_add_module_use,	KSI_CALL_ARG3, 2 },
  { "@make-module",		ksi_make_module,	KSI_CALL_ARG4, 2 },
  { "@find-module",		ksi_find_module,	KSI_CALL_ARG1, 1 },
  { "@module?",			ksi_module_p,		KSI_CALL_ARG1, 1 },
  { "@module-options",		ksi_module_opts,	KSI_CALL_ARG1, 1 },
  { "@get-module-option",	ksi_get_module_opt,	KSI_CALL_ARG2, 2 },
  { "@set-module-option!",	ksi_set_module_opt,	KSI_CALL_ARG3, 3 },
  { "@clear-module",		ksi_clear_module,	KSI_CALL_ARG1, 1 },

  { 0 }
};

void
ksi_init_modules ()
{
  ksi_ignore (rcsid);
  ksi_reg_unit (defs);
}


 /* End of code */
