/*
 * ksi_read.c
 * reader
 *
 * Copyright (C) 1997-2000,2001, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Wed Aug  6 19:36:40 1997
 * Last Update:   Sat Nov 10 00:25:48 2001
 *
 */

/* Comments:
 *
 */

#include "ksi_int.h"
#include "ksi_comp.h"

static char rcsid[] = "$Id: ksi_read.c,v 1.5 2001/11/25 14:30:01 demakov Exp $";


#define DELIMITERS  "()[]{}'`,;\""

typedef struct
{
  ksi_port	port;
  ksi_obj	sharp_hook;
  char		*tok_buf;
  unsigned	tok_size, tok_len;
  unsigned	tok_pos, tok_line;
  unsigned	kw_mode_prefix : 1;
  unsigned	kw_mode_postfix : 1;
  unsigned	case_sens : 1;
  unsigned	annotate : 1;
  unsigned	make_const : 1;
} read_state;


static ksi_obj ksi_read_obj (read_state* st);


static char*
port_name (ksi_port p)
{
  char *ptr;

  if (p->is_fd)
    ptr = ((struct Ksi_FdPort*) p) -> name;
  else
    ptr = (char*) ksi_obj2str ((ksi_obj) p);

  return ptr;
}

static ksi_obj
unannotate (ksi_obj x)
{
  if (PAIR_P (x))
    return CONS (unannotate (CAR (x)), unannotate (CDR (x)));

  if (KSI_VEC_P (x))
    {
      int i;
      ksi_obj v = ksi_alloc_vector (KSI_VEC_LEN (x), KSI_TAG_VECTOR);
      for (i = 0; i < KSI_VEC_LEN (x); i++)
	KSI_VEC_REF (v, i) = unannotate (KSI_VEC_REF (x, i));
      return v;
    }

  if (KSI_OBJ_IS (x, KSI_TAG_ANNOTATION))
    return KSI_VEC_REF (x, 0);

  return x;
}

static ksi_obj
annotation (ksi_obj x, read_state *st)
{
  if (st->annotate)
    {
      ksi_obj a = ksi_alloc_vector (3, KSI_TAG_ANNOTATION);
      KSI_VEC_REF (a, 0) = unannotate (x);
      KSI_VEC_REF (a, 1) = x;
      KSI_VEC_REF (a, 2) = KSI_LIST5 (ksi_str02string (port_name (st->port)),
				      ksi_int2num (st->tok_line),
				      ksi_int2num (st->tok_pos),
				      ksi_int2num (st->port->read_line),
				      ksi_int2num (st->port->read_pos));
      return a;
    }

  return x;
}


static void
expand_tok_buf (read_state *st, int sz)
{
  char *p;

  if (sz >= st->tok_size)
    {
      sz = ((sz + 127) & ~127) + 128;
      p = (char*) ksi_malloc_data (sz);
      if (st->tok_len)
	memcpy (p, st->tok_buf, st->tok_len);

      st->tok_size = sz;
      st->tok_buf  = p;
    }
}


static inline int
read_next (read_state *st)
{
  return ksi_port_getc (st->port);
}

static inline int
peek_next (read_state *st)
{
  if (st->port->unread_num)
    return (unsigned char) st->port->unread_chars [st->port->unread_num - 1];
  else
    {
      int c = read_next (st);
      if (c >= 0)
	ksi_port_ungetc (st->port, c);
      return c;
    }
}

static int
skip_ws (read_state *st, int store_pos)
{
  int c, pc, line, pos, depth;

  for (;;)
    {
      switch (c = read_next (st))
	{
	case -1:
	  return -1;

	case ';':
	  while ((c = read_next (st)) != '\n')
	    if (c == -1)
	      return -1;
	  break;

	case '#':
	  line = st->port->read_line + 1;
	  pos  = st->port->read_pos;
	  c = peek_next (st);
	  if (c == '!') /* skip balanced comment, i.e.  #! comment !# */
	    {
	      depth = 0;
	      pc = '#';
	      for (;;)
		{
		  c = read_next (st);
		  if (c == -1)
		    ksi_error ("read: %s: not found matching `!#' "
			       "for `#!' at %d:%d",
			       port_name (st->port), line, pos);

		  if (pc == '#' && c == '!')
		    {
		      ++depth;
		      c = read_next (st);
		    }
		  else if (pc == '!' && c == '#')
		    {
		      if (--depth == 0)
			break;
		      c = read_next (st);
		    }
		  pc = c;
		} /* for (;;) */

	      break; /* case '#' */
	    } /* if (c == '!') */

	  if (store_pos)
	    {
	      st->tok_line = line;
	      st->tok_pos  = pos;
	    }
	  return '#';

	default:
	  if (isspace (c))
	    break;

	  line = st->port->read_line + 1;
	  pos  = st->port->read_pos;
	  if (c < ' ')
	    ksi_error ("read: %s: illegal char 0x%x at %d:%d",
		       port_name (st->port), c, line, pos);

	  if (store_pos)
	    {
	      st->tok_line = line;
	      st->tok_pos  = pos;
	    }
	  return c;
	}

    } /* for (;;) */
}


static char*
read_c_string (read_state *st, int end_char, char* read_token)
{
  st->tok_len = 0;
  for (;;)
    {
      int k, n;
      int c = read_next (st);
      if (c == -1)
	{
	eof_err:
	  ksi_error ("read: %s: eof while reading a %s at %d:%d",
		     port_name (st->port), read_token,
		     st->tok_line, st->tok_pos);
	}

    next_char:
      expand_tok_buf (st, st->tok_len);

      if (c == end_char)
        break;

      if (c == '\\')
        {
          if ((c = read_next (st)) == -1)
	    goto eof_err;

          switch (c)
            {
            case 'e' : c = 0x1b; break;     /* esc  */
            case 'b' : c = '\b'; break;     /* bs   */
            case 'n' : c = '\n'; break;     /* lf   */
            case 'r' : c = '\r'; break;     /* cr   */
            case 't' : c = '\t'; break;     /* tab  */
            case 'f' : c = '\f'; break;     /* ff   */
	    case 'a' : c = '\a'; break;	    /* bel  */
	    case 'v' : c = '\v'; break;	    /* vtab */

            case '0': case '1': case '2': case '3':
	    case '4': case '5': case '6': case '7':
              for (k=n=0; ; k++)
                {
                  if (isdigit(c) && (c < '8') && k < 3) /* Max = 3 digits */
                    n = n * 8 + c - '0';
                  else
                    {
                      st->tok_buf [st->tok_len++] = n;
                      goto next_char;
                    }

                  if ((c = read_next (st)) == -1)
		    goto eof_err;
                }
	      /* never reached */
	      ksi_error ("read_c_string: intergal error");

            case 'x' : case 'X':
              for (k=n=0; ; k++)
                {
                  if ((c = read_next (st)) == -1)
		    goto eof_err;

                  if (isxdigit(c) && k < 2) /* Max = 2 digits */
                    {
                      if (isdigit(c))
                        c = c - '0';
                      else if (isupper(c))
                        c = c - 'A' + 0xA;
                      else
                        c = c - 'a' + 0xa;
                      n = n * 16 + c;
                    }
                  else
                    {
                      st->tok_buf[st->tok_len++] = n;
                      goto next_char;
                    }
                }
	      /* never reached */
	      ksi_error ("read_c_string: intergal error");

            } /* switch */

        } /* if (c == '\\') */

      st->tok_buf [st->tok_len++] = c;
    } /* for */

  st->tok_buf [st->tok_len] = '\0';
  return st->tok_buf;
}

static char*
read_token (read_state *st, int c, int lower)
{
  if (c == -1) c = read_next (st);

  st->tok_len = 0;
  if (c != -1)
    for (;;)
      {
	expand_tok_buf (st, st->tok_len);
	st->tok_buf[st->tok_len++] = (lower ? tolower (c) : c);

	c = peek_next (st);
	if (c == -1 || strchr (DELIMITERS, c))
	  break;

	c = read_next (st);
	if (isspace (c))
	  break;
      }

  expand_tok_buf (st, st->tok_len);
  st->tok_buf [st->tok_len] = 0;

  return st->tok_buf;
}


static ksi_obj
read_list (read_state *st, int close_char, char* read_item)
{
  ksi_obj  list = ksi_nil, tmp;
  ksi_obj* last = &list;

  for (;;)
    {
      int c = skip_ws (st, 0);
      if (c == -1)
	{
	eof_err:
	  ksi_error ("read: %s: eof while reading a %s at %d:%d",
		     port_name (st->port), read_item,
		     st->tok_line, st->tok_pos);
	}

      if (c == close_char)
	return annotation (list, st);

      if (c == '.')
	{
	  int c1 = read_next (st);
	  if (c1 == -1)
	    goto eof_err;

	  if (isspace (c1) || strchr (DELIMITERS, c1))
	    {
	      ksi_port_ungetc (st->port, c1);
	      *last = ksi_read_obj (st);
	      if (*last == ksi_eof)
		goto eof_err;

	      c = skip_ws (st, 0);
	      if (c != close_char)
		ksi_error ("read: %s: missing `%c' while reading %s at %d:%d",
			   port_name (st->port), close_char, read_item,
			   st->tok_line, st->tok_pos);

	      return annotation (list, st);
	    }

	  ksi_port_ungetc (st->port, c1);
        }

      ksi_port_ungetc (st->port, c);
      tmp = ksi_read_obj (st);
      if (tmp == ksi_eof)
	goto eof_err;

      *last = CONS (tmp, ksi_nil);
      if (st->make_const)
	(*last)->itag = KSI_TAG_CONST_PAIR;

      last = & CDR (*last);
    }

  ksi_error ("read: internal error");
  return 0;
}


static ksi_obj
read_ch (read_state *st)
{
  ksi_obj x;

  read_token (st, -1, 0);
  if (st->tok_len == 1)
    return KSI_MK_CHAR (st->tok_buf[0]);

  x = ksi_str2char (st->tok_buf, st->tok_len);
  if (x != ksi_false)
    return x;

  ksi_error ("read: %s: unknown character name `#\\%s' at %d:%d",
	     port_name (st->port), st->tok_buf,
	     st->tok_line, st->tok_pos);
  return 0;
}

static ksi_obj
read_vector (read_state *st, int c)
{
  ksi_obj x = read_list (st, c, "vector");
  ksi_obj v = (st->annotate ? KSI_VEC_REF (x, 0) : x);

#if CAUTIOUS >= 2
  if (!KSI_LIST_P (v))
    ksi_error ("read: %s: malformed vector #%s at %d:%d",
	       port_name (st->port), ksi_obj2str (v),
	       st->tok_line, st->tok_pos);
#endif

  v = ksi_list2vector (v);
  if (st->make_const)
    v->itag = KSI_TAG_CONST_VECTOR;
  if (st->annotate)
    KSI_VEC_REF (x, 0) = v;
  else
    x = v;

  return x;
}

static int
token_p (char *tok, char *templ)
{
  while (1)
    {
      if (*templ == '\0')
	return (*tok == '\0');
      if (*tok++ != *templ++)
	return 0;
    }
}

static ksi_obj
read_sharp (read_state *st)
{
  ksi_obj o;
  int c;

  c = read_next (st);
  switch (c)
    {
    case '(':
      return read_vector (st, ')');

    case '[':
      return read_vector (st, ']');

    case '\\':
      return annotation (read_ch (st), st);

    case ':':
      c = read_next (st);
      if (c == '|')
	{
	  read_c_string (st, '|', "keyword");
	  o = ksi_str2key (st->tok_buf, st->tok_len);
	}
      else
	{
	  read_token (st, c, !st->case_sens);
	  o = ksi_str2key (st->tok_buf, st->tok_len);
	}
      return annotation (o, st);

    case '|':
      read_c_string (st, '|', "symbol");
      o = ksi_intern (st->tok_buf, st->tok_len);
      return annotation (o, st);

    case '"':
      /* bigloo string */
      read_c_string (st, '"', "string");
      o = ksi_str2string (st->tok_buf, st->tok_len);
      if (st->make_const)
	o->itag = KSI_TAG_CONST_STRING;
      return annotation (o, st);

    case 'a': case 'A':
      /* bigloo char */
      read_token (st, 'a', 0);
      o = ksi_str02num (st->tok_buf + 1, 10);
      if (o != ksi_false)
	return annotation (ksi_integer2char (o), st);
      break;

    case 'f': case 'F':
    case 't': case 'T':
      read_token (st, tolower (c), 1);
      if (st->tok_len == 1)
	{
	  o = st->tok_buf[0] == 't' ? ksi_true : ksi_false;
	  return annotation (o, st);
	}
      break;

    case 'b': case 'B': case 'o': case 'O':
    case 'd': case 'D': case 'x': case 'X':
    case 'i': case 'I': case 'e': case 'E':
      ksi_port_ungetc (st->port, c);
      read_token (st, '#', 0);
      o = ksi_str02num (st->tok_buf, 0);
      if (o != ksi_false)
	return annotation (o, st);
      memmove (st->tok_buf, st->tok_buf + 1, st->tok_len);
      break;

    default:
      if (isalnum (c))
	read_token (st, c, 1);
      else
	{
	  expand_tok_buf (st, 2);
	  st->tok_buf [0] = c;
	  st->tok_buf [1] = '\0';
	  st->tok_len = 1;
	}
    }

  if (PROC_P (st->sharp_hook))
    {
      o = ksi_apply_2 (st->sharp_hook,
		       ksi_str2string (st->tok_buf, st->tok_len),
		       (ksi_obj) st->port);
      if (o != ksi_void)
	return annotation (o, st);
    } 

  if (token_p (st->tok_buf, "void"))
    return ksi_void;

  if (token_p (st->tok_buf, "error"))
    return ksi_err;

  if (token_p (st->tok_buf, "cons"))
    return ksi_glob_val (KSI->sym_cons, 0);

  if (token_p (st->tok_buf, "car"))
    return ksi_glob_val (KSI->sym_car, 0);

  if (token_p (st->tok_buf, "cdr"))
    return ksi_glob_val (KSI->sym_cdr, 0);

  if (token_p (st->tok_buf, "not"))
    return ksi_glob_val (KSI->sym_not, 0);

  if (token_p (st->tok_buf, "eq?"))
    return ksi_glob_val (KSI->sym_eq_p, 0);

  if (token_p (st->tok_buf, "eqv?"))
    return ksi_glob_val (KSI->sym_eqv_p, 0);

  if (token_p (st->tok_buf, "equal?"))
    return ksi_glob_val (KSI->sym_equal_p, 0);

  if (token_p (st->tok_buf, "memq"))
    return ksi_glob_val (KSI->sym_memq, 0);

  if (token_p (st->tok_buf, "memv"))
    return ksi_glob_val (KSI->sym_memv, 0);

  if (token_p (st->tok_buf, "member"))
    return ksi_glob_val (KSI->sym_member, 0);

  if (token_p (st->tok_buf, "null?"))
    return ksi_glob_val (KSI->sym_null_p, 0);

  if (token_p (st->tok_buf, "pair?"))
    return ksi_glob_val (KSI->sym_pair_p, 0);

  if (token_p (st->tok_buf, "list?"))
    return ksi_glob_val (KSI->sym_list_p, 0);

  if (token_p (st->tok_buf, "list"))
    return ksi_glob_val (KSI->sym_list, 0);

  if (token_p (st->tok_buf, "apply"))
    return ksi_glob_val (KSI->sym_apply, 0);

  if (token_p (st->tok_buf, "call/cc"))
    return ksi_glob_val (KSI->sym_call_cc, 0);

  if (token_p (st->tok_buf, "call/vs"))
    return ksi_glob_val (KSI->sym_call_vs, 0);

  if (token_p (st->tok_buf, "vector?"))
    return ksi_glob_val (KSI->sym_vector_p, 0);

  ksi_error ("read: %s: unknown read syntax: `#%s' at %d:%d",
	     port_name (st->port), st->tok_buf,
	     st->tok_line, st->tok_pos);
  return 0;
}


static ksi_obj
read_not_impl (read_state *st, int ch)
{
  ksi_error ("read: %s: `%c' syntax not implemented at %d:%d",
	     port_name (st->port), ch, st->tok_line, st->tok_pos);
  return 0;
}


static ksi_obj
ksi_read_obj (read_state *st)
{
  ksi_obj res;
  int c, prev_line = st->tok_line, prev_pos = st->tok_pos;

  CHECK_EVENTS;
  c = skip_ws (st, 1);
  switch (c)
    {
    case -1:
      res = ksi_eof;
      break;

    case '"':
      read_c_string (st, '"', "string");
      res = ksi_str2string (st->tok_buf, st->tok_len);
      if (st->make_const)
	res->itag = KSI_TAG_CONST_STRING;
      res = annotation (res, st);
      break;

    case '#':
      res = read_sharp (st);
      break;

    case '(':
      res = read_list (st, ')', "list");
      break;

    case '[':
      res = read_list (st, ']', "list");
      break;

    case ')':
    case ']':
      ksi_error ("read: %s: unexpected `%c' at %d:%d",
		 port_name (st->port), c, st->tok_line, st->tok_pos);
      /* never reached */

    case '\'':
      res = ksi_read_obj (st);
      if (res == ksi_eof)
        ksi_error ("read: %s: eof while reading a quote at %d:%d",
		   port_name (st->port), st->tok_line, st->tok_pos);

      res = annotation (LIST2 (KSI->sym_quote, res), st);
      break;

    case '`':
      res = ksi_read_obj (st);
      if (res == ksi_eof)
        ksi_error ("read: %s: eof while reading a backquote at %d:%d",
		   port_name (st->port), st->tok_line, st->tok_pos);

      res = annotation (LIST2 (KSI->sym_quasiquote, res), st);
      break;

    case ',':
      c = read_next (st);
      if (c == -1)
        ksi_error ("read: %s: eof while reading an unquote at %d:%d",
		   port_name (st->port), st->tok_line, st->tok_pos);

      if (c == '@')
        {
          res = ksi_read_obj (st);
          if (res == ksi_eof)
            ksi_error ("read: %s: eof while reading ,@ at %d:%d",
		       port_name (st->port), st->tok_line, st->tok_pos);

          res = annotation (LIST2 (KSI->sym_unquote_splicing, res), st);
        }
      else
        {
	  ksi_port_ungetc (st->port, c);
	  res = ksi_read_obj (st);
          if (res == ksi_eof)
            ksi_error ("read: %s: eof while reading an unquote at %d:%d",
		       port_name (st->port), st->tok_line, st->tok_pos);

          res = annotation (LIST2 (KSI->sym_unquote, res), st);
        }
      break;

    case '{': case '}': case '|':
      res = read_not_impl (st, c);
      break;

    default:
      read_token (st, c, !st->case_sens);

      /* number? */
      if (st->tok_buf[0] == '-' || st->tok_buf[0] == '+'
	  || st->tok_buf[0] == '.' || isdigit (st->tok_buf[0]))
	{
	  res = ksi_str02num (st->tok_buf, 0);
	  if (res != ksi_false)
	    {
	      res = annotation (res, st);
	      break;
	    }
	}

      /* keyword? */
      if (st->kw_mode_prefix && st->tok_buf[0] == ':')
	res = annotation (ksi_str2key (st->tok_buf+1, st->tok_len-1), st);
      else if (st->kw_mode_postfix && st->tok_buf[st->tok_len-1] == ':')
	res = annotation (ksi_str2key (st->tok_buf, st->tok_len-1), st);
      else /* it is symbol */
	res = annotation (ksi_intern (st->tok_buf, st->tok_len), st);
    }

  st->tok_line = prev_line;
  st->tok_pos  = prev_pos;
  return res;
}

static ksi_obj
ksi_read_inner (ksi_port p, int annotate, int make_const)
{
  read_state st;
  ksi_obj kw, cs;
  char buf[80];

  cs = ksi_spec_val ("*read-case-sensitive*", ksi_false, 0);
  kw = ksi_spec_val ("*read-keywords*", ksi_false, 0);

  st.port            = p;
  st.sharp_hook      = ksi_spec_val ("*read-sharp-hook*", ksi_false, 0);
  st.tok_buf         = buf;
  st.tok_size        = sizeof buf;
  st.tok_len         = 0;
  st.tok_pos         = p->read_pos;
  st.tok_line        = p->read_line;
  st.kw_mode_prefix  = (kw == KSI->sym_prefix  ? 1 : 0);
  st.kw_mode_postfix = (kw == KSI->sym_postfix ? 1 : 0);
  st.case_sens       = (cs == ksi_false ? 0 : 1);
  st.annotate        = (annotate != 0);
  st.make_const      = (make_const != 0);
  return ksi_read_obj (&st);
}

ksi_obj
ksi_read (ksi_obj p)
{
  if (!p) p = ksi_current_input_port ();
  TYPE_CHECK (p, KSI_INPUT_PORT_P (p), "read: invalid port");

  return ksi_read_inner ((ksi_port) p, 0, 0);
}

ksi_obj
ksi_read_const (ksi_obj p)
{
  if (!p) p = ksi_current_input_port ();
  TYPE_CHECK (p, KSI_INPUT_PORT_P (p), "read: invalid port");

  return ksi_read_inner ((ksi_port) p, 0, 1);
}

static ksi_obj
ksi_read_annotation (ksi_obj p)
{
  if (!p) p = ksi_current_input_port ();
  TYPE_CHECK (p, KSI_INPUT_PORT_P (p), "read-annotation: invalid port");

  return ksi_read_inner ((ksi_port) p, 1, 0);
}

ksi_obj
ksi_str2obj (const char* ptr, int len)
{
  ksi_obj port, str;
  ksi_obj val = ksi_nil, *loc = &val;

  if (ptr == 0)
    return ksi_void;
  if (len == 0)
    return ksi_nil;

  str = ksi_str2string (ptr, len);
  port = (ksi_obj) ksi_new_str_port (str);
  ((ksi_port) port) -> input = 1;

  while (1)
    {
      ksi_obj x = ksi_read (port);
      if (x == ksi_eof)
	{
	  if (PAIR_P (val) && CDR (val) == ksi_nil)
	    return CAR (val);
	  return val;
	}
      *loc = CONS (x, ksi_nil);
      loc = & (CDR (*loc));
    }
}

ksi_obj
ksi_str02obj (const char* ptr)
{
  if (ptr == 0)
    return ksi_void;

  return ksi_str2obj (ptr, strlen (ptr));
}

ksi_obj
ksi_object2string (ksi_obj x)
{
  const char* str = ksi_obj2str (x);
  return ksi_str02string (str);
}


static ksi_obj
ksi_annotation_ref (ksi_obj a, ksi_obj x)
{
  int i = KSI_SINT_COD (x);

  TYPE_CHECK (a, KSI_OBJ_IS (a, KSI_TAG_ANNOTATION),
	      "annotation-ref: invalid annotation in arg1");

  TYPE_CHECK (x, KSI_SINT_P (x) && 0 <= i && i < 3,
	      "annotation-ref: invalid index in arg2");

  return KSI_VEC_REF (a, i);
}


static struct Ksi_Prim_Def defs [] =
{
  {"read",		ksi_read,		KSI_CALL_ARG1, 0},
  {"object->string",	ksi_object2string,	KSI_CALL_ARG1, 1},

  {"read-const",	ksi_read_const,		KSI_CALL_ARG1, 0},
  {"read-annotation",	ksi_read_annotation,	KSI_CALL_ARG1, 0},
  {"annotation-ref",	ksi_annotation_ref,	KSI_CALL_ARG2, 2},

  { 0 }
};

void
ksi_init_read (void)
{
  ksi_ignore (rcsid);
  ksi_reg_unit (defs);
}

 /* End of code */
