/*
 * ksi_shel.c
 * shell utils
 *
 * Copyright (C) 1997-2000,2001,2002, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Fri Dec  5 12:56:37 1997
 * Last Update:   Fri Feb  8 18:25:03 2002
 *
 */

/* Comments:
 *
 */

#include "ksi_int.h"

static char rcsid[] = "$Id: ksi_shel.c,v 1.8 2002/02/08 12:51:37 ksion Exp $";


#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif
#ifdef HAVE_PWD_H
#  include <pwd.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#  include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
#  include <sys/stat.h>
#endif
#ifdef HAVE_STDIO_H
#  include <stdio.h>
#endif

#ifdef __WATCOMC__
#  include <direct.h>
#  define NAMLEN(dirent) strlen((dirent)->d_name)
#elif defined(_MSC_VER)
#  include <io.h>
#  include <direct.h>
#  define F_OK 0
#elif defined(HAVE_DIRENT_H)
#  include <dirent.h>
#  define NAMLEN(dirent) strlen((dirent)->d_name)
#else
#  define dirent direct
#  define NAMLEN(dirent) (dirent)->d_namlen
#  if HAVE_SYS_NDIR_H
#    include <sys/ndir.h>
#  endif
#  if HAVE_SYS_DIR_H
#    include <sys/dir.h>
#  endif
#  if HAVE_NDIR_H
#    include <ndir.h>
#  endif
#endif

#ifndef __WATCOMC__
extern char	**environ;
#endif


#ifndef S_ISCHR
#  define S_ISCHR(m)    (((m) & S_IFMT) == S_IFCHR)
#endif

#ifndef S_ISDIR
#  define S_ISDIR(m)    (((m) & S_IFMT) == S_IFDIR)
#endif

#ifndef S_ISREG
#  define S_ISREG(m)    (((m) & S_IFMT) == S_IFREG)
#endif


int
ksi_has_suffix (const char* fname, const char* suffix)
{
  int l1 = strlen (fname);
  int l2 = strlen (suffix);
  if (l1 < l2)
    return 0;

  while (l2)
    {
#if defined(MSWIN32) || defined(OS2) || defined(MSDOS) \
    || defined(__CYGWIN32__) || defined(__CYGWIN__)

      if (tolower (fname[l1]) != tolower (suffix[l2]))
	return 0;
      --l1; --l2;
#else
      if (fname[--l1] != suffix[--l2])
        return 0;
#endif
    }

  return 1;
}


static const char *
tilde_expand (const char *name)
{
  int len;
  const char *p;
  char *res;

#ifdef unix
  struct passwd *pwPtr;
#endif

  if (name[0] != '~')
    return name;

  if ((name[1] == '/') || (name[1] == '\0'))
    {
      p = getenv ("HOME");
      if (p == NULL)
	p = ".";

      return ksi_aprintf ("%s%s", p, name+1);
    } 

#ifdef unix

  for (p = &name[1]; (*p != 0) && (*p != '/'); p++)
    {
      /* Null body;  just find end of name. */
    }

  len = p - (name+1);
  res = (char*) ksi_malloc_data (len + 1);
  memcpy (res, name+1, len);
  res [len] = '\0';

  pwPtr = getpwnam (res);
  if (pwPtr == NULL)
    {
      endpwent ();
      name = ".";
    }
  else
    name = pwPtr->pw_dir;

  res = ksi_aprintf ("%s%s", name, p);
  endpwent ();

  return res;

#else

  return name;

#endif
}

static char*
get_cwd ()
{
  int len = MAX_PATH_LENGTH;
  char *dir = (char*) ksi_malloc_data (len);

  while (1)
    {
      if (getcwd (dir, len))
	break;
      len *= 2;
      dir = (char *) ksi_realloc (dir, len);
    }

#if defined(MSWIN32) || defined(OS2) || defined(MSDOS)
  {
    int i;
    for (i = 0; dir [i]; i++)
      if (dir [i] == '\\')
	dir [i] = '/';
  }
#endif

  return dir;
}

static char*
absolute_path (char* fname)
{
  char *cwd, *dir, *dst, *src;

#if defined(MSWIN32) || defined(OS2) || defined(MSDOS)
  if (fname [0] == '/' || (fname [0] != '\0' && fname [1] == ':'))
    dir = fname;
#else
  if (fname [0] == '/')
    dir = fname;
#endif
  else
    {
      cwd = get_cwd ();
      dir = (char*) ksi_malloc_data (strlen (fname) + strlen (cwd) + 2);
      strcpy (dir, cwd);
      strcat (dir, "/");
      strcat (dir, fname);
    }

  dst = dir;
  src = dir;
  while (*src)
    {
      if (*src == '/')
	{
	  src++;
	  if (*src == '.')
	    {
	      if (*(src+1) == '/' || *(src+1) == '\0')	/* skip "/./" */
		src++;
	      else if (*(src+1) == '.' && (*(src+2) == '/' || *(src+2) == '\0'))
		{			/* have "/../" */
		  src += 2;
		  while (dst > dir)    /* back up to the previous '/'  */
		    {
		      --dst;
#if defined(MSWIN32) || defined(OS2) || defined(MSDOS)
		      if (*dst == '/' || *dst == ':')
#else
		      if (*dst == '/')
#endif
			break;
		    }
		}
	      else
		*dst++ = '/';
	    }
	  else if (*src != '/')		/* skip "//" */
	    *dst++ = '/';
	}
      else
	*dst++ = *src++;
    }
  *dst = '\0';                /* zero terminator */

  return dir;
}


char*
ksi_expand_file_name (const char* fname)
{
  int len = strlen (fname);
  char* ptr = (char*) ksi_malloc_data (len+1);
  strcpy (ptr, fname);

#if defined(MSWIN32) || defined(OS2) || defined(MSDOS)
  {
    int i;
    for (i = 0; i < len; ++i)
      if (ptr[i] == '\\')
	ptr[i] = '/';
  }
#endif

  ptr = (char*) tilde_expand (ptr);
  ptr = absolute_path (ptr);

#if defined(MSWIN32) || defined(OS2) || defined(MSDOS)
  {
    int i;
    len = strlen (ptr);
    for (i = 0; i < len; ++i)
      {
	if (ptr[i] == '/')
	  ptr[i] = '\\';
	else
	  ptr[i] = tolower (ptr[i]);
      }
  }
#elif defined(__CYGWIN32__) || defined(__CYGWIN__)
  {
    int i;
    len = strlen (ptr);
    for (i = 0; i < len; ++i)
      ptr[i] = tolower (ptr[i]);
  }
#endif

  return ptr;
}

const char*
ksi_mk_filename (ksi_obj x, char* name)
{
  if (KSI_STR_P (x))
    return (char*) KSI_STR_PTR (x);
  if (KSI_SYM_P (x))
    return (char*) KSI_SYM_PTR (x);

  if (KSI_PAIR_P (x))
    {
      if (CDR (x) == ksi_nil)
	return ksi_mk_filename (CAR (x), name);
      else
	{
	  const char* ptr1 = ksi_mk_filename (CAR (x), name);
	  const char* ptr2 = ksi_mk_filename (CDR (x), name);
	  return ksi_aprintf ("%s%s%s", ptr1, DIR_SEP, ptr2);
	}
    }

  ksi_exn_error ("type", x, "%s: invalid file-name", name);
  return 0;
}


static ksi_obj
ksi_getenv (ksi_obj str)
{
  int		i;
  char		*ptr;
  ksi_obj	res, x, y;

  if (str)
    {
      TYPE_CHECK (str, KSI_STR_P (str), "getenv: invalid string in arg1");

      ptr = KSI_STR_PTR (str);
      if (ptr)
	ptr = getenv (ptr);
      res = (ptr ? ksi_str02string (ptr) : ksi_false);
    }
  else
    {
      res = ksi_nil;
      for (i = 0; environ[i] != 0; i++)
	{
	  ptr = strchr (environ[i], '=');
	  if (ptr)
	    {
	      x = ksi_str2string (environ[i], ptr - environ[i]);
	      y = ksi_str02string (ptr + 1);
	      res = CONS (CONS (x, y), res);
	    }
	}
    }

  return res;
}

static ksi_obj
ksi_syscall (ksi_obj str)
{
  char *ptr;
  TYPE_CHECK (str, KSI_STR_P (str), "system: invalid string in arg1");

  ptr = KSI_STR_PTR (str);
  return KSI_MK_SINT (system (ptr));
}


static ksi_obj
ksi_getcwd (void)
{
  char *ptr = get_cwd ();

#if defined(MSWIN32) || defined(OS2) || defined(MSDOS)
  {
    int i;
    for (i = 0; ptr [i]; i++)
      if (ptr [i] == '/')
	ptr [i] = '\\';
  }
#endif

  return ksi_str02string (ptr);
}

static ksi_obj
ksi_chdir (ksi_obj dir)
{
  const char *ptr = ksi_mk_filename (dir, "chdir");
  ptr = tilde_expand (ptr);

  if (chdir (ptr) == 0)
    return ksi_true;

  ksi_exn_error ("system", ksi_str02string (ptr),
		 "chdir: %s", strerror (errno));
  return ksi_false;
}

static ksi_obj
ksi_mkdir (ksi_obj dir, ksi_obj msk)
{
  const char *ptr = ksi_mk_filename (dir, "mkdir");
  ptr = tilde_expand (ptr);

#if defined(MSWIN32) || defined(OS2) || defined(MSDOS)
  if (mkdir (ptr) == 0)
    return ksi_true;
#else
  TYPE_CHECK (msk, !msk || KSI_SINT_P (msk), "mkdir: invalid mode in arg2");
  if (mkdir (ptr, msk ? KSI_SINT_COD (msk) : 0777) == 0)
    return ksi_true;
#endif

  ksi_exn_error ("system", ksi_str02string (ptr),
		 "mkdir: %s", strerror (errno));
  return ksi_false;
}

static ksi_obj
ksi_rmdir (ksi_obj dir)
{
  const char *ptr = ksi_mk_filename (dir, "rmdir");
  ptr = tilde_expand (ptr);

  if (rmdir (ptr) == 0)
    return ksi_true;

  ksi_exn_error ("system", ksi_str02string (ptr),
		 "rmdir: %s", strerror (errno));
  return ksi_false;
}

static ksi_obj
ksi_file_exists (ksi_obj fn)
{
  const char *ptr = ksi_mk_filename (fn, "file-exists?");
  ptr = tilde_expand (ptr);

  if (access (ptr, F_OK) == 0)
    return ksi_true;

  return ksi_false;
}

static ksi_obj
ksi_delete_file (ksi_obj fn)
{
  const char *ptr = ksi_mk_filename (fn, "delete-file");
  ptr = tilde_expand (ptr);

  if (unlink (ptr) == 0)
    return ksi_true;

  ksi_exn_error ("system", ksi_str02string (ptr),
		 "delete-file: %s", strerror (errno));
  return ksi_false;
}

static ksi_obj
ksi_rename_file (ksi_obj old, ksi_obj new)
{
  const char *ptr_old = ksi_mk_filename (old, "rename-file");
  const char *ptr_new = ksi_mk_filename (new, "rename-file");
  ptr_old = tilde_expand (ptr_old);
  ptr_new = tilde_expand (ptr_new);

#ifdef HAVE_RENAME
  if (rename (ptr_old, ptr_new) == 0)
    return ksi_true;
#else
  if (link (ptr_old, ptr_new) == 0)
    {
      if (unlink (ptr_old) == 0)
	return ksi_true;
      /* unlink failed. remove new filename */
      unlink (ptr_new);
    }
#endif

  ksi_exn_error ("system", ksi_str02string (ptr_old),
		 "rename-file: %s", strerror (errno));
  return ksi_false;
}


static ksi_obj
ksi_stat (ksi_obj x)
{
  struct stat st;
  const char *fname;

  fname = ksi_mk_filename (x, "stat");
  fname = tilde_expand (fname);

  if (stat (fname, &st) == 0)
    {
      ksi_obj r = ksi_alloc_vector (15, KSI_TAG_VECTOR);

      KSI_VEC_REF (r, 0)  = ksi_uint2num (st.st_dev);
      KSI_VEC_REF (r, 1)  = ksi_uint2num (st.st_ino);
      KSI_VEC_REF (r, 2)  = ksi_uint2num (st.st_mode);
      KSI_VEC_REF (r, 3)  = ksi_uint2num (st.st_nlink);
      KSI_VEC_REF (r, 4)  = ksi_uint2num (st.st_uid);
      KSI_VEC_REF (r, 5)  = ksi_uint2num (st.st_gid);
#ifdef HAVE_STRUCT_STAT_ST_RDEV
      KSI_VEC_REF (r, 6)  = ksi_uint2num (st.st_rdev);
#else
      KSI_VEC_REF (r, 6)  = ksi_false;
#endif
      KSI_VEC_REF (r, 7)  = ksi_ulong2num (st.st_size);
      KSI_VEC_REF (r, 8)  = ksi_ulong2num (st.st_atime);
      KSI_VEC_REF (r, 9)  = ksi_ulong2num (st.st_mtime);
      KSI_VEC_REF (r, 10) = ksi_ulong2num (st.st_ctime);
#ifdef HAVE_STRUCT_STAT_ST_BLKSIZE
      KSI_VEC_REF (r, 11) = ksi_uint2num (st.st_blksize);
#else
      KSI_VEC_REF (r, 11) = ksi_uint2num (4096);
#endif
#ifdef HAVE_STRUCT_STAT_ST_BLOCKS
      KSI_VEC_REF (r, 12) = ksi_uint2num (st.st_blocks);
#else
      KSI_VEC_REF (r, 12) = ksi_uint2num ((st.st_size + 511) / 512);
#endif

      if (S_ISREG (st.st_mode))
	KSI_VEC_REF (r, 13) = ksi_ptr02sym ("regular");
#ifdef S_ISDIR
      else if (S_ISDIR (st.st_mode))
	KSI_VEC_REF (r, 13) = ksi_ptr02sym ("directory");
#endif
#ifdef S_ISLNK
      else if (S_ISLNK (st.st_mode))
	KSI_VEC_REF (r, 13) = ksi_ptr02sym ("symlink");
#endif
#ifdef S_ISBLK
      else if (S_ISBLK (st.st_mode))
	KSI_VEC_REF (r, 13) = ksi_ptr02sym ("block-special");
#endif
#ifdef S_ISCHR
      else if (S_ISCHR (st.st_mode))
	KSI_VEC_REF (r, 13) = ksi_ptr02sym ("char-special");
#endif
#ifdef S_ISFIFO
      else if (S_ISFIFO (st.st_mode))
	KSI_VEC_REF (r, 13) = ksi_ptr02sym ("fifo");
#endif
#ifdef S_ISSOCK
      else if (S_ISSOCK (st.st_mode))
	KSI_VEC_REF (r, 13) = ksi_ptr02sym ("socket");
#endif
      else
	KSI_VEC_REF (r, 13) = ksi_ptr02sym ("unknown");

      KSI_VEC_REF (r, 14) = ksi_uint2num ((~S_IFMT) & st.st_mode);

      return r;
    }

  ksi_exn_error ("system", ksi_str02string (fname),
		 "stat: %s", strerror (errno));
  return ksi_false;
}


static struct Ksi_Tag tc_dir =
{
  "directory",
  ksi_default_tag_equal,
  ksi_default_tag_print
};

struct Ksi_Dir
{
  struct Ksi_ExtObj	ko;
#ifdef _MSC_VER
  long			hd;
  int			eof;
  struct _finddata_t	file;
#else
  DIR			*dir;
#endif
};

static ksi_obj
ksi_opendir (ksi_obj name)
{
  const char		*dirname;
#ifdef _MSC_VER
  char			*fname;
#else
  DIR			*dir;
#endif
  struct Ksi_Dir	*obj;

  dirname = ksi_mk_filename (name, "opendir");
  dirname = tilde_expand (dirname);

#ifdef _MSC_VER

  fname = (char*) ksi_malloc_data (strlen (dirname) + 6);
  strcpy (fname, dirname);
  strcat (fname, "\\*.*");

  obj = (struct Ksi_Dir*) ksi_malloc (sizeof *obj);
  obj->ko.itag = KSI_TAG_EXTENDED;
  obj->ko.etag = &tc_dir;
  obj->hd      = _findfirst (fname, &obj->file);
  obj->eof     = 0;
  if (obj->hd != -1)
    return (ksi_obj) obj;

#else

  dir = opendir (dirname);
  if (dir)
    {
      obj = (struct Ksi_Dir*) ksi_malloc (sizeof *obj);
      obj->ko.itag = KSI_TAG_EXTENDED;
      obj->ko.etag = &tc_dir;
      obj->dir     = dir;
      return (ksi_obj) obj;
    }

#endif

  return ksi_false;
}

static ksi_obj
ksi_readdir (ksi_obj dir)
{
#ifdef _MSC_VER
  ksi_obj str;
#else
  struct dirent *rd;
#endif

  TYPE_CHECK (dir, KSI_EXT_IS (dir, &tc_dir), "readdir: invalid directory");

#ifdef _MSC_VER

  if (((struct Ksi_Dir*) dir) -> hd == -1)
    KSI_EXN_ERR ("system", dir, "readdir: directory is closed");

  if (((struct Ksi_Dir*) dir)->eof)
    return ksi_eof;

  str = ksi_str02string (((struct Ksi_Dir*) dir)->file.name);

  if (_findnext (((struct Ksi_Dir*) dir)->hd, &((struct Ksi_Dir*) dir)->file) != 0)
    ((struct Ksi_Dir*) dir)->eof = 1;

  return str;

#else

  if (((struct Ksi_Dir*) dir) -> dir == 0)
    KSI_EXN_ERR ("system", dir, "readdir: directory is closed");

  rd = readdir (((struct Ksi_Dir*) dir) -> dir);
  return (rd ? ksi_str2string (rd->d_name, NAMLEN (rd)) : ksi_eof);

#endif
}

static ksi_obj
ksi_closedir (ksi_obj dir)
{
  KSI_CHECK (dir, KSI_EXT_IS (dir, &tc_dir), "closedir: invalid dir");

#ifdef _MSC_VER
  if (((struct Ksi_Dir*) dir) -> hd == -1)
    KSI_EXN_ERR ("system", dir, "closedir: directory is closed");
  _findclose (((struct Ksi_Dir*) dir) -> hd);
  ((struct Ksi_Dir*) dir) -> hd = -1;
#else
  if (((struct Ksi_Dir*) dir) -> dir == 0)
    KSI_EXN_ERR ("system", dir, "closedir: directory is closed");
  closedir (((struct Ksi_Dir*) dir) -> dir);
  ((struct Ksi_Dir*) dir) -> dir = 0;
#endif

  return ksi_void;
}


static ksi_obj
ksi_exp_fname (ksi_obj x, ksi_obj dir)
{
  const char *fname;

  if (dir && dir != ksi_false)
    x = CONS (dir, x);

  fname = ksi_mk_filename (x, "path-list->file-name");
  if (dir != ksi_false)
    fname = ksi_expand_file_name (fname);
  return ksi_str02string (fname);
}

static ksi_obj
ksi_split_fname (ksi_obj x)
{
  char* ptr;
  int len, i;
  ksi_obj res;

  TYPE_CHECK (x, KSI_STR_P (x), "split-file-name: invalid string in arg1");

  ptr = KSI_STR_PTR (x);
  len = KSI_STR_LEN (x);
  res = ksi_nil;
  for (i = 0; i < len; /**/)
    {
#if defined(MSDOS) || defined(MSWIN32) || defined(OS2)
      if (ptr [i] == '\\' || ptr [i] == '/')
#else
      if (ptr [i] == DIR_SEP [0])
#endif
	{
	  res = CONS (ksi_str2string (ptr, i), res);
	  ptr += i+1;
	  len -= i+1;
	  i = 0;
	}
      else
	i++;
    }
  res = CONS (ksi_str2string (ptr, i), res);

  return ksi_reverse_x (res);
}

static ksi_obj
ksi_split_path (ksi_obj x)
{
  char* ptr;
  int len, i;
  ksi_obj res;

  TYPE_CHECK (x, KSI_STR_P (x), "split-path: invalid string in arg1");

  ptr = KSI_STR_PTR (x);
  len = KSI_STR_LEN (x);
  res = ksi_nil;
  for (i = 0; i < len; /**/)
    {
      if (ptr [i] == PATH_SEP [0])
	{
	  res = CONS (ksi_str2string (ptr, i), res);
	  ptr += i+1;
	  len -= i+1;
	  i = 0;
	}
      else
	i++;
    }
  res = CONS (ksi_str2string (ptr, i), res);

  return ksi_reverse_x (res);
}

static ksi_obj
ksi_has_suffix_p (ksi_obj name, ksi_obj suff)
{
  const char *ptr = ksi_mk_filename (name, "file-name-has-suffix?");

  TYPE_CHECK (suff, KSI_STR_P (suff),
	      "file-name-has-suffix?: invalid string in arg2");

  return (ksi_has_suffix (ptr, KSI_STR_PTR (suff)) ? ksi_true : ksi_false);
}


static struct Ksi_Prim_Def defs [] = 
{
  { "getenv",			ksi_getenv,		KSI_CALL_ARG1, 0 },
  { "system",			ksi_syscall,		KSI_CALL_ARG1, 1 },

  { "getcwd",			ksi_getcwd,		KSI_CALL_ARG0, 0 },
  { "chdir",			ksi_chdir,		KSI_CALL_ARG1, 1 },
  { "mkdir",			ksi_mkdir,		KSI_CALL_ARG2, 1 },
  { "rmdir",			ksi_rmdir,		KSI_CALL_ARG1, 1 },

  { "stat",			ksi_stat,		KSI_CALL_ARG1, 1 },

  { "file-exists?",		ksi_file_exists,	KSI_CALL_ARG1, 1 },
  { "delete-file",		ksi_delete_file,	KSI_CALL_ARG1, 1 },
  { "rename-file",		ksi_rename_file,	KSI_CALL_ARG2, 2 },

  { "opendir",			ksi_opendir,		KSI_CALL_ARG1, 1 },
  { "readdir",			ksi_readdir,		KSI_CALL_ARG1, 1 },
  { "closedir",			ksi_closedir,		KSI_CALL_ARG1, 1 },

  { "path-list->file-name",	ksi_exp_fname,		KSI_CALL_ARG2, 1 },
  { "split-file-name",		ksi_split_fname,	KSI_CALL_ARG1, 1 },
  { "split-path",		ksi_split_path,		KSI_CALL_ARG1, 1 },
  { "file-name-has-suffix?",	ksi_has_suffix_p,	KSI_CALL_ARG2, 2 },

  { 0 }
};

void
ksi_init_shel (void)
{
  ksi_ignore (rcsid);
  ksi_reg_unit (defs);
}

 /* End of code */
