/*
 * ksi_time.c
 * time related function
 *
 * Copyright (C) 1997-2000,2001, ivan demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Mon Nov 17 05:38:36 1997
 * Last Update:   Fri Nov  9 23:48:13 2001
 *
 */

/* Comments:
 *
 */

#include "ksi_int.h"

# ifdef HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif

# ifdef TIME_WITH_SYS_TIME
#  include <sys/time.h>
#  include <time.h>
# else
#  ifdef HAVE_SYS_TIME_H
#   include <sys/time.h>
#  else
#   ifdef HAVE_TIME_H
#    include <time.h>
#   endif
#  endif
# endif

#ifdef HAVE_SYS_TIMES_H
# include <sys/times.h>
#endif

#ifdef HAVE_SYS_TIMEB_H
# include <sys/timeb.h>
#endif

static char rcsid[] = "$Id: ksi_time.c,v 1.5 2001/11/25 14:30:01 demakov Exp $";


/* This should be figured out by autoconf.  */
#ifdef CLK_TCK
# define CLKTCK CLK_TCK
# ifdef CLOCKS_PER_SEC
#  if defined (unix) || defined (__unix)
#   define LACK_CLOCK
    /* This is because clock() might be POSIX rather than ANSI.
       This occurs on HP-UX machines */
#  endif
# endif
#else
# ifdef CLOCKS_PER_SEC
#  define CLKTCK CLOCKS_PER_SEC
# else
#  define LACK_CLOCK
#  define CLKTCK 60
# endif
#endif

#ifndef __STDC__
# define time_t long
#endif


#if defined(MSWIN32)

static int proc_times_supported = 0;

static inline double
ft2d (FILETIME *ft)
{
  /* convert FILETIME to seconds */
  return (ft->dwHighDateTime * 4294967296.0 + ft->dwLowDateTime) / 10000000.0;
}
#endif


double
ksi_cpu_time ()
{
#if defined(MSWIN32)

  FILETIME dummy, user;

  if (proc_times_supported)
    {
      GetProcessTimes (GetCurrentProcess (), &dummy, &dummy, &dummy, &user);
      return ft2d (&user);
    }

  return GetTickCount () / 1000.0;

#elif defined(HAVE_TIMES)

  struct tms time_buffer;
  times (&time_buffer);
  return ((double) time_buffer.tms_utime) / CLKTCK;

#elif !defined(LACK_CLOCK)

  return ((double) clock ()) / CLKTCK;

#else

  static time_t base = time (0);
  return time (0) - base;

#endif
}


double
ksi_eval_time ()
{
#if defined(MSWIN32)

  FILETIME dummy, user, sys;
  DWORD    tmp;

  if (proc_times_supported)
    {
      GetProcessTimes (GetCurrentProcess (), &dummy, &dummy, &sys, &user);
      user.dwHighDateTime += sys.dwHighDateTime;
      tmp = user.dwLowDateTime + sys.dwLowDateTime;
      if (tmp < user.dwLowDateTime) user.dwHighDateTime += 1;
      user.dwLowDateTime = tmp;

      return ft2d (&user);
    }

  return GetTickCount () / 1000.0;

#elif defined(HAVE_TIMES)

  struct tms time_buffer;
  times (&time_buffer);
  return ((double) time_buffer.tms_utime + time_buffer.tms_stime) / CLKTCK;

#elif !defined(LACK_CLOCK)

  return ((double) clock ()) / CLKTCK;

#else

  static time_t base = time (0);
  return time (0) - base;

#endif
}


double
ksi_real_time ()
{
#ifdef HAVE_GETTIMEOFDAY

  struct timeval tbuf;
  gettimeofday (&tbuf, 0);
  return (tbuf.tv_usec / 1000000.0) + tbuf.tv_sec;

#elif defined(HAVE_FTIME)

  struct timeb buf;
  ftime (&buf);
  return buf.time + (buf.millitm / 1000.0);

#else

  return time (0);

#endif
}


static ksi_obj
ksi_scm_current_time ()
{
  time_t t = time (0);
  return ksi_long2num (t);
}

static ksi_obj
ksi_scm_cpu_time ()
{
  double d = ksi_cpu_time ();
  return ksi_double2num (d);
}

static ksi_obj
ksi_scm_eval_time ()
{
  double d = ksi_eval_time ();
  return ksi_double2num (d);
}

static ksi_obj
ksi_scm_real_time ()
{
  double d = ksi_real_time ();
  return ksi_double2num (d);
}


static ksi_obj
tm2obj (struct tm *bt)
{
  ksi_obj v = ksi_alloc_vector (9, KSI_TAG_VECTOR);
  KSI_VEC_REF (v, 0) = KSI_MK_SINT (bt->tm_sec);
  KSI_VEC_REF (v, 1) = KSI_MK_SINT (bt->tm_min);
  KSI_VEC_REF (v, 2) = KSI_MK_SINT (bt->tm_hour);
  KSI_VEC_REF (v, 3) = KSI_MK_SINT (bt->tm_mday);
  KSI_VEC_REF (v, 4) = KSI_MK_SINT (bt->tm_mon);
  KSI_VEC_REF (v, 5) = KSI_MK_SINT (bt->tm_year);
  KSI_VEC_REF (v, 6) = KSI_MK_SINT (bt->tm_wday);
  KSI_VEC_REF (v, 7) = KSI_MK_SINT (bt->tm_yday);
  KSI_VEC_REF (v, 8) = KSI_MK_SINT (bt->tm_isdst);

  return v;
}

static int
obj2tm (ksi_obj v, struct tm *bt)
{
  if (KSI_VEC_P (v) && KSI_VEC_LEN (v) == 9
      && KSI_SINT_P (KSI_VEC_REF (v, 1))
      && KSI_SINT_P (KSI_VEC_REF (v, 2))
      && KSI_SINT_P (KSI_VEC_REF (v, 3))
      && KSI_SINT_P (KSI_VEC_REF (v, 4))
      && KSI_SINT_P (KSI_VEC_REF (v, 5))
      && KSI_SINT_P (KSI_VEC_REF (v, 6))
      && KSI_SINT_P (KSI_VEC_REF (v, 7))
      && KSI_SINT_P (KSI_VEC_REF (v, 8)))
    {
      bzero (bt, sizeof *bt);

      bt->tm_sec   = KSI_SINT_COD (KSI_VEC_REF (v, 0));
      bt->tm_min   = KSI_SINT_COD (KSI_VEC_REF (v, 1));
      bt->tm_hour  = KSI_SINT_COD (KSI_VEC_REF (v, 2));
      bt->tm_mday  = KSI_SINT_COD (KSI_VEC_REF (v, 3));
      bt->tm_mon   = KSI_SINT_COD (KSI_VEC_REF (v, 4));
      bt->tm_year  = KSI_SINT_COD (KSI_VEC_REF (v, 5));
      bt->tm_wday  = KSI_SINT_COD (KSI_VEC_REF (v, 6));
      bt->tm_yday  = KSI_SINT_COD (KSI_VEC_REF (v, 7));
      bt->tm_isdst = KSI_SINT_COD (KSI_VEC_REF (v, 8));

      return 1;
    }

  return 0;
}

static ksi_obj
ksi_gmtime (ksi_obj time)
{
  time_t    it;
  struct tm *bt;
  ksi_obj res;

  TYPE_CHECK (time, KSI_REAL_P (time), "gmtime: invalid time");
  
  it = (time_t) ksi_num2long (time);
  bt = gmtime (&it);
  res = (bt ? tm2obj (bt) : ksi_false);

  return res;
}

static ksi_obj
ksi_localtime (ksi_obj time)
{
  time_t    it;
  struct tm *bt;
  ksi_obj res;

  TYPE_CHECK (time, KSI_REAL_P (time), "localtime: invalid time");
  
  it = (time_t) ksi_num2long (time);
  bt = localtime (&it);
  res = (bt ? tm2obj (bt) : ksi_false);

  return res;
}

static ksi_obj
ksi_strftime (ksi_obj format, ksi_obj time)
{
  struct tm bt;
  char *buf;
  int size, ok = obj2tm (time, &bt);

  TYPE_CHECK (format, KSI_STR_P (format), "strftime: invalid string in arg1");
  TYPE_CHECK (time, ok, "strftime: invalid broken-time in arg2");

  size = 64;
  buf = (char*) ksi_malloc_data (size);
  while (strftime (buf, size, KSI_STR_PTR (format), &bt) >= size)
    {
      ksi_free (buf);
      size *= 2;
      buf = (char*) ksi_malloc_data (size);
    }

  return ksi_str02string (buf);
}

static ksi_obj
ksi_mktime (ksi_obj time)
{
  time_t    it;
  struct tm bt;
  int ok = obj2tm (time, &bt);

  TYPE_CHECK (time, ok, "mktime: invalid broken-time");

  it = mktime (&bt);
  return (it == ((time_t) -1) ? ksi_false : ksi_long2num (it));
}


static struct Ksi_Prim_Def defs [] =
{
  { "current-time",	ksi_scm_current_time,	KSI_CALL_ARG0, 0 },
  { "cpu-time",		ksi_scm_cpu_time,	KSI_CALL_ARG0, 0 },
  { "eval-time",	ksi_scm_eval_time,	KSI_CALL_ARG0, 0 },
  { "real-time",	ksi_scm_real_time,	KSI_CALL_ARG0, 0 },
  { "gmtime",		ksi_gmtime,		KSI_CALL_ARG1, 1 },
  { "localtime",	ksi_localtime,		KSI_CALL_ARG1, 1 },
  { "strftime",		ksi_strftime,		KSI_CALL_ARG2, 2 },
  { "mktime",		ksi_mktime,		KSI_CALL_ARG1, 1 },

  { 0 }
};

void
ksi_init_time (void)
{
#if defined(MSWIN32)
  FILETIME x;
  proc_times_supported = GetProcessTimes (GetCurrentProcess (), &x, &x, &x, &x);
#endif

  ksi_ignore (rcsid);
  ksi_reg_unit (defs);
}

 /* End of code */
