/*
 * ksi_vtab.c
 * symtab
 *
 * Copyright (C) 1997-2000,2001, Ivan Demakov
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without a written agreement
 * is hereby granted, provided that the above copyright notice and this
 * paragraph and the following two paragraphs appear in all copies.
 * Modifications to this software may be copyrighted by their authors
 * and need not follow the licensing terms described here, provided that
 * the new terms are clearly indicated on the first page of each file where
 * they apply.
 *
 * IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS
 * DOCUMENTATION, EVEN IF THE AUTHORS HAS BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAS NO OBLIGATIONS
 * TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *
 * Author:        Ivan Demakov <demakov@users.sourceforge.net>
 * Creation date: Fri Aug  1 04:46:01 1997
 * Last Update:   Sat Mar 17 13:17:15 2001
 *
 */

/* Comments:
 *
 */

#include "ksi_int.h"
#include "ksi_vtab.h"
#include "ksi_mod.h"
#include "ksi_comp.h"

static char rcsid[] = "$Id: ksi_vtab.c,v 1.4 2001/03/17 20:07:20 demakov Exp $";


unsigned ksi_primes [] =
{
  17, 41, 113, 349, 1049, 3089, 6247, 8209, 14057, 21089, 31627,
  47431, 71143, 106721, 160073, 240101, 360163, 540217, 810343,
  810343, 1215497, 0
};

#define PRIMES_NUM	(sizeof (ksi_primes) / sizeof (ksi_primes [0]) - 1)
#define REBUILD_COUNT	3
#define MIN_SIZE	(ksi_primes [0])
#define MAX_SIZE	(ksi_primes [PRIMES_NUM-1])


#define HASH_FUN 0

static inline
unsigned
hash_str (const char* str, int len, unsigned n)
{
#if HASH_FUN == 0

  unsigned h = 0;
  while (len)
    h = ((h << 8) + ((unsigned) (str [--len]))) % n;
  return h;

#elif HASH_FUN == 1

  unsigned h = 0;
  while (len)
    h = (h*16777619) ^ ((unsigned) str [--len]);
  return h % n;

#elif HASH_FUN == 2

  unsigned h = 0;
  while (len)
    h = 0x63c63cd9 * h + 0x9c39c33d + ((unsigned) str [--len]);
  return h % n;

#elif HASH_FUN == 3

  if (len == 0)
    return 0;

#define HASHC h = *str++ + 65599 * h
  unsigned h = 0;
  unsigned loop = (len + 8 - 1) >> 3;
  switch (len & (8 - 1))
    {
    case 0:
      do {  HASHC;
      case 7: HASHC;
      case 6: HASHC;
      case 5: HASHC;
      case 4: HASHC;
      case 3: HASHC;
      case 2: HASHC;
      case 1: HASHC;
      } while (--loop);
    }
  return h % n;

#else

  register unsigned h = 0;
  while (len)
    h += (h << 3) + ((unsigned) str [--len]);
  return h % n;

#endif
}

unsigned
ksi_hash_str (const char* str, int len, unsigned n)
{
  return hash_str (str, len, n);
}


ksi_valtab
ksi_new_valtab (unsigned sz, ksi_hash_fun h, ksi_cmp_fun c, void* d)
{
  ksi_valtab tab = (ksi_valtab) ksi_malloc (sizeof *tab);

  tab -> count = 0;
  tab -> hash = h;
  tab -> cmp = c;
  tab -> data = d;

  if (sz < MIN_SIZE)
    tab -> size = 1;
  else
    {
      register int i;
      for (i = 0; i < PRIMES_NUM && ksi_primes [i] < sz; ++i);
      tab -> size = ksi_primes [i];
    }

  tab -> table = ksi_malloc (tab -> size * sizeof *tab->table);

  return tab;
}

void
ksi_clear_vtab (ksi_valtab tab)
{
  ksi_free (tab -> table);

  tab -> count = 0;
  tab -> size  = 1;
  tab -> table = ksi_malloc (tab -> size * sizeof *tab->table);
}

void*
ksi_lookup_vtab (ksi_valtab tab, void* val)
{
  register struct Ksi_Tabrec* rec;
  unsigned index = (tab -> size <= 1
		    ? 0 : tab -> hash (val, tab -> size, tab -> data));

  KSI_ASSERT (index < tab -> size);

  for (rec = tab -> table [index]; rec; rec = rec -> next)
    {
      if (tab -> cmp (rec -> val, val, tab -> data) == 0)
        return rec -> val;
    }

  return 0;
}

void
ksi_append_vtab (ksi_valtab tab, void* val)
{
  register struct Ksi_Tabrec* rec;
  unsigned index = (tab -> size <= 1
		    ? 0 : tab -> hash (val, tab -> size, tab -> data));

  KSI_ASSERT (index < tab -> size);

  rec = (struct Ksi_Tabrec*) ksi_malloc (sizeof *rec);
  rec -> val = val;

  rec -> next = tab -> table [index];
  tab -> table [index] = rec;
  tab -> count += 1;

  if (tab -> count >= MIN_SIZE
      && tab -> count > tab -> size * REBUILD_COUNT
      && tab -> size < MAX_SIZE)
    ksi_rebuild_vtab (tab, tab -> size);
}

void*
ksi_remove_vtab (ksi_valtab tab, void* val)
{
  register struct Ksi_Tabrec *rec, **ptr;
  unsigned index = (tab -> size <= 1
		    ? 0 : tab -> hash (val, tab -> size, tab -> data));

  KSI_ASSERT (index < tab -> size);

  ptr = & tab -> table [index];
  for (rec = tab -> table [index]; rec; rec = rec -> next)
    {
      if (tab -> cmp (rec -> val, val, tab -> data) == 0)
	{
	  val = rec -> val;
	  tab -> count -= 1;

	  *ptr = rec->next;
	  return val;
	}
      ptr = & rec->next;
    }

  return 0;
}

void*
ksi_iterate_vtab (ksi_valtab tab, ksi_iter_fun fun, void* iter_data)
{
  register int i;
  register struct Ksi_Tabrec* rec;

  for (i = 0; i < tab -> size; i++)
    for (rec = tab -> table [i]; rec; rec = rec -> next)
      if (fun (rec -> val, iter_data))
	return rec -> val;

  return 0;
}

void
ksi_rebuild_vtab (ksi_valtab tab, unsigned sz)
{
  struct Ksi_Tabrec *rec, **newtab;
  unsigned i, index;

  if (sz >= MAX_SIZE)
    return;

  for (i = 0; ksi_primes [i] <= sz; ++i);
  sz = ksi_primes [i];

  newtab = (struct Ksi_Tabrec**) ksi_malloc (sz * sizeof *newtab);

  for (i = 0; i < tab -> size; ++i)
    {
      while (tab -> table [i])
        {
          rec = tab -> table [i];
          tab -> table [i] = rec -> next;

          index = tab -> hash (rec -> val, sz, tab -> data);
	  KSI_ASSERT (index < tab -> size);

          rec -> next = newtab [index];
          newtab [index] = rec;
        }
    }

  ksi_free (tab -> table);
  tab -> table = newtab;
  tab -> size = sz;
}


/* ------------------------------------------------------------------ */

static unsigned
hash_sym (void* obj, unsigned num, void* unused)
{
  return hash_str (KSI_SYM_PTR (obj), KSI_SYM_LEN (obj), num);
}

static int
cmp_sym (void* x1, void* x2, void* unused)
{
  int len1 = SYM_LEN (x1);
  int len2 = SYM_LEN (x2);
  const char* str1 = SYM_PTR (x1);
  const char* str2 = SYM_PTR (x2);

#if 1
  int cmp = memcmp (str1, str2, MIN (len1, len2));
  return cmp ? cmp : len1 - len2;
#else
  while (len1 > 0 && len2 > 0)
    {
      int cmp = *str1++ - *str2++;
      if (cmp)
	return cmp;
      --len1; --len2;
    }
  return len1 - len2;
#endif
}


ksi_valtab
ksi_new_symtab (unsigned init_size)
{
  ksi_valtab tab = ksi_new_valtab (init_size, hash_sym, cmp_sym, 0);
  return tab;
}

ksi_symbol
ksi_lookup_sym (ksi_valtab tab, const char* key, int len)
{
  struct Ksi_Symbol sym;

  sym . len = len;
  sym . ptr = key;

  return (ksi_symbol) ksi_lookup_vtab (tab, &sym);
}

void
ksi_append_sym (ksi_valtab tab, ksi_symbol sym)
{
  ksi_append_vtab (tab, sym);
}

ksi_obj
ksi_abbrev_sym (ksi_valtab tab, const char* key, int len)
{
  ksi_obj list = ksi_nil;

  if (key && len >= 0)
    {
      struct Ksi_Tabrec *rec;
      int i;

      for (i = 0; i < tab -> size; i++)
	for (rec = tab -> table [i]; rec; rec = rec -> next)
	  {
	    ksi_symbol sym = (ksi_symbol) rec -> val;
	    if (len <= sym->len && memcmp (key, sym->ptr, len) == 0)
	      list = CONS (sym, list);
	  }
    }

  return list;
}

 
/* ------------------------------------------------------------------- */

static unsigned
hash_key (void* obj, unsigned num, void* unused)
{
  return hash_str (KSI_KEY_PTR (obj), KSI_KEY_LEN (obj), num);
}

static int
cmp_key (void* x1, void* x2, void* unused)
{
  int len1 = KSI_KEY_LEN (x1);
  int len2 = KSI_KEY_LEN (x2);
  const char* str1 = KSI_KEY_PTR (x1);
  const char* str2 = KSI_KEY_PTR (x2);

  int cmp;
  while (len1 > 0 && len2 > 0)
    {
      if ((cmp = *str1++ - *str2++) != 0)
	return cmp;
      --len1;
      --len2;
    }
  return len1 - len2;
}


ksi_valtab
ksi_new_keytab (unsigned init_size)
{
  return ksi_new_valtab (init_size, hash_key, cmp_key, 0);
}

ksi_keyword
ksi_lookup_key (ksi_valtab tab, const char* ptr, int len)
{
  struct Ksi_Keyword key;
  key . len = len;
  key . ptr = ptr;

  return (ksi_keyword) ksi_lookup_vtab (tab, &key);
}

void
ksi_append_key (ksi_valtab tab, ksi_keyword key)
{
  ksi_append_vtab (tab, key);
}


/* ------------------------------------------------------------------- */


static void
tab_stat (ksi_obj out, char* tabname, ksi_valtab tab)
{
  struct Ksi_Tabrec* rec;
  int i, num, worse = 0, wn = 0, ones = 0, zeros = 0, miss = 0;

  ksi_printf (out, "%s %dcount/%dsize=%.2f ",
	      tabname, tab->count, tab->size, (double) tab->count / tab->size);

  for (i = 0; i < tab->size; i++)
    {
      num = 0;
      for (rec = tab->table [i]; rec; rec = rec->next)
	num++;
      if (num > worse)
	worse = num, wn = 1;
      else if (num == worse)
	wn++;
      if (num > 1)
	miss += num-1;
      else if (num == 1)
	ones++;
      else if (num == 0)
	zeros++;
    }

  ksi_printf (out, "%dempty %done %dworse(%d) %dmiss\n",
	      zeros, ones, wn, worse, miss);
}

static ksi_obj
ksi_tab_stat ()
{
  ksi_obj out = ksi_current_output_port ();
  if (KSI_OUTPUT_PORT_P (out))
    {
      tab_stat (out, "symtab", KSI->obarray);
      tab_stat (out, "keytab", KSI->keyarray);

      ksi_flush_port (out);
    }

  return ksi_void;
}



static struct Ksi_Prim_Def defs [] =
{
  { "ksi:tab-stat",	ksi_tab_stat,	KSI_CALL_ARG0, 0 },

  { 0 }
};

/*
 * ksi_init_symtab
 *	  ksi_init_interp  .
 */

void
ksi_init_symtab (void)
{
  SYMTAB_CHECK_INIT;
  KEYTAB_CHECK_INIT;

  ksi_ignore (rcsid);

  ksi_reg_unit (defs);
}


/* End of code */
